/* lexcomp -- phrase composition from lexicon */

/* Mark Huckvale - University College London */

/* version 1.0 - March 1998 */

#define PROGNAME "lexcomp"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"

/* input file */
char		filename[256];	/* input file (from sdump -H) */

/* lexicon file */
char		lfilename[256];	/* lexicon file (from lexparse) */
NSL_File	lexf;
NSL_Doctype	ldct=NULL;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;
const char *	attrHREF;
const char *	attrSTRENGTH;

/* token to process */
char	*token;

/* word records */
struct word_rec {
	char	*str;
	int	flags;
	NSL_Item *sylseq;
	struct word_rec *next;
};

#define WFLAG_STRESS	1
struct word_rec *wroot;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		LTNSL_ERROR(0,"out of memory\n");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* check a word for stress mark */
int	checkword(char *src)
{
	int	stress=0;
	char	*dst=src;

	while (*src) {
		if (*src=='`')
			stress=WFLAG_STRESS;
		else
			*dst++ = *src;
		src++;
	}
	*dst='\0';
	return(stress);
}

/* check for a word match */
int	matchword(char *str)
{
	struct word_rec *w=wroot;

	while (w) {
		if (strcmp(w->str,str)==0) return(1);
		w = w->next;
	}
	return(0);
}

/* process a syllable sequence */
void process_sylseq(char *hw,NSL_Item *sylseqitem)
{
	struct word_rec *w=wroot;

	while (w) {
		if (strcmp(w->str,hw)==0)
			w->sylseq = CopyItem(sylseqitem);
		w = w->next;
	}
}

/* process a pronunciation */
void process_pron(char *hw,NSL_Item *prnitem)
{
	NSL_Data	*data;
	NSL_Item	*pitem;

	/* search for first PRON */
	for (data=prnitem->data;data!=NULL;data=data->next) {
		if (data->type==NSL_item_data) {
			pitem = data->first;
			if (strcmp(pitem->label,"SYLSEQ")==0) {
				/* found syllable sequence */
				process_sylseq(hw,pitem);
				break;
			}
		}
	}
}

/* process a pronunciation sequence */
void process_pron_seq(char *hw,NSL_Item *seqitem)
{
	NSL_Data	*data;
	NSL_Item	*pitem;

	/* search for first PRON */
	for (data=seqitem->data;data!=NULL;data=data->next) {
		if (data->type==NSL_item_data) {
			pitem = data->first;
			if (strcmp(pitem->label,"PRON")==0) {
				/* found pronunciation sequence */
				process_pron(hw,pitem);
				break;
			}
		}
	}
}

/* find words in lexicon */
void findwords()
{
	NSL_Bit		*bit;
	NSL_Item	*item;
	NSL_Data	*data;
	NSL_Item	*pitem;
	NSL_Data	*pdata;
	int		isreq;
	char		*hw;

	/* read first tag - should be <LEXICON> */
	if ((bit=GetNextBit(lexf))==NULL)
		LTNSL_ERROR(0,"Could not read <LEXICON> start tag");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"LEXICON")))
		LTNSL_ERROR(0,"Could not read <LEXICON> start tag");
	FreeBit(bit);

	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(lexf))!=NULL) {
		
		if (bit->type==NSL_start_bit) {
			
			/* should be an entry */
			item=ItemParse(lexf,bit->value.item);
			if (strcmp(item->label,"ENTRY"))
				LTNSL_ERROR(0,"Expected <ENTRY> tag not found");

			/* check HEADWORD entry */
			isreq = 0;
			for (data=item->data;data!=NULL;data = data->next) {
				if (data->type==NSL_item_data) {
					pitem = data->first;
					if (strcmp(pitem->label,"HW")==0) {
						/* found headword entry */
						pdata = pitem->data;
						if (pdata->type==NSL_text_data) {
							isreq = matchword(pdata->first);
							hw = pdata->first;
							break;
						}
					}
				}
			}

			if (isreq) {
				/* found match to headword */
				/* check process PRONSEQ entry */
				for (data=item->data;data!=NULL;data = data->next) {
					if (data->type==NSL_item_data) {
						pitem = data->first;
						if (strcmp(pitem->label,"PRONSEQ")==0) {
							/* found pronunciation sequence */
							process_pron_seq(hw,pitem);
							break;
						}
					}
				}
			}

			FreeItem(item);
		}
		else
			FreeBit(bit);
	}

}

/* compose from lexicon */
NSL_Item *compose(char *token)
{
	char		*p;
	struct word_rec *w;
	struct word_rec **wlast=&wroot;
	NSL_Item	*ipitem;
	NSL_Item	*agitem;
	NSL_Item	*ftitem;
	NSL_Data	*data;
	const char	*str;
	
	/* split token into words */
	p = strtok(token," ");
	while (p && *p) {
		/* get a new word record */
		if ((w = (struct word_rec *)calloc(1,sizeof(struct word_rec)))==NULL)
			LTNSL_ERROR(0,"out of memory");
		w->str = strsave(p);
		w->flags = checkword(w->str);
		w->sylseq = NULL;
		w->next = NULL;

		/* add to end of chain */
		*wlast = w;
		wlast = &w->next;
		p = strtok(NULL," ");
	}

#if 0
	w = wroot;
	while (w) {
		printf("str='%s' flags=%d\n",w->str,w->flags);
		w = w->next;
	}
#endif

	/* match words against lexicon */
	findwords();

	/* create new intonational phrase */
	ipitem = newItem("IP");

	/* create new accent group */
	agitem = newItem("AG");

	/* add words into FOOT */
	ftitem = newItem("FOOT");
	w = wroot;
	while (w) {
		/* check we found word in lexicon */
		if (!w->sylseq)
			LTNSL_ERROR1(0,"No entry in lexicon for '%s'",w->str);
		/* process each syllable in turn */
		for (data=w->sylseq->data;data!=NULL;data=data->next) {
			if (data->type==NSL_item_data) {
				str = GetAttrStringVal(data->first,attrSTRENGTH);
				/* check if phrase & lexically stressed */
				if ((w->flags & WFLAG_STRESS) &&
				    (strcmp(str,"STRONG")==0)) {

					/* start a new foot */
					if (ftitem->type!=NSL_empty) {
						/* save old one */
						AddItemToEnd(agitem,ftitem);
						ftitem = newItem("FOOT");
					}
				}
				else {
					/* set strength to WEAK */
					PutAttrVal(data->first,attrSTRENGTH,"WEAK");
				}
				/* add syllable to foot */
				AddItemToEnd(ftitem,data->first);
			}
		}
		w = w->next;
	}

	if (ftitem->type != NSL_empty)
		AddItemToEnd(agitem,ftitem);

	AddItemToEnd(ipitem,agitem);

	return(ipitem);

}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	inf;
	NSL_Doctype 	idct=NULL;
	NSL_Item 	*item;
	NSL_Item 	*anitem;
	NSL_Item 	*ipitem;
	NSL_Data	*data;
	NSL_Bit		*bit;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Il:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: phrase composition from lexicon Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'l':	/* lexicon name */
		strcpy(lfilename,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		LTNSL_ERROR1(0,"usage: %s (-I) -l lexicon.xml input.xml (out.xml)",PROGNAME);

	/* get the name of the DTD */
	strcpy(dtdname,"/");
	if (argv[0][1]==':')
		strcat(dtdname,&argv[0][2]);
	else
		strcat(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((inf = SFopen(dtdname, odct, NSL_read))==NULL)
		LTNSL_ERROR1(0,"unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(inf);
	SFclose(inf);
	
	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, idct, NSL_read))==NULL)
			LTNSL_ERROR1(0,"unable to open '%s'",filename);
	}
	else
		LTNSL_ERROR(0,"no input file specified");
	idct = DoctypeFromFile(inf);

	/* Open the output NSGML stream using the segparse doctype declaration */
	optind++;
	if (optind < argc)
		outf = SFopen(argv[optind], odct, NSL_write_pretty);
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* get output attribute names */
	attrHREF = AttrUniqueName(odct,"HREF",0);

	/* open the lexicon */
	if (lfilename[0]=='\0')
		LTNSL_ERROR(0,"no lexicon specified");
	if ((lexf = SFopen(lfilename,ldct,NSL_read))==NULL)
		LTNSL_ERROR1(0,"unable to open '%s'",lfilename);
	ldct = DoctypeFromFile(lexf);
	attrSTRENGTH = AttrUniqueName(ldct,"STRENGTH",0);

	/* read first tag - should be <SFSFILE> */
	if ((bit=GetNextBit(inf))==NULL)
		LTNSL_ERROR(0,"Could not read <SFSFILE> start tag");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"SFSFILE")))
		LTNSL_ERROR(0,"Could not read <SFSFILE> start tag");
	FreeBit(bit);

	/* create output annotation referenced to input file */
	anitem = newItem("AN");
	PutAttrVal(anitem,attrHREF,filename);

	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(inf))!=NULL) {
		if ((bit->type==NSL_start_bit) || (bit->type==NSL_empty_bit)) {
			/* scan entries looking for FILE & TOKEN */
			item=ItemParse(inf,bit->value.item);
			if (strcmp(item->label,"FILE")==0) {
				
				/* add the FILE entry straight */
				AddItemToEnd(anitem,CopyItem(item));
			}
			else if (strcmp(item->label,"TOKEN")==0) {

				/* add the TOKEN entry straight */
				AddItemToEnd(anitem,CopyItem(item));
				
				/* process the TOKEN data */
				data = item->data;
				if (data->type==NSL_text_data) {
					token = strsave(data->first);
					ipitem = compose(token);
					AddItemToEnd(anitem,ipitem);
					free(token);
				}
			}
			/* else skip item */
			FreeItem(item);
		}
		else
			FreeBit(bit);
	}

	/* print the annotation */
	PrintItem(outf,anitem);

	/* that's all folks */
	SFclose(inf);
	SFclose(lexf);
	SFclose(outf);
	return 0;
}
