/* lexparse -- add syllable structure to lexicon */

/* Mark Huckvale - University College London */

/* version 1.0 - February 1998 */

/* version 1.1 - April 2001
	- use generic samparse module
*/

#define PROGNAME "lexparse"
#define PROGVERS "1.1"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"
#include "samparse.h"

/* input file */
char		filename[256];

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;

/* tag values */
const char	*tagIPA;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* process a pronunciation */
void process_pron(NSL_Item *ppitem)
{
	NSL_Data	*ppdata;
	NSL_Item	*pppitem;
	NSL_Item	*seqitem;

	/* found a pronunciation */
	for (ppdata=ppitem->data;ppdata!=NULL;ppdata=ppdata->next) {
		if (ppdata->type==NSL_item_data) {
			pppitem = ppdata->first;
			if (strcmp(pppitem->label,"IPA")==0) {
				/* found a transcription */
				seqitem = segparse(pppitem->data->first);
				AddItemToEnd(ppitem,seqitem);
			}
		}
	}
}

/* process a pronunciation sequence */
void process_pron_seq(NSL_Item *pitem)
{
	NSL_Data	*pdata;
	NSL_Item	*ppitem;

	/* found pronunciation sequence */
	for (pdata=pitem->data;pdata!=NULL;pdata = pdata->next) {
		if (pdata->type==NSL_item_data) {
			ppitem = pdata->first;
			if (strcmp(ppitem->label,"PRON")==0)
				process_pron(ppitem);
		}
	}
}


/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	inf;
	NSL_Doctype 	idct=NULL;
	NSL_Item 	*item;
	NSL_Item 	*pitem;
	NSL_Data	*data;
	NSL_Bit		*bit;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"I"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: parse lexical transcription Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) file.lex.xml (out.lex.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* initialise the segment parser with the DTD */
	odct = initparse(dtdname);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, idct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else
		error("no input file specified");
	idct = DoctypeFromFile(inf);

	/* Open the output NSGML stream using the segparse doctype declaration */
	optind++;
	if (optind < argc)
		outf = SFopen(argv[optind], odct, NSL_write_pretty);
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");
	
	/* read first tag - should be <LEXICON> */
	if ((bit=GetNextBit(inf))==NULL)
		error("Could not read <LEXICON> start tag");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"LEXICON")))
		error("Could not read <LEXICON> start tag");
	PrintBit(outf,bit);
	FreeBit(bit);

	/* get references to elements and attributes */
	tagIPA = ElementUniqueName(idct,"IPA",0);

	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(inf))!=NULL) {
		if (bit->type==NSL_start_bit) {
			/* should be an entry */
			item=ItemParse(inf,bit->value.item);
			if (strcmp(item->label,"ENTRY"))
				error("Expected <ENTRY> tag not found");

			/* find and process PRON entries */
			for (data=item->data;data!=NULL;data = data->next) {
				if (data->type==NSL_item_data) {
					pitem = data->first;
					if (strcmp(pitem->label,"PRONSEQ")==0) {
						/* found pronunciation sequence */
						process_pron_seq(pitem);
					}
				}
			}
	
			/* print extended entry */
			PrintItem(outf,item);
			FreeItem(item);
		}
		else
			PrintBit(outf,bit);
		FreeBit(bit);
	}

	/* that's all folks */
	SFclose(outf);
	return 0;
}
