/* phnparse -- parse a marked-up SAMPA transcription */

/* Mark Huckvale - University College London */

/* version 1.0 - February 2000 */

/*
 * allowed markup:
 *	/ = intonation phrase boundary
 *	` = accent group and foot boundary before this syllable
 *	\ = foot boundary (but not accent group) before this syllable
 */

#define PROGNAME "phnparse"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "segment.h"
#include "sylonoff.h"
#include "local.h"

/* DTD */
static NSL_Doctype 	odct=NULL;

/* tag values */
static const char	*tagIP;
static const char	*tagAG;
static const char	*tagFOOT;
static const char	*tagSYL;
static const char	*tagONSET;
static const char	*tagRHYME;
static const char	*tagNUC;
static const char	*tagCODA;
static const char	*tagVOC;
static const char	*tagCNS;

/* attribute values */
static const char	*attrID;
static const char	*attrAMBI;
static const char	*attrSTRENGTH;
static const char	*attrWEIGHT;
static const char	*attrCHECKED;
static const char	*attrLONG;
static const char	*attrCNT;
static const char	*attrSON;
static const char	*attrRHO;
static const char	*attrCNSGRV;
static const char	*attrCNSCMP;
static const char	*attrVOCGRV;
static const char	*attrVOCHEIGHT;
static const char	*attrVOCRND;
static const char	*attrVOI;
static const char	*attrGRV;
static const char	*attrHEIGHT;
static const char	*attrRND;
const char	*attrFPOS;
const char	*attrRFPOS;

/* input buffer */
#define MAXBUFLEN	4096
char	buf[MAXBUFLEN];

/* coded record of onsets and offsets */
static int		onset[NUMSAMPAONSET][5];
static int		offset[NUMSAMPAOFFSET][6];

/* record of a segment */
struct segment_entry {
	int	idx;
};

/* duplicate a string */
static char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
static NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* look up a segment name */
static int findsegment(const char *sym)
{
	int	lo,hi,k,c;
	char	buf[6];
	int	len;

	strncpy(buf,sym,4);
	for (len=3;len>=1;len--) {
		buf[len]='\0';
		lo = 0;
		hi = (sizeof(segtab)/sizeof(struct segment_rec))-1;
		do {
			k = (lo+hi)/2;
			c = strcmp(buf,segtab[k].sym);
			if (c < 0)
				hi = k-1;
			else if (c > 0)
				lo = k+1;
			else
				return(k);
		} while (lo <= hi);
	}
	error("unknown symbol: '%s'",sym);
	return(0);
}

/* load SAMPA onsets & offsets */
static void loadsylonoff()
{
	int	i,j;
	const char	*str;

	for (i=0;i<NUMSAMPAONSET;i++) {
		str = SAMPAonset[i];
		for (j=0;(j < 5) && *str;j++) {
			onset[i][j] = findsegment(str);
			str += strlen(segtab[onset[i][j]].sym);
		}
	}
	for (i=0;i<NUMSAMPAOFFSET;i++) {
		str = SAMPAoffset[i];
		for (j=0;(j < 6) && *str;j++) {
			offset[i][j] = findsegment(str);
			str += strlen(segtab[offset[i][j]].sym);
		}
	}
}

/* get consonant sequence */
static char *getconseq(char *data,struct segment_entry *seq,int *eof,int *stress)
{
	int		idx;

	*stress = 0;
	while (data && *data) {
		/* find segment */
		idx = findsegment(data);
		if (segtab[idx].attr & SKIP) {
			/* check stress indicator */
			if (segtab[idx].attr & STRESS1) *stress |= STRESS1;
			if (segtab[idx].attr & STRESS2) *stress |= STRESS2;
			if (segtab[idx].attr & FTBOUND) *stress |= FTBOUND;
			if (segtab[idx].attr & AGBOUND) *stress |= AGBOUND;
			if (segtab[idx].attr & IPBOUND) *stress |= IPBOUND;
			if (segtab[idx].attr & SYLBOUND) {
				seq->idx = 0;
				*eof = 0;
				data += strlen(segtab[idx].sym);
				return(data);
			}
		}
		else {
			if (segtab[idx].attr & NUC) {
				seq->idx = 0;
				*eof = 0;
				return(data);
			}
			seq->idx = idx;
			seq++;
		}
		data += strlen(segtab[idx].sym);
	}
	seq->idx = 0;
	*eof = 1;
	return(data);
}

/* get nuclear sequence */
static char *getnucleus(char *data,struct segment_entry *seq,int *stress)
{
	int	idx;

	*stress=0;
	if (!data || !*data) {
		seq->idx = 0;
		return(NULL);
	}
	else {
		do {
			idx = findsegment(data);
			if (segtab[idx].attr & SKIP) {
				/* check stress indicator */
				if (segtab[idx].attr & STRESS1) *stress |= STRESS1;
				if (segtab[idx].attr & STRESS2) *stress |= STRESS2;
				if (segtab[idx].attr & FTBOUND) *stress |= FTBOUND;
				if (segtab[idx].attr & AGBOUND) *stress |= AGBOUND;
				if (segtab[idx].attr & IPBOUND) *stress |= IPBOUND;
			}
			else
				break;
			data += strlen(segtab[idx].sym);
		} while (*data);
		if (!*data) { seq->idx=0; return(NULL); }
		if (segtab[idx].attr & NUC) {
			seq->idx = idx;
			seq++;
			seq->idx = 0;
			data += strlen(segtab[idx].sym);
			if (!*segtab[idx].elem2) {
				/* whoops! only half a vowel */
				idx = findsegment(data);
				if (((segtab[idx].attr & NUC)==0)||(*segtab[idx].elem2))
					error("unmatched half-vowel segment found");
				seq->idx = idx;
				seq++;
				seq->idx = 0;
				data += strlen(segtab[idx].sym);
			}
			return(data);
		}
		seq->idx = 0;
		return(data);
	}
}

/* check segment sequence overlap */
static int overlapf(struct segment_entry *sep,int *cp)
{
	int	cnt=0;
	while ((sep->idx) && *cp && (sep->idx==*cp)) { cnt++; sep++; cp++; }
	return(cnt);
}
static int overlapr(struct segment_entry *sep,int *cp)
{
	int	cnt=0;
	struct segment_entry *sep2;
	int	*cp2;

	for (sep2=sep;sep2[0].idx && sep2[1].idx;sep2++) /* skip */;
	for (cp2=cp;*cp2 && *(cp2+1);cp2++) /* skip */;
	while ((sep2>=sep) && (cp2>=cp) && (sep2->idx==*cp2)) { cnt++; sep2--; cp2--; }
	return(cnt);
}

/* parse consonant sequence */
static void parseconseq(struct segment_entry *cseq,
	    struct segment_entry *cseq1,
	    struct segment_entry *ambi1,
	    struct segment_entry *ambi2,
	    struct segment_entry *cseq2)
{
	int	i,len;
	int	onmax,offmax;
	int	nambi;

	/* find length */
	for (len=0;cseq[len].idx!=0;len++) /* skip */;

	/* find longest matching offset */
	offmax=0;
	for (i=0;i<NUMSAMPAOFFSET;i++) {
		if (overlapf(cseq,offset[i])>offmax) {
			offmax = overlapf(cseq,offset[i]);
		}
	}
	/* find longest matching onset */
	onmax=0;
	for (i=0;i<NUMSAMPAONSET;i++) {
		if (overlapr(cseq,onset[i])>onmax) {
			onmax = overlapr(cseq,onset[i]);
		}
	}

	/* find overlap between onset and offset */
	nambi = onmax + offmax - len;

	/* copy relevant parts */
	for (i=0;i<len-onmax;i++) {
		cseq1[0] = cseq[i];
		cseq1++;
	}
	for (;i<len-onmax+nambi;i++) {
		ambi1[0] = ambi2[0] = cseq[i];
		ambi1++;
		ambi2++;
	}
	for (;i<len;i++) {
		cseq2[0] = cseq[i];
		cseq2++;
	}

	/* terminate strings */
	cseq1[0].idx=0;
	ambi1[0].idx=0;
	ambi2[0].idx=0;
	cseq2[0].idx=0;

}

/* check an attribute is Y or N */
static int 	checkAttr(NSL_Item *item,const char *name)
{
	const char	*val;

	val = GetAttrStringVal(item,name);
	if (val && *val=='Y')
		return(1);
	else
		return(0);
}

/* check if a Nucleus is heavy */
static int	heavyNUC(NSL_Item *nucitem)
{
	const char	*tagval;

	/* see if has weight set */
	tagval = GetAttrStringVal(nucitem,attrWEIGHT);
	if (tagval && strcmp(tagval,"HEAVY")==0) return(1);

	/* see if long */
	if (checkAttr(nucitem,attrLONG)) return(1);

	return(0);
}

/* check if a CODA is heavy */
static int	heavyCODA(NSL_Item *codaitem)
{
	NSL_Data	*data;
	NSL_Item	*item;
	int		ncons=0;

	/* see if has 2 consonants */
	for (data=codaitem->data;data;data=data->next) {
		if (data->type!=NSL_item_data) return(0);
		item=data->first;
		if (item->label==tagCNS) ncons++;
	}
	if (ncons < 2)
		return(0);
	else
		return(1);
}

/* check if a Rhyme is heavy */
static int	heavyRHYME(NSL_Item *rhymeitem)
{
	NSL_Data	*data;
	NSL_Item	*item;
	const char	*tagval;

	/* see if weight set already */
	tagval = GetAttrStringVal(rhymeitem,attrWEIGHT);
	if (tagval && strcmp(tagval,"HEAVY")==0) return(1);

	/* check daughters */
	for (data=rhymeitem->data;data;data=data->next) {
		if (data->type!=NSL_item_data) return(0);
		item=data->first;
		/* check NUC */
		if (item->label==tagNUC)
			if (heavyNUC(item)) return(1);
		/* check CODA */
		if (item->label==tagCODA)
			if (heavyCODA(item)) return(1);
	}
	return(0);
}

/* check if a Syllable is heavy */
static int	heavySYL(NSL_Item *sylitem)
{
	NSL_Data	*data;
	NSL_Item	*item;
	const char	*tagval;

	/* see if weight set already */
	tagval = GetAttrStringVal(sylitem,attrWEIGHT);
	if (tagval && strcmp(tagval,"HEAVY")==0) return(1);

	/* check daughters */
	for (data=sylitem->data;data;data=data->next) {
		if (data->type!=NSL_item_data) return(0);
		item=data->first;
		/* check RHYME */
		if (item->label==tagRHYME)
			if (heavyRHYME(item)) return(1);
	}
	return(0);
}

/* is a consonant a glide */
static int isGlide(NSL_Item *item)
{
	const char *valCNT;
	const char *valSON;

	valCNT = GetAttrStringVal(item,attrCNT);
	valSON = GetAttrStringVal(item,attrSON);
	if (valCNT && (*valCNT=='Y') && valSON && (*valSON=='Y')) return(1);
	return(0);
}

/* is a consonant a liquid */
static int isLiquid(NSL_Item *item)
{
	const char *valCNT;
	const char *valSON;
	const char *valRHO;
	const char *valCNSGRV;
	const char *valCNSCMP;

	valCNT = GetAttrStringVal(item,attrCNT);
	valSON = GetAttrStringVal(item,attrSON);
	valRHO = GetAttrStringVal(item,attrRHO);
	valCNSGRV = GetAttrStringVal(item,attrCNSGRV);
	valCNSCMP = GetAttrStringVal(item,attrCNSCMP);
	if (valCNT && (*valCNT=='Y') &&
	    valSON && (*valSON=='Y') &&
	    valRHO && (*valRHO=='N') &&
	    valCNSGRV && (*valCNSGRV=='N') &&
	    valCNSCMP && (*valCNSCMP=='N')) return(1);
	return(0);
}

/* transfer VOC features throughout onset or coda */
static void xferVOC(NSL_Item *src,NSL_Item *item)
{
	NSL_Data	*data;
	const char	*valVOCGRV;
	const char	*valVOCHEIGHT;
	const char	*valVOCRND;

	/* get values from glide */
	if (src->label==tagVOC) {
		valVOCGRV = GetAttrStringVal(src,attrGRV);
		valVOCHEIGHT = GetAttrStringVal(src,attrHEIGHT);
		valVOCRND = GetAttrStringVal(src,attrRND);
	}
	else {
		valVOCGRV = GetAttrStringVal(src,attrVOCGRV);
		valVOCHEIGHT = GetAttrStringVal(src,attrVOCHEIGHT);
		valVOCRND = GetAttrStringVal(src,attrVOCRND);
	}
	if (!valVOCGRV || !valVOCHEIGHT || !valVOCRND) return;

	/* process item contents */
	for (data=item->data;data;data=data->next) {
		if (data->type==NSL_item_data) {
			item=data->first;
			if (item->label==tagCNS) {
				PutAttrVal(item,attrVOCGRV,strsave(valVOCGRV));
				PutAttrVal(item,attrVOCHEIGHT,strsave(valVOCHEIGHT));
				PutAttrVal(item,attrVOCRND,strsave(valVOCRND));
			}
		}
	}
}

/* parse a single utterance */
NSL_Item * segparse(char *transcription)
{
	char		*data;
	NSL_Item	*seqitem=newItem("SYLSEQ");
	NSL_Item	*sylitem;
	NSL_Item	*onsetitem;
	NSL_Item	*nucitem;
	NSL_Item	*rhymeitem;
	NSL_Item	*codaitem;
	NSL_Item	*acodaitem;
	NSL_Item	*vocitem;
	NSL_Item	*cnsitem;
	int		i;
	struct segment_entry	cseq[32],dummy1[32],dummy2[32];
	struct segment_entry 	ambi1[32],cseq1[32],cseq2[32],ambi2[32];
	struct segment_entry	vseq[32];
	int		eostring;
	int		cstress=0,vstress=0,dstress=0;
	NSL_Item	*glide;
	NSL_Item	*voc1,*voc2;
	NSL_Item	*lastcnsitem;
	char		textattr[32];

	memset(cseq,0,sizeof(cseq));
	memset(vseq,0,sizeof(vseq));
	memset(cseq1,0,sizeof(cseq1));
	memset(cseq2,0,sizeof(cseq2));
	memset(dummy1,0,sizeof(dummy1));
	memset(dummy2,0,sizeof(dummy2));
	memset(ambi1,0,sizeof(ambi1));
	memset(ambi2,0,sizeof(ambi2));

	/* get initial consonant sequence */
	data = getconseq(transcription,cseq1,&eostring,&cstress);

	/* get vowel sequence */
	data = getnucleus(data,vseq,&vstress);

	if (!vseq[0].idx) return(NULL);

	/* get final consonant sequence */
	getconseq(data,cseq,&eostring,&dstress);
	if (eostring)
		memcpy(cseq2,cseq,sizeof(cseq2));
	else
		parseconseq(cseq,cseq2,ambi2,dummy1,dummy2);

	/* got the start of a syllable */
	while (vseq[0].idx) {

		/* start a syllable */
		sylitem = newItem("SYL");

		/* do onset */
		onsetitem=NULL;
		glide=NULL;
		if (ambi1[0].idx || cseq1[0].idx) {
			onsetitem = newItem("ONSET");
			for (i=0;ambi1[i].idx;i++) {
				cnsitem = GetItemFromString(segtab[ambi1[i].idx].elem,odct);
				PutAttrVal(cnsitem,attrAMBI,strsave("Y"));
				if (isGlide(cnsitem)) glide=cnsitem;
				AddItemToEnd(onsetitem,cnsitem);
			}
			for (i=0;cseq1[i].idx;i++) {
				cnsitem = GetItemFromString(segtab[cseq1[i].idx].elem,odct);
				if (isGlide(cnsitem)) glide=cnsitem;
				AddItemToEnd(onsetitem,cnsitem);
			}
			if (glide!=NULL) xferVOC(glide,onsetitem);
			AddItemToEnd(sylitem,onsetitem);
		}

		/* start rhyme */
		rhymeitem = newItem("RHYME");

		/* do nucleus */
		nucitem = newItem("NUC");
		voc1=NULL; voc2=NULL;
		for (i=0;vseq[i].idx;i++) {
			vocitem = GetItemFromString(segtab[vseq[i].idx].elem,odct);
			voc1 = voc2 = vocitem;
			AddItemToEnd(nucitem,vocitem);
			if (*segtab[vseq[i].idx].elem2) {
				vocitem = GetItemFromString(segtab[vseq[i].idx].elem2,odct);
				voc2 = vocitem;
				if (segtab[vseq[i].idx].attr & LONG) PutAttrVal(nucitem,attrLONG,strsave("Y"));
				AddItemToEnd(nucitem,vocitem);
			}
		}
		if (heavyNUC(nucitem)) PutAttrVal(nucitem,attrWEIGHT,strsave("HEAVY"));
		if (onsetitem && !glide) xferVOC(voc1,onsetitem);

		/* do coda */
		lastcnsitem = cnsitem=NULL;
		glide=NULL;
		if ((cseq2[0].idx && !(segtab[cseq2[0].idx].attr & BLOCK)) || ambi2[0].idx) {
			codaitem = newItem("CODA");
			for (i=0;cseq2[i].idx;i++) if (!(segtab[cseq2[i].idx].attr & BLOCK)) {
				cnsitem = GetItemFromString(segtab[cseq2[i].idx].elem,odct);
				if (isLiquid(cnsitem)) glide=cnsitem;
				if (segtab[cseq2[i].idx].attr & ACODA) {
					acodaitem = newItem("ACODA");
					AddItemToEnd(acodaitem,cnsitem);
					AddItemToEnd(codaitem,acodaitem);
				}
				else
					AddItemToEnd(codaitem,cnsitem);
			}
			for (i=0;ambi2[i].idx;i++) {
				cnsitem = GetItemFromString(segtab[ambi2[i].idx].elem,odct);
				if (isLiquid(cnsitem)) glide=cnsitem;
				if (!eostring) PutAttrVal(cnsitem,attrAMBI,strsave("Y"));
#if 0
				if ((cseq2[0].idx || (i>0)) && (!ambi2[i+1].idx) && (segtab[ambi2[i].idx].attr & ACODA)) {
					acodaitem = newItem("ACODA");
					AddItemToEnd(acodaitem,cnsitem);
					AddItemToEnd(codaitem,acodaitem);
				}
				else
#endif
				{
					lastcnsitem= cnsitem;
					AddItemToEnd(codaitem,cnsitem);
				}
			}
			if (glide!=NULL)
				xferVOC(glide,codaitem);
			else
				xferVOC(voc2,codaitem);

			AddItemToEnd(rhymeitem,nucitem);
			AddItemToEnd(rhymeitem,codaitem);
			if (heavyCODA(codaitem)) PutAttrVal(rhymeitem,attrWEIGHT,strsave("HEAVY"));
		}
		else {
			AddItemToEnd(rhymeitem,nucitem);
		}

		if (heavyRHYME(rhymeitem)) PutAttrVal(rhymeitem,attrWEIGHT,strsave("HEAVY"));
		if (lastcnsitem) {
			PutAttrVal(codaitem,attrVOI,strsave(GetAttrStringVal(lastcnsitem,attrVOI)));
			PutAttrVal(nucitem,attrVOI,strsave(GetAttrStringVal(lastcnsitem,attrVOI)));
			PutAttrVal(rhymeitem,attrVOI,strsave(GetAttrStringVal(lastcnsitem,attrVOI)));
			PutAttrVal(nucitem,attrCHECKED,strsave("Y"));
			PutAttrVal(rhymeitem,attrCHECKED,strsave("Y"));
		}
		else {
			/* syllable does not have coda */
			PutAttrVal(nucitem,attrCHECKED,strsave("N"));
			PutAttrVal(nucitem,attrVOI,strsave("Y"));
			PutAttrVal(rhymeitem,attrCHECKED,strsave("N"));
			PutAttrVal(rhymeitem,attrVOI,strsave("Y"));
		}
		AddItemToEnd(sylitem,rhymeitem);
		if (heavySYL(sylitem)) PutAttrVal(sylitem,attrWEIGHT,strsave("HEAVY"));
		if ((cstress&1) || (vstress&1)) {
			if (onsetitem) PutAttrVal(onsetitem,attrSTRENGTH,strsave("STRONG"));
			PutAttrVal(rhymeitem,attrSTRENGTH,strsave("STRONG"));
			PutAttrVal(nucitem,attrSTRENGTH,strsave("STRONG"));
		}
		sprintf(textattr,"%d",(cstress|vstress));
		PutAttrVal(sylitem,attrSTRENGTH,strsave(textattr));
		AddItemToEnd(seqitem,sylitem);

		/* get next syllable */
		data = getconseq(data,cseq,&eostring,&cstress);
		ambi1[0].idx=cseq1[0].idx=0;
		parseconseq(cseq,dummy1,dummy2,ambi1,cseq1);

		data = getnucleus(data,vseq,&vstress);
		getconseq(data,cseq,&eostring,&dstress);
		ambi2[0].idx=cseq2[0].idx=0;
		if (eostring)
			memcpy(cseq2,cseq,sizeof(cseq2));
		else
			parseconseq(cseq,cseq2,ambi2,dummy1,dummy2);
	}

	return(seqitem);
}

/* add reverse foot position attributes to syl seq under foot */
void addRFPOS(NSL_Item *foot,int footcnt)
{
	NSL_Data	*data;
	char		buf[32];

	for (data=foot->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			sprintf(buf,"%d",footcnt--);
			PutAttrVal(data->first,attrRFPOS,strsave(buf));
		}
	}
}

/* IP parse - parse a syllable sequence into one IP */
NSL_Item * ipparse(NSL_Item *sylseq)
{
	NSL_Data	*data;
	NSL_Item	*ipitem=newItem("IP");
	NSL_Item	*agitem=newItem("AG");
	NSL_Item	*ftitem=newItem("FOOT");
	NSL_Item	*sylitem;
	const char	*p;
	int		flags;
	int		footcnt=0;
	char	buf[32];

	for (data=sylseq->data;data!=NULL;data=data->next) {
		sylitem = (NSL_Item *)data->first;
		if (strcmp(sylitem->label,"SYL")!=0)
			error("illegal syllable sequence found in ipparse()\n");
		/* get stress value */
		p = GetAttrStringVal(sylitem,attrSTRENGTH);
		if (!p || !isdigit(*p))
			error("illegal STRENGTH attribute on SYL in ipparse()\n");
		flags = atoi(p);
		if (flags & FTBOUND) {
			if (ftitem->type!=NSL_empty) {
				addRFPOS(ftitem,footcnt);
				AddItemToEnd(agitem,ftitem);
				ftitem=newItem("FOOT");
				footcnt=0;
			}
		}
		if (flags & AGBOUND) {
			if (agitem->type!=NSL_empty) {
				AddItemToEnd(ipitem,agitem);
				agitem=newItem("AG");
			}
		}
		if (flags & IPBOUND) {
			if (ipitem->type!=NSL_empty)
				break;
		}
		if (flags & STRESS1)
			PutAttrVal(sylitem,attrSTRENGTH,strsave("STRONG"));
		else
			PutAttrVal(sylitem,attrSTRENGTH,strsave("WEAK"));
		sprintf(buf,"%d",++footcnt);
		PutAttrVal(sylitem,attrFPOS,strsave(buf));
		AddItemToEnd(ftitem,sylitem);
	}
	if (ftitem->type!=NSL_empty) {
		addRFPOS(ftitem,footcnt);
		AddItemToEnd(agitem,ftitem);
	}
	if (agitem->type!=NSL_empty) AddItemToEnd(ipitem,agitem);
	return(ipitem);
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	char		filename[256];
	FILE		*ip;
	int		pos;
	int		readstdin=0;

	NSL_File	inf,outf;
	NSL_Item 	*sylseq;
	NSL_Item	*ipitem;
	NSL_Item	*utitem;

	/* Initialise the NSL API */
	NSLInit(0);

	/* encode syllable onsets and offsets */
	loadsylonoff();

	/* option decoding */
	while ((c=getopt(argc,argv,"Ii"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: parse phonetic transcription Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'i':
		readstdin=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-i|photran.txt) (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((inf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(inf);
	SFclose(inf);

	/* open input file */
	if (readstdin) {
		strcpy(filename,"stdin");
		ip = stdin;
	}
	else if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((ip=fopen(filename,"r"))==NULL)
			error("unable to open '%s'",filename);
	}
	else
		error("no input file specified");

	/* Open the output NSGML stream using the doctype declaration */
	optind++;
	if (optind < argc)
		outf = SFopen(argv[optind], odct, NSL_write_pretty);
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* get references to elements and attributes */
	tagIP = ElementUniqueName(odct,"IP",0);
	tagAG = ElementUniqueName(odct,"AG",0);
	tagFOOT = ElementUniqueName(odct,"FOOT",0);
	tagSYL = ElementUniqueName(odct,"SYL",0);
	tagONSET = ElementUniqueName(odct,"ONSET",0);
	tagRHYME = ElementUniqueName(odct,"RHYME",0);
	tagNUC = ElementUniqueName(odct,"NUC",0);
	tagCODA = ElementUniqueName(odct,"CODA",0);
	tagVOC = ElementUniqueName(odct,"VOC",0);
	tagCNS = ElementUniqueName(odct,"CNS",0);
	attrAMBI = AttrUniqueName(odct,"AMBI",0);
	attrWEIGHT = AttrUniqueName(odct,"WEIGHT",0);
	attrSTRENGTH = AttrUniqueName(odct,"STRENGTH",0);
	attrCHECKED = AttrUniqueName(odct,"CHECKED",0);
	attrLONG = AttrUniqueName(odct,"LONG",0);
	attrCNT = AttrUniqueName(odct,"CNT",0);
	attrSON = AttrUniqueName(odct,"SON",0);
	attrRHO = AttrUniqueName(odct,"RHO",0);
	attrCNSGRV = AttrUniqueName(odct,"CNSGRV",0);
	attrCNSCMP = AttrUniqueName(odct,"CNSCMP",0);
	attrVOCGRV = AttrUniqueName(odct,"VOCGRV",0);
	attrVOCHEIGHT = AttrUniqueName(odct,"VOCHEIGHT",0);
	attrVOCRND = AttrUniqueName(odct,"VOCRND",0);
	attrVOI = AttrUniqueName(odct,"VOI",0);
	attrGRV = AttrUniqueName(odct,"GRV",0);
	attrHEIGHT = AttrUniqueName(odct,"HEIGHT",0);
	attrRND = AttrUniqueName(odct,"RND",0);
	attrFPOS = AttrUniqueName(odct,"FPOS",0);
	attrRFPOS = AttrUniqueName(odct,"RFPOS",0);

	/* get enclosure */
	PrintStartTag(outf,"PROSYNTH");

	/* Loop round reading IP from input text */
	while (!feof(ip)) {
		/* get an IP */
		pos=0;
		while ((c=getc(ip))!=EOF) if (!isspace(c)) {
			if ((pos>0) && (c=='/')) break;
			if ((pos==0) && (c!='/')) buf[pos++]='/';
			buf[pos++]=c;
			if (pos >= MAXBUFLEN)
				error("input line too big");
		}
		buf[pos]='\0';
		/* parse it into syllables */
		if (pos > 0) {
			utitem = newItem("UTT");
			sylseq = segparse(buf);
			ipitem = ipparse(sylseq);
			AddItemToEnd(utitem,ipitem);
			PrintItem(outf,utitem);
			FreeItem(utitem);
		}
	}
	PrintEndTag(outf,"PROSYNTH");

	/* that's all folks */
	SFclose(outf);
	if (ip!=stdin) fclose(ip);
	return 0;
}

