/* prxlex -- lexical analyser and symbol table manipulation for ProXML */

/* Mark Huckvale - University College London */

/* version 0.1 - February 1999 */

#undef IAG

/* external definitions */
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "prx.h"
#include "process.h"
#include "builtin.h"

/* symbol table */
#define MAXSYMBOL 4000
Symbol 	*symtab[MAXSYMBOL];
int	sympos;

/* reserved words */
static struct {
	char	*name;
	short	tok;
} reserved[] = {			/* KEEP SORTED */
{	"assert",	ASSERT_TOK },
{	"break",	BREAK_TOK },
{	"const",	CONST_TOK },
{	"else",		ELSE_TOK },
{	"false",	FALSE_TOK },
{	"for",		FOR_TOK },
{	"foreach",	FOREACH_TOK },
{	"if",		IF_TOK },
{	"include",	INCLUDE_TOK },
{	"nil",		NIL_TOK },
{	"node",		NODE_TOK },
{	"resize",	RESIZE_TOK },
{	"return",	RETURN_TOK },
{	"sizeof",	SIZEOF_TOK },
{	"true",		TRUE_TOK },
{	"var",		VAR_TOK },
{	"while",	WHILE_TOK },
{	0,		0 },		/* KEEP SORTED */

};

/* built-in functions */
static struct {
	char	*name;
	void	(*func)(void);
	short	etype[2];
	int	nparm;
} functions[] = {
{	"char",		BI_tochar,	{ VALUE, 0 }, 1 },
{	"element",	BI_element,	{ NODE, 0 }, 1 },
{	"exit",		BI_exit,	{ VALUE, VALUE }, 2 },
{	"float",	BI_tofloat,	{ VALUE, 0 }, 1 },
{	"format",	BI_format,	{ VALUE, VALUE }, 2 },
{	"integer",	BI_tointeger,	{ VALUE, 0 }, 1 },
{	"iselement",	BI_iselement,	{ NODE, 0 }, 1 },
{	"istext",	BI_istext,	{ NODE, 0 }, 1 },
{	"log10",	BI_log10,	{ VALUE, 0 }, 1 },
{	"newattribute",	BI_newattribute, { NODE, VALUE }, 2 },
{	"numchild",	BI_numchild,	{ NODE, 0 }, 1 },
{	"output",	BI_output,	{ VALUE, 0 }, 1 },
{	"pow10",	BI_pow10,	{ VALUE, 0 }, 1 },
{	"random",	BI_random,	{ 0, 0 }, 0 },
{	"streamclose",	BI_streamclose,	{ VALUE, 0 }, 1 },
{	"streamget",	BI_streamget,	{ VALUE, 0 }, 1 },
{	"streamgetline",	BI_streamgetline,	{ VALUE, 0 }, 1 },
{	"streamopen",	BI_streamopen,	{ VALUE, VALUE }, 2 },
{	"streamput",	BI_streamput,	{ VALUE, VALUE }, 2 },
{	"string",	BI_tostring,	{ VALUE, 0 }, 1 },
{	"text",		BI_text,	{ NODE, 0 }, 1 },
{	0,		0,  { 0, 0 }, 0 },
};

/* built-in node functions */
static struct {
	char	*name;
	void	(*func)(void);
	short	etype[3];
	int	nparm;
} nfunctions[] = {
{	"ancestor",	BI_ancestor,	{ NODE, VALUE, 0 }, 2 },
{	"appendchild",	BI_appendchild,	{ NODE, VALUE, 0 }, 2 },
{	"appenditem",	BI_appenditem,	{ NODE, NODE, 0 }, 2 },
{	"deleteitem",	BI_deleteitem,	{ NODE, 0, 0 }, 1 },
{	"descendant",	BI_descendant,	{ NODE, VALUE, 0 }, 2 },
{	"following",	BI_following,	{ NODE, VALUE, 0 }, 2 },
{	"newitem",	BI_newitem,	{ VALUE, 0, 0 }, 1 },
{	"parse",	BI_parse,	{ VALUE, VALUE, VALUE }, 3 },
{	"preceding",	BI_preceding,	{ NODE, VALUE, 0 }, 2 },
{	"sortchildren",	BI_sortchildren,	{ NODE, VALUE, 0 }, 2 },
{	0,		0,  { 0, 0, 0 }, 0 },
};

/* save a string in dynamic memory */
char *strsave(const char *str)
{
	char *p;
	if (!str) return(NULL);
	if ((p=malloc(strlen(str)+1))==NULL)
		prxerror("out of memory",NULL);
	strcpy(p,str);
	return(p);
}

/* install a symbol into symbol table */
int	installsymbol(char *name,int type,int status,char * valstr)
{
	Symbol	*sym;

	if (sympos >= MAXSYMBOL)
		prxerror("too many symbols","");

	if ((sym=(Symbol *)malloc(sizeof(Symbol)))==NULL)
		prxerror("out of memory",NULL);
	memset((char *)sym,0,sizeof(Symbol));

	sym->name = strsave(name);
	sym->type = type;
	sym->status = status;
	if (valstr) {
		switch (type) {
		case CHAR:	sym->val.u.c = valstr[0]; break;
		case INTEGER:	sym->val.u.i = atoi(valstr); break;
		case FLOAT:	sym->val.u.d = atof(valstr); break;
		case STRING:	sym->val.u.s = strsave(valstr); break;
		}
		sym->val.type = type;
	}
	symtab[sympos++] = sym;
	return(sympos-1);
}

/* free symbol */
void freesymbol(Symbol *sym)
{
	if (sym->name) free(sym->name);
	if (sym->type==STRING) free(sym->val.u.s);
	free(sym);
}

void clearlocals(int epos)
{
	while (epos < sympos) {
		if ((symtab[epos]->status==LOCAL) &&
		    ((symtab[epos]->type==VARIABLE)||(symtab[epos]->type==NODE)||
		     (symtab[epos]->type==ARRAY)||(symtab[epos]->type==FUNCTION)||
		     (symtab[epos]->type==NODE_FUNCTION))) {
			if (symtab[epos]->name) free(symtab[epos]->name);
			symtab[epos]->name=NULL;
		}	
		epos++;
	}
}

/* initialise symbol table */
void initsymbol()
{
	int	i;
	int	s;
	
	/* clear any existing contents of table */
	for (i=0;i<sympos;i++) free(symtab[i]);
	sympos=0;

	/* install reserved words */	
	for (i=0;reserved[i].name;i++)
		installsymbol(reserved[i].name,reserved[i].tok,RESERVED,NULL);

	/* install built-in functions */	
	for (i=0;functions[i].name;i++) {
		s = installsymbol(functions[i].name,FUNCTION,RESERVED,NULL);
		symtab[s]->val.type = FUNCTION;
		symtab[s]->val.u.v = (void *)(functions[i].func);
		symtab[s]->etype = functions[i].etype[0];
		symtab[s]->nparm = functions[i].nparm;
	}

	/* install built-in node functions */
	for (i=0;nfunctions[i].name;i++) {
		s = installsymbol(nfunctions[i].name,NODE_FUNCTION,RESERVED,NULL);
		symtab[s]->val.type = NODE_FUNCTION;
		symtab[s]->val.u.v = (void *)(nfunctions[i].func);
		symtab[s]->etype = nfunctions[i].etype[0];
		symtab[s]->nparm = nfunctions[i].nparm;
	}
}

/* symbol table lookup */
int	lookupsymbol(char *name)
{
	int	i;
	Symbol	*sym;

	for (i=sympos-1;i>=0;i--) {
		sym = symtab[i];
		if (sym->name && (strcmp(name,sym->name)==0)) {
			return(i);
		}
	}

	return(-1);
}

/* dump symbol name and value (for debugging) */
void dumpsymbol(int s)
{
	if ((0 <= s) && (s < sympos)) {
		printf("Name='%s' Value=",(symtab[s]->name)?symtab[s]->name:"<nil>");
		switch (symtab[s]->type) {
		case CHAR:	printf("%c",symtab[s]->val.u.c); break;
		case INTEGER:	printf("%d",symtab[s]->val.u.i); break;
		case FLOAT:	printf("%g",symtab[s]->val.u.d); break;
		case STRING:	printf("%s",symtab[s]->val.u.s); break;
		default:	printf("%X",symtab[s]->val.u.i); break;
		}
	}
}

/* input file stack */
#define FSTACKSIZE	10
struct {
	char	fname[80];
	FILE	*fp;
	int	lno;
	int	col;
} fstack[FSTACKSIZE];
int		fsp;
extern FILE	*ip;
extern char	filename[];
extern char	basename[];
extern int	domap;
extern FILE	*mfp;
int		lno;
int		errcomments;
static int	lastlno=1;
static int 	last2c;
static int 	lastc;
static int 	lastcol;
static int	col;
static int	oldcodepos;

void resetlex()
{
	fsp = 0;
	lastlno=1;
	lastcol=0;
	last2c = lastc = 0;
	errcomments=0;
	sympos=0;
}

char *findfile(char *name)
{
	static char	filename[256];

	/* try base directory */
	if (basename[0]) {
		sprintf(filename,"%s%s",basename,name);
#ifdef WIN32
		if (_access(filename,0)==0)
#else
		if (access(filename,0)==0)
#endif
			return(filename);
	}
		
	/* try file straight */
	strncpy(filename,name,256);
#ifdef WIN32
	if (_access(filename,0)==0)
#else
	if (access(filename,0)==0)
#endif
		return(filename);

	/* try /usr/local directory */
	strcpy(filename,"/usr/local/include/");
	strcat(filename,name);
#ifdef WIN32
	if (_access(filename,0)==0)
#else
	if (access(filename,0)==0)
#endif
		return(filename);
	else
		return(name);
}

void pushfcontext(char *fname)
{
	strcpy(fstack[fsp].fname,filename);
	fstack[fsp].fp = ip;
	fstack[fsp].lno = lastlno;
	fstack[fsp].col = lastcol;
	fsp++;
	strcpy(filename,findfile(fname));
	if ((ip=fopen(filename,"r"))==NULL)
		prxerror("unable to find '%s'",fname);
	lastlno=1;
	lastcol=1;
}
int popfcontext()
{
	if (fsp==0)
		return(0);
	else {
		fclose(ip);
		fsp--;
		strcpy(filename,fstack[fsp].fname);
		ip = fstack[fsp].fp;
		lastlno = fstack[fsp].lno;
		lastcol = fstack[fsp].col;
	}
	return(1);
}

/* get a source character */
int charget()
{
	int	c;

	lno = lastlno;
	col = lastcol;

	if (lastc) {
		c = lastc;
		lastc = last2c;
		last2c = 0;
		return(c);
	}
	else do {
		if ((c = getc(ip))!=EOF) {
			if (c=='\n') {
				if (domap) {
					fprintf(mfp,"PC: %04d File: %s Line: %d\n",oldcodepos,filename,lastlno);
					oldcodepos = codepos;
				}
				lastlno++;
				lastcol=1;
			}
			else if (c=='\t') do {
				lastcol++;
			} while ((lastcol%8)!=1);
			else
				lastcol++;
			return(c);
		}
	} while (popfcontext());
	return(EOF);
}

/* unget a source character */
void charunget(int c)
{
	last2c = lastc;
	lastc = c;
}

/* yylex token positioning */
YYLTYPE yylloc;

/* yylex definition */
#ifdef IAG
int myylex();
int yylex()
{
	int c=myylex();
	printf("yylex()=>%d ",c);
	dumpsymbol(yylval);
	printf("\n");
	return(c);
}
int myylex()
#else
int yylex()
#endif
{
	char	sbuf[100], *p;
	int	c,c2;
	int	s;
	int	numtype;

	/* store token position */
	yylloc.first_line = lno;
	yylloc.first_column = col;

	/* get a character */
	if ((c = charget())==EOF)
		return(0);

	/* get following character */
	c2 = charget();

	/* skip whitespace and comments */
	while ((c == ' ') || (c == '\t') || (c=='\n') || (c=='\r') ||
		((c=='/') && (c2=='*')) || ((c=='/') && (c2=='/'))) {
		/* strip comments here */
		if ((c=='/') && (c2=='*')) {
			/* skip comment until * + / */
			if ((c=charget())==EOF)
				return(0);
			c2=charget();
			while (!((c=='*') && (c2=='/'))) {
				if ((c==EOF) || (c2==EOF))
					return(0);
				if ((c=='/') && (c2=='*'))
					warning(&yylloc,"nested comments?","");
				c = c2;
				c2 = charget();
			}
			if ((c=charget())==EOF)
				return(0);
			c2=charget();
		}
		if ((c=='/') && (c2=='/')) {
			/* skip comment until \n */
			if ((c=charget())==EOF)
				return(0);
			while (!(c=='\n')) {
				if (c==EOF)
					return(0);
				c=charget();
			}
			if ((c=charget())==EOF)
				return(0);
			c2=charget();
		}
		while ((c==' ') || (c=='\t') || (c=='\n') || (c=='\r')) {
			if ((c = c2)==EOF)
				return(0);

			/* store token position */
			yylloc.first_line = lno;
			yylloc.first_column = col;

			c2 = charget();
		}
		if (c==EOF) {
			fprintf(stderr,"?PRX Compiler error, unexpected EOF\n");
			return(0);
		}

	}

	/* process numbers */
	if (isdigit(c)) {
		numtype=INTEGER;
		p = sbuf;
		*p++ = c;
		if ((c = c2)==EOF)
			return(0);
		c2 = charget();
		while (isdigit(c)) {
			if (p >= (sbuf + sizeof(sbuf) - 1)) {
				*p ='\0';
				warning(&yylloc,"number too long",sbuf);
			}
			*p++ = c;
			if ((c = c2)==EOF)
				return(0);
			c2 = charget();
		}
		if (c=='.') {
			/* add in decimals */
			numtype=FLOAT;
			*p++ = c;
			if ((c = c2)==EOF)
				return(0);
			c2 = charget();
			while (isdigit(c)) {
				if (p >= (sbuf + sizeof(sbuf) - 1)) {
					*p ='\0';
					warning(&yylloc,"number too long",sbuf);
				}
				*p++ = c;
				if ((c = c2)==EOF)
					return(0);
				c2 = charget();
			}
		}
		if ((c=='e') || (c=='E')) {
			/* add in exponent */
			numtype=FLOAT;
			*p++ = c;
			if ((c = c2)==EOF)
				return(0);
			c2 = charget();
			if (c=='-') {
				*p++ = c;
				if ((c = c2)==EOF)
					return(0);
				c2 = charget();
			}
			while (isdigit(c)) {
				if (p >= (sbuf + sizeof(sbuf) - 1)) {
					*p ='\0';
					warning(&yylloc,"number too long",sbuf);
				}
				*p++ = c;
				if ((c = c2)==EOF)
					return(0);
				c2 = charget();
			}
		}
		*p = '\0';
		s = installsymbol(sbuf,(short)numtype,CONSTANT,sbuf);
		yylval = s;
		charunget(c2);
		charunget(c);
		return(symtab[s]->type);
	}

	/* process identifiers */
	if (isalpha(c) || (c=='_')) {
		p = sbuf;
		*p++ = c;
		if ((c = c2)==EOF)
			return(0);
		c2 = charget();
		while (isalnum(c) || (c=='_')) {
			if (p >= (sbuf + sizeof(sbuf) - 1)) {
				*p ='\0';
				warning(&yylloc,"name too long",sbuf);
			}
			*p++ = c;
			if ((c = c2)==EOF)
				return(0);
			c2 = charget();
		}
		*p='\0';
		if ((s=lookupsymbol(sbuf)) < 0) {
			/* add new symbol as generic identifier */
			s = installsymbol(sbuf,IDENTIFIER,LOCAL,NULL);
		}
		yylval = s;
		charunget(c2);
		charunget(c);
#ifdef IAG
fprintf(stderr,"[Symbol='%s',line=%ld,col=%ld]\n",
	symtab[s]->name,yylloc.first_line,yylloc.first_column);
#endif
		return(symtab[s]->type);
	}

	/* quoted strings */
	if (c=='"') {
		if ((c = c2)==EOF)
			return(0);
		c2 = charget();
		for (p=sbuf;c != '"'; p++) {
			if ((c=='\n') || (c== EOF)) {
				warning(&yylloc,"missing quotes","");
				break;
			}
			if (p>=sbuf + sizeof(sbuf) -1) {
				*p='\0';
				warning(&yylloc,"string too long",sbuf);
				break;
			}
			if (c=='\\') {
				c = backslash(c2);
				c2 = charget();
			}
			*p = c;
			if ((c = c2)==EOF)
				return(0);
			c2 = charget();
		}
		*p = '\0';
		s = installsymbol("",STRING,CONSTANT,sbuf);
		yylval = s;
		charunget(c2);
		return(STRING);
	}

	/* quoted character */
	if (c=='\'') {
		if ((c = c2)==EOF)
			return(0);
		c2 = charget();
		if (c=='\\') {
			c = backslash(c2);
			c2 = charget();
		}
		if (c2!='\'')
			warning(&yylloc,"missing quote","");
		sbuf[0]=c;
		sbuf[1]='\0';
		s = installsymbol("",CHAR,CONSTANT,sbuf);
		yylval = s;
		return(CHAR);
	}

	yylval = -1;

	/* graphs and digraphs */
	switch (c) {
	case '>':	return follow(c2,'=',GE_TOK,GT_TOK);
	case '<':	return follow(c2,'=',LE_TOK,LT_TOK);
	case '=':	return follow(c2,'=',EQ_TOK,ASSIGN_TOK);
	case '!':	return follow(c2,'=',NE_TOK,NOT_TOK);
	case '&':	return follow(c2,'&',AND_TOK,BIT_AND_TOK);
	case '|':	return follow(c2,'|',OR_TOK,BIT_OR_TOK);
	case '+':	if (c2=='+') return(CONCAT_TOK);
			else return follow(c2,'=',PLUS_ASSIGN_TOK,PLUS_TOK);	
	case '-':	charunget(c2);
			return(MINUS_TOK);
	case '/':	charunget(c2);
			return(DIVIDE_TOK);
	case '*':	return follow(c2,'=',TIMES_ASSIGN_TOK,TIMES_TOK);
	case '.':	if (c2!='.') {
				charunget(c2);
				return(c);
			}
			else {
				c = c2;
				c2 = charget();
				return follow(c2,'.',DESCENDANT_TOK,PARENT_TOK);
			}
	default:
		charunget(c2);
		return(c);
	}
}

/* interpret backslash char */
int backslash(int c)
{
	char 	*strchr(),*p;
	static char transtab[]="b\bf\fn\nr\rt\t\"\"''";
	if ((p=strchr(transtab,c)))
		return(*(p+1));
	return(c);
}

/* look ahead for digraphs */
int follow(int c2,int expect,int ifyes,int ifno)
{
	if (c2==expect)
		return(ifyes);
	charunget(c2);
	return(ifno);
}

/* assign a normal variable a stack position */
int assignlocal(int sym,int vno)
{
	symtab[sym]->type = VARIABLE;
	symtab[sym]->status = LOCAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a node variable a stack position */
int assignnode(int sym,int vno)
{
	symtab[sym]->type = NODE;
	symtab[sym]->status = LOCAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignlocalfunc(int sym,int vno)
{
	symtab[sym]->type = FUNCTION;
	symtab[sym]->status = LOCAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a node variable a stack position */
int assignnodefunc(int sym,int vno)
{
	symtab[sym]->type = NODE_FUNCTION;
	symtab[sym]->status = LOCAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignglobal(int sym,int vno)
{
	symtab[sym]->type = VARIABLE;
	symtab[sym]->status = GLOBAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a node variable a stack position */
int assignglobalnode(int sym,int vno)
{
	symtab[sym]->type = NODE;
	symtab[sym]->status = GLOBAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignglobalconstant(int sym,int vno)
{
	symtab[sym]->type = CONSTSYM;
	symtab[sym]->status = GLOBAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignlocalvector(int sym,int vno)
{
	symtab[sym]->type = ARRAY;
	symtab[sym]->status = LOCAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignlocalnodevector(int sym,int vno)
{
	symtab[sym]->type = NARRAY;
	symtab[sym]->status = LOCAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignglobalvector(int sym,int vno)
{
	symtab[sym]->type = ARRAY;
	symtab[sym]->status = GLOBAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* assign a normal variable a stack position */
int assignglobalnodevector(int sym,int vno)
{
	symtab[sym]->type = NARRAY;
	symtab[sym]->status = GLOBAL;
	symtab[sym]->stackpos = vno;
	return(0);
}

/* correct parameter indices */
int correctparams(int nparam)
{
	int	i;

	for (i=0;i<nparam;i++)
		symtab[sympos-i-1]->stackpos = -(2+i);
	return(0);
}

#ifdef EMO
FILE	*ip;
Symbol	*yylval;
main()
{
	Symbol 	*sym;
	int	ch;

	ip=stdin;

	initsymbol();

	while (ch=yylex()) {
		if (yylval) {
			sym = yylval;
			printf("name='%s' type=%d status=%d\n",
				sym->name,sym->type,sym->status);
		}
		else if (ch < 128)
			printf("name='%c'\n",ch);
		else
			printf("name=%d\n",ch);
	}
}

#endif

