/* segparse -- parse a segment level transcription */

/* Mark Huckvale - University College London */

/* version 1.0 - December 1997 */

#define PROGNAME "segparse"
#define PROGVERS "1.0"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "lt-memory.h"
#include "nsl.h"
#include "segment.h"
#include "conoff.h"

/* input file */
char		filename[256];
const char	*tagSEG;
const char	*attrID;

/* output file */

/* utterance item */
NSL_Item	*utt;
char		uttID[16];
char		uttHREF[256];
int		uttno=0;

/* syllable item */
NSL_Item	*syl;
char		sylID[16];
char		sylHREF[256];
int		sylno=0;

/* onset item */
NSL_Item	*ons;
char		onsID[16];
char		onsHREF[256];
int		onsno=0;

/* rhyme item */
NSL_Item	*rhyme;
char		rhymeID[16];
char		rhymeHREF[256];
int		rhymeno=0;

/* nucleus item */
NSL_Item	*nuc;
char		nucID[16];
char		nucHREF[256];
int		nucno=0;

/* coda item */
NSL_Item	*coda;
char		codaID[16];
char		codaHREF[256];
int		codano=0;

/* segment item */
NSL_Item	*seg;
char		segID[16];
char		segHREF[256];
char		segAMBI[4];
int		segno=0;

/* record of a segment */
struct segment_entry {
	int	idx;
	const char	*id;
};

/* look up a segment name */
int findsegment(char *sym)
{
	int	lo,hi,k,c;

	lo = 0;
	hi = (sizeof(segtab)/sizeof(struct segment_rec));
	do {
		k = (lo+hi)/2;
		c = strcmp(sym,segtab[k].sym);
		if (c < 0)
			hi = k-1;
		else if (c > 0)
			lo = k+1;
		else
			return(k);
	} while (lo <= hi);
	LTNSL_ERROR1(0,"unknown symbol: '%s'",sym);
	return(0);
}

NSL_Data *getconseq(NSL_Data *data,struct segment_entry *seq)
{
	NSL_Item	*sitem;
	int		idx;

	while (data!=NULL) {
		/* should be a SEG tag */
		sitem = (NSL_Item *)(data->first);
		if (sitem->label != tagSEG)
			LTNSL_ERROR(0,"Utterance contains other tag than <SEG>");
		idx = findsegment(sitem->data->first);
		if (segtab[idx].attr & NUC) {
			seq->idx = 0;
			seq->id = NULL;
			return(data);
		}
		seq->idx = idx;
		seq->id = GetAttrStringVal(sitem,attrID);
		seq++;
		data = data->next;
	}
	seq->idx = 0;
	seq->id = NULL;
	return(data);
}

NSL_Data	*getnucleus(NSL_Data *data,struct segment_entry *seq)
{
	NSL_Item	*sitem;
	int	idx;

	if (data==NULL) {
		seq->idx = 0;
		seq->id = NULL;
		return(NULL);
	}
	else {
		/* should be a SEG tag */
		sitem = (NSL_Item *)(data->first);
		if (sitem->label != tagSEG)
			LTNSL_ERROR(0,"Utterance contains other tag than <SEG>");
		idx = findsegment(sitem->data->first);

		if (segtab[idx].attr & NUC) {
			seq->idx = idx;
			seq->id = GetAttrStringVal(sitem,attrID);
			seq++;
			seq->idx = 0;
			seq->id = NULL;
			return(data->next);
		}
		else {
			seq->idx = 0;
			seq->id = NULL;
			return(data);
		}
	}
}

/* check segment sequence overlap */
int overlapf(struct segment_entry *sep,char *cp)
{
	int	cnt=0;
	while ((sep->idx) && *cp && (sep->idx==*cp)) { cnt++; sep++; cp++; }
	return(cnt);
}
int overlapr(struct segment_entry *sep,char *cp)
{
	int	cnt=0;
	struct segment_entry *sep2;
	char	*cp2;
	
	for (sep2=sep;sep2[0].idx && sep2[1].idx;sep2++) /* skip */;
	for (cp2=cp;*cp2 && *(cp2+1);cp2++) /* skip */;
	while ((sep2>=sep) && (cp2>=cp) && (sep2->idx==*cp2)) { cnt++; sep2--; cp2--; }
	return(cnt);
}

/* parse consonant sequence */
void parseconseq(struct segment_entry *cseq,
	    struct segment_entry *cseq1,
	    struct segment_entry *ambi1,
	    struct segment_entry *ambi2,
	    struct segment_entry *cseq2)
{
	int	i,len;
	int	onmax,offmax;
	int	nambi;

	for (len=0;cseq[len].idx!=0;len++) /* skip */;

	/* find longest matching offset */
	offmax=0;
	for (i=0;offset[i][0];i++) {
		if (overlapf(cseq,offset[i])>offmax) {
			offmax = overlapf(cseq,offset[i]);
		}
	}
	/* find longest matching onset */
	onmax=0;
	for (i=0;onset[i][0];i++) {
		if (overlapr(cseq,onset[i])>onmax) {
			onmax = overlapr(cseq,onset[i]);
		}
	}

	/* find overlap between onset and offset */
	nambi = onmax + offmax - len;

	/* copy relevant parts */
	for (i=0;i<len-onmax;i++) {
		cseq1[0] = cseq[i];
		cseq1++;
	}
	for (;i<len-onmax+nambi;i++) {
		ambi1[0] = ambi2[0] = cseq[i];
		ambi1++;
		ambi2++;
	}
	for (;i<len;i++) {
		cseq2[0] = cseq[i];
		cseq2++;
	}

	/* terminate strings */
	cseq1[0].idx=0; cseq1[0].id=NULL; 
	ambi1[0].idx=0; ambi1[0].id=NULL;
	ambi2[0].idx=0; ambi2[0].id=NULL;
	cseq2[0].idx=0; cseq2[0].id=NULL;

}

/* print sequence */
void printseq(NSL_File outf,struct segment_entry *seq,int ambi)
{
	while (seq->id) {
		sprintf(segID,"SEG%d",++segno);
		sprintf(segHREF,"%s#ID(%s)",filename,seq->id);
		sprintf(segAMBI,"%s",(ambi)?"yes":"no");
		PrintItem(outf,seg);
		PrintText(outf,segtab[seq->idx].sym);
		PrintEndTag(outf,seg->label);
		seq++;
	}
}

/* find index of last entry */
int slast(struct segment_entry *seq)
{
	int	i;
	for (i=0;(seq[i].idx!=0) && (seq[i+1].idx!=0);i++) /* skip */;
	return(i);
}

/* parse a single utterance */
void segparse(NSL_File outf,NSL_Item *item)
{
	NSL_Data	*data;
	struct segment_entry	cseq[32],dummy1[32],dummy2[32];
	struct segment_entry 	ambi1[32],cseq1[32],cseq2[32],ambi2[32];
	struct segment_entry	vseq[32];
	const char	*first,*last;
	
	/* get initial consonant sequence */
	ambi1[0].idx = 0;
	ambi1[0].id = NULL;
	data = getconseq(item->data,cseq1);

	/* get vowel sequence */
	data = getnucleus(data,vseq);

	if (!vseq[0].idx) return;

	/* get final consonant sequence */
	getconseq(data,cseq);
	parseconseq(cseq,cseq2,ambi2,dummy1,dummy2);
	
	/* got the start of a syllable */
	while (vseq[0].idx) {
		/* start a syllable */
		sprintf(sylID,"SYL%d",++sylno);
		first = ambi1[0].id;
		if (!first) first = cseq1[0].id;
		if (!first) first = vseq[0].id;
		
		last = ambi2[slast(ambi2)].id;
		if (!last) last = cseq2[slast(cseq2)].id;
		if (!last) last = vseq[slast(vseq)].id;
		sprintf(sylHREF,"%s#ID(%s)..ID(%s)",filename,first,last);
		PrintItem(outf,syl);
		
		/* do onset */
		if (ambi1[0].idx || cseq1[0].idx) {
			sprintf(onsID,"ONSET%d",++onsno);
			first = ambi1[0].id;
			if (!first) first = cseq1[0].id;
		
			last = cseq1[slast(cseq1)].id;
			if (!last) last = ambi1[slast(ambi1)].id;
			sprintf(onsHREF,"%s#ID(%s)..ID(%s)",filename,first,last);
			PrintItem(outf,ons);
			printseq(outf,ambi1,1);
			printseq(outf,cseq1,0);
			PrintEndTag(outf,ons->label);
		}

		/* start rhyme */
		sprintf(rhymeID,"RHYME%d",++rhymeno);
		first = vseq[0].id;
		last = ambi2[slast(ambi2)].id;
		if (!last) last = cseq2[slast(cseq2)].id;
		if (!last) last = vseq[slast(vseq)].id;
		sprintf(rhymeHREF,"%s#ID(%s)..ID(%s)",filename,first,last);
		PrintItem(outf,rhyme);

		/* do nucleus */
		sprintf(nucID,"NUC%d",++nucno);
		first = vseq[0].id;
		last = vseq[slast(vseq)].id;
		sprintf(nucHREF,"%s#ID(%s)..ID(%s)",filename,first,last);
		PrintItem(outf,nuc);
		printseq(outf,vseq,0);
		PrintEndTag(outf,nuc->label);

		/* do coda */
		if (cseq2[0].idx || ambi2[0].idx) {
			sprintf(codaID,"CODA%d",++codano);
			first = cseq2[0].id;
			if (!first) first = ambi2[0].id;
			last = ambi2[slast(ambi2)].id;
			if (!last) last = cseq2[slast(cseq2)].id;
			sprintf(codaHREF,"%s#ID(%s)..ID(%s)",filename,first,last);
			PrintItem(outf,coda);
			printseq(outf,cseq2,0);
			printseq(outf,ambi2,1);
			PrintEndTag(outf,coda->label);
		}

		PrintEndTag(outf,rhyme->label);
		PrintEndTag(outf,syl->label);
		ForceNewline(outf);

		data = getconseq(data,cseq);
		parseconseq(cseq,dummy1,dummy2,ambi1,cseq1);

		data = getnucleus(data,vseq);
		getconseq(data,cseq);
		parseconseq(cseq,cseq2,ambi2,dummy1,dummy2);
	}

}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	inf, outf;
	NSL_Doctype 	idct=NULL;
	NSL_Doctype 	odct=NULL;
	NSL_Item 	*item;
	NSL_Bit		*bit;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"I"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: parse segmental transcription Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	default:
		errflg=1;
	}
	if (errflg)
		LTNSL_ERROR1(0,"usage: %s (-I) file.seg.xml",PROGNAME);

	/* get the name of the DTD */
	strcpy(dtdname,"/");
	if (argv[0][1]==':')
		strcat(dtdname,&argv[0][2]);
	else
		strcat(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((inf = SFopen(dtdname, odct, NSL_read))==NULL)
		LTNSL_ERROR1(0,"unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(inf);
	SFclose(inf);
	
	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, idct, NSL_read))==NULL)
			LTNSL_ERROR1(0,"unable to open '%s'",filename);
	}
	else
		LTNSL_ERROR(0,"no input file specified");
	idct = DoctypeFromFile(inf);

	/* Open the output NSGML stream using the segparse doctype declaration */
	outf = SFFopen(stdout, odct, NSL_write_normal, "stdout");

	/* read first tag - should be <AN> */
	if ((bit=GetNextBit(inf))==NULL)
		LTNSL_ERROR(0,"Could not read <AN> start tag");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"AN")))
		LTNSL_ERROR(0,"Could not read <AN> start tag");
	PrintBit(outf,bit);
	FreeBit(bit);

	/* get references to input elements and attributes */
	tagSEG = ElementUniqueName(idct,"SEG",0);
	attrID = AttrUniqueName(idct,"ID",0);

	/* make an utterance item */
	utt = NewNullNSLItem(odct,"UTT",0);
	NewAttrVal(utt,AttrUniqueName(odct,"ID",0),uttID);
	NewAttrVal(utt,AttrUniqueName(odct,"HREF",0),uttHREF);

	/* make a syllable item */
	syl = NewNullNSLItem(odct,"SYL",0);
	NewAttrVal(syl,AttrUniqueName(odct,"ID",0),sylID);
	NewAttrVal(syl,AttrUniqueName(odct,"HREF",0),sylHREF);
	
	/* make a onset item */
	ons = NewNullNSLItem(odct,"ONSET",0);
	NewAttrVal(ons,AttrUniqueName(odct,"ID",0),onsID);
	NewAttrVal(ons,AttrUniqueName(odct,"HREF",0),onsHREF);
	
	/* make a nucleus item */
	nuc = NewNullNSLItem(odct,"NUC",0);
	NewAttrVal(nuc,AttrUniqueName(odct,"ID",0),nucID);
	NewAttrVal(nuc,AttrUniqueName(odct,"HREF",0),nucHREF);
	
	/* make a rhyme item */
	rhyme = NewNullNSLItem(odct,"RHYME",0);
	NewAttrVal(rhyme,AttrUniqueName(odct,"ID",0),rhymeID);
	NewAttrVal(rhyme,AttrUniqueName(odct,"HREF",0),rhymeHREF);
	
	/* make a coda item */
	coda = NewNullNSLItem(odct,"CODA",0);
	NewAttrVal(coda,AttrUniqueName(odct,"ID",0),codaID);
	NewAttrVal(coda,AttrUniqueName(odct,"HREF",0),codaHREF);
	
	/* make a segment item */
	seg = NewNullNSLItem(odct,"SEG",0);
	NewAttrVal(seg,AttrUniqueName(odct,"ID",0),segID);
	NewAttrVal(seg,AttrUniqueName(odct,"HREF",0),segHREF);
	NewAttrVal(seg,AttrUniqueName(odct,"AMBI",0),segAMBI);
	
	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(inf))!=NULL) {
		if (bit->type==NSL_start_bit) {
			/* should be an utterance */
			item=ItemParse(inf,bit->value.item);
			if (strcmp(item->label,"UTT"))
				LTNSL_ERROR(0,"Expected <UTT> tag not found");

			/* output utterance start tag */
			sprintf(uttID,"UTT%d",++uttno);
			sprintf(uttHREF,"%s#ID(%s)",filename,GetAttrStringVal(item,attrID));
			PrintItem(outf,utt);
			
			/* parse segment sequence */
			segparse(outf,item);

			/* output utterance end tag */
			PrintEndTag(outf,utt->label);
			ForceNewline(outf);

			FreeItem(item);
		}
		else
			PrintBit(outf,bit);
		FreeBit(bit);
	}

	/* that's all folks */
	SFclose(outf);
	return 0;
}
