/* uttalign -- align an utterance to a set of annotations */

#undef IAG

/* Mark Huckvale - University College London */

/* version 1.0 - April 1998 */
/* version 1.1 - August 1998
	- split duration of nuclei of two equal segments
	  into two equal intervals
*/
/* version 1.2 - October 1998
	- add in segment /?/ as kind of vowel
	- be flexible over annotations
*/
/* version 1.3 - April 1999
	- add in special detection of stop closures & releases
*/
/* version 1.4 - January 2000
	- improve alignment for vowel sequences
*/
/* version 2.0 - April 2001
	- allow alignment to annotations held in XML structure
*/

/*
 * uttalign takes the syllable structure from an
 * utterance file and matches it to a SAMPA transcription.
 * The annotation can either be in an annotation item of an
 * SFS file or can be stored in the LBDATA data section of
 * the XML file.
 * The program then stores START and STOP times as features
 * onto element within th IP the utterance.
 */
 
#define PROGNAME "uttalign"
#define PROGVERS "2.0"
char *progname=PROGNAME;

#include "SFSCONFG.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "sfs.h"
#include "lt-memory.h"
#include "nsl.h"
#include "segment.h"

/* input file */
char		filename[256];	/* input file (from uttcomp) */
char		ofilename[256];	/* output filename */
char		tfilename[256];	/* temporary output filename */
const char	*attrWREF;
const char	*attrID;
const char	*attrSTART;
const char	*attrSTOP;
const char	*attrRELEASE;
const char	*attrTIME;
const char	*attrNAME;
int debug=0;

/* SFS file */
char		sfilename[256];
struct item_header	anitem;
struct an_rec	*an;
int		dosfs=0;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;

/* phone features */
const char *featname[]={
	"CNT",
	"VOI",
	"NAS",
	"STR",
	"SON",
	"RHO",
	"CNSGRV",
	"CNSCMP",
	"VOCGRV",
	"VOCHEIGHT",
	"VOCRND",
	"GRV",
	"HEIGHT",
	"RND",
};
#define NFEATNAME (sizeof(featname)/sizeof(char *))

const char *attrFEAT[NFEATNAME];

/* phone records */
struct phone_rec {
	const char *feat[NFEATNAME];
	NSL_Item *item;
	double	stime;
	double	etime;
	int	vowpos;		/* position in vowel nucleus */
	int	stopgap;
	double	rtime;
};

struct phone_rec *uphone;	/* utterance phones */
int	ucount;
struct phone_rec *aphone;	/* annotation phones */
int	acount;

/* DP alignment */
int	**dptab;

#define LEFT	1
#define DIAG	2
#define UP	3

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Item *item,const char *name)
{
	NSL_Data	*data;
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
			else if ((item=MyObtainItem(item,name))!=NULL)
				return(item);
		}
	}
	return(NULL);
}

/* look up a segment name */
int findsegment(const char *sym)
{
	int	lo,hi,k,c;
	char	buf[16];
	int	len;

	strncpy(buf,sym,15);
	buf[15]='\0';
	for (len=strlen(buf);len>=1;len--) {
		buf[len]='\0';
		lo = 0;
		hi = (sizeof(segtab)/sizeof(struct segment_rec))-1;
		do {
			k = (lo+hi)/2;
			c = strcmp(buf,segtab[k].sym);
			if (c < 0)
				hi = k-1;
			else if (c > 0)
				lo = k+1;
			else
				return(k);
		} while (lo <= hi);
	}
	error("unknown symbol: '%s'",sym);
	return(0);
}

/* count number of CNS and VOC units */
int countphone(NSL_Item *item,int count)
{
	NSL_Data *data;

	if (!item) return(count);

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if ((strcmp(item->label,"VOC")==0) ||
			    (strcmp(item->label,"CNS")==0)) count++;
			else
				count = countphone(item,count);
		}
	}
	return(count);
}

/* copy over CNS and VOC units */
void copyphone(NSL_Item *item)
{
	int	i,v=0;
	NSL_Data *data;

	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if ((strcmp(item->label,"VOC")==0) ||
			    (strcmp(item->label,"CNS")==0)) {
			    	uphone[ucount].item = item;
			    	for (i=0;i<NFEATNAME;i++)
			    		uphone[ucount].feat[i] = GetAttrStringVal(item,attrFEAT[i]);
				if (strcmp(item->label,"VOC")==0) uphone[ucount].vowpos = ++v;
			    	ucount++;
			}
			else
				copyphone(item);
		}
	}
}

/* distance metric */
int distance(struct phone_rec *p1,struct phone_rec *p2)
{
	int	i;
	int	dist=0;

	for (i=0;i<NFEATNAME;i++) {
		if (p1->feat[i]) {
			if (p2->feat[i]) {
				if (strcmp(p1->feat[i],p2->feat[i])==0)
					dist += 0; /* same feature - same value */
				else
					dist += 2; /* same feature - different value */
			}
			else
				dist += 4; /* different feature */
		}
		else {
			if (p2->feat[i])
				dist += 4; /* different feature */
			else
				dist += 0; /* neither has feature */
		}
	}
	if (p1->vowpos != p2->vowpos) dist++;
	return(dist << 2);
}

/* do DP alignment */
void dpalign()
{
	int	i,j;
	int	left,diag,up;

	/* allocate memory */
	dptab = (int **)calloc(ucount,sizeof(int *));
	if (!dptab)
		error("out of memory");
	for (i=0;i<ucount;i++) {
		dptab[i] = (int	*)calloc(acount,sizeof(int));
		if (!dptab[i])
			error("out of memory");
	}

	/* do dp alignment */
	for (i=0;i<ucount;i++) {
		for (j=0;j<acount;j++) {
			if ((i==0) && (j==0)) {
				left = 1000;
				diag = 0;
				up = 1000;
			}
			else if (i==0) {
				left = 4 + (dptab[i][j-1] & ~3);
				diag = 1000;
				up = 1000;
			}
			else if (j==0) {
				left = 1000;
				diag = 1000;
				up = 4 + (dptab[i-1][j] & ~3);
			}
			else {
				left = 4 + (dptab[i][j-1] & ~3);
				diag = dptab[i-1][j-1] & ~3;
				up = 4 + (dptab[i-1][j] & ~3);
			}
			dptab[i][j] = distance(&uphone[i],&aphone[j]);
			if (left < diag) {
				if (left < up)
					dptab[i][j] += left + LEFT;
				else
					dptab[i][j] += up + UP;
			}
			else {
				if (diag < up)
					dptab[i][j] += diag + DIAG;
				else
					dptab[i][j] += up + UP;
			}
		}
	}
#ifdef IAG
	printf("DPTABLE:\n");
	for (i=0;i<ucount;i++) {
		for (j=0;j<acount;j++)
			printf("%4d",dptab[i][j]);
		printf("\n");
	}
#endif
	/* transfer times */
	i = ucount-1;
	j = acount-1;
	while ((i>=0) && (j>=0)) {
		if (uphone[i].etime==0.0)
			uphone[i].etime = aphone[j].etime;
		if (aphone[j].stopgap) {
			uphone[i].stopgap = 1;
			uphone[i].rtime = aphone[j].rtime;
		}
		uphone[i].stime = aphone[j].stime;
		switch (dptab[i][j]&3) {
		case 1:	j--; break;
		case 2: i--; j--; break;
		case 3: i--; break;
		case 0: error("DP alignment fault");
		}
	}
	if (debug) {
		for (i=0;i<acount;i++) printf("%d stime=%g etime=%g\n",i,aphone[i].stime,aphone[i].etime);
		for (i=0;i<ucount;i++) printf("%d stime=%g etime=%g\n",i,uphone[i].stime,uphone[i].etime);
	}
}

double myatof(const char *str)
{
	if (str && *str && isdigit(*str))
		return(atof(str));
	else {
		fprintf(stderr,"?atof(%s)\n",(str)?str:"<null>");
		return(0.0);
	}
}

/* copy times on leaves up to root */
void uptimes(NSL_Item *item,double *start,double *stop)
{
	NSL_Item 	*first=NULL;
	NSL_Item 	*last=NULL;
	NSL_Data 	*data;
	char		buf[32];

	if (!item) {
		/* whoops ! */
		*start=0;
		*stop=0;
		return;
	}
	else if ((strcmp(item->label,"VOC")==0) ||
		 (strcmp(item->label,"CNS")==0)) {
		/* is a leaf node - should be timed already */
		*start = myatof(GetAttrStringVal(item,attrSTART));
		*stop = myatof(GetAttrStringVal(item,attrSTOP));
		return;
	}

	/* find first and last children of node */
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			/* recurse into children */
			uptimes(data->first,start,stop);
			/* keep record of first and last */
			if (!first) first=data->first; else last=data->first;
		}
	}
	if (!first)
		error("tried to find children of node with no children");
	if (!last) last=first;

	/* get times from first and last children */
	*start = myatof(GetAttrStringVal(first,attrSTART));
	*stop = myatof(GetAttrStringVal(last,attrSTOP));

	/* store times on this node */
	sprintf(buf,"%.4f",*start);
	PutAttrVal(item,attrSTART,strsave(buf));
	sprintf(buf,"%.4f",*stop);
	PutAttrVal(item,attrSTOP,strsave(buf));

}
		
/* process an utterance */
void process(NSL_Item *ipitem)
{
	int		i,j,idx;
	int		ulen;
	int		alen;
	NSL_Item	*item;
	char		buf[32],*p;
	double		start,stop,dur;
	
	/* load the segments into temporary structure */
	ulen = countphone(ipitem,0);
	alen = anitem.numframes;

	/* get memory */
	uphone = (struct phone_rec *)calloc(ulen,sizeof(struct phone_rec));
	aphone = (struct phone_rec *)calloc(2*alen,sizeof(struct phone_rec));
	if (!uphone || !aphone)
		error("out of memory");

	/* copy over the utterance phones */
	ucount=0;
	copyphone(ipitem);
	if (debug) {
		fprintf(stderr,"UTTERANCE:\n");
		for (i=0;i<ucount;i++) {
			fprintf(stderr,"%2d. ",i);
			for (j=0;j<NFEATNAME;j++)
				fprintf(stderr,"%s:%s ",attrFEAT[j],uphone[i].feat[j]);
			fprintf(stderr,"\n");
		}
	}

	/* copy over the annotation phones */
	acount=0;
	for (i=0;i<alen;i++) {
	   strcpy(buf,an[i].label);
	   p = strtok(buf,"+");
	   while (p && *p) {
		idx = findsegment(p);
		if (!(segtab[idx].attr & SKIP)) { 
			item = GetItemFromString(segtab[idx].elem,odct);
		    	for (j=0;j<NFEATNAME;j++)
			    	aphone[acount].feat[j] = GetAttrStringVal(item,attrFEAT[j]);
			aphone[acount].stime = anitem.offset + an[i].posn * anitem.frameduration;
			if (i==(alen-1))
				aphone[acount].etime = anitem.offset + (an[i].posn+an[i].size) * anitem.frameduration;
			else
				aphone[acount].etime = anitem.offset + an[i+1].posn * anitem.frameduration;
			if (segtab[idx].attr & STOPGAP) {
				aphone[acount].stopgap = 1;
				aphone[acount].rtime = aphone[acount].etime;
			}
			else
				aphone[acount].stopgap = 0;
			if (strcmp(item->label,"VOC")==0)
				aphone[acount].vowpos=1;
			else
				aphone[acount].vowpos=0;
			if ((acount==0)||(distance(&aphone[acount-1],&aphone[acount])!=0))
				/* keep this one */
				acount++;
			else {
				/* merge symbol with last */
				if (aphone[acount].stime < aphone[acount-1].stime) aphone[acount-1].stime = aphone[acount].stime;
				if (aphone[acount].etime > aphone[acount-1].etime) aphone[acount-1].etime = aphone[acount].etime;
				if (aphone[acount].stopgap) {
					aphone[acount-1].stopgap = aphone[acount].stopgap;
					aphone[acount-1].rtime = aphone[acount].rtime;
				}
			}
			if (*segtab[idx].elem2) {
				item = GetItemFromString(segtab[idx].elem2,odct);
			    	for (j=0;j<NFEATNAME;j++)
				    	aphone[acount].feat[j] = GetAttrStringVal(item,attrFEAT[j]);
				aphone[acount].stime = anitem.offset + an[i].posn * anitem.frameduration;
				if (i==(alen-1))
					aphone[acount].etime = anitem.offset + (an[i].posn+an[i].size) * anitem.frameduration;
				else
					aphone[acount].etime = anitem.offset + an[i+1].posn * anitem.frameduration;
				if (strcmp(item->label,"VOC")==0)
					aphone[acount].vowpos=2;
				else
					aphone[acount].vowpos=0;
				if ((acount==0)||(distance(&aphone[acount-1],&aphone[acount])!=0))
					/* keep this one */
					acount++;
				else {
					/* merge symbol with last */
					if (aphone[acount].stime < aphone[acount-1].stime) aphone[acount-1].stime = aphone[acount].stime;
					if (aphone[acount].etime > aphone[acount-1].etime) aphone[acount-1].etime = aphone[acount].etime;
				}
			}
		}
		p = strtok(NULL,"+");
	   }
	}
	if (debug) {
		fprintf(stderr,"ANNOTATION:\n");
		for (i=0;i<acount;i++) {
			fprintf(stderr,"%2d. ",i);
			for (j=0;j<NFEATNAME;j++)
				fprintf(stderr,"%s:%s ",attrFEAT[j],aphone[i].feat[j]);
			fprintf(stderr," s=%.4f e=%.4f sg=%d\n",aphone[i].stime,aphone[i].etime,aphone[i].stopgap);
		}
	}

	/* do DP alignment */
	dpalign();

	/* store start and end times on segments */
	for (i=0;i<ucount;i++) {
		if ((i<ucount-1) &&
		    ((uphone[i].stime==uphone[i+1].stime) ||
		    (uphone[i].etime==uphone[i+1].etime)) &&
		    (strcmp(uphone[i].item->label,"VOC")==0)) {
			/* two vowels with identical starts or stops - split interval */
			dur = uphone[i+1].etime - uphone[i].stime;
			sprintf(buf,"%.4f",uphone[i].stime);
			PutAttrVal(uphone[i].item,attrSTART,strsave(buf));
			sprintf(buf,"%.4f",uphone[i].stime+dur/2);
			PutAttrVal(uphone[i].item,attrSTOP,strsave(buf));
			PutAttrVal(uphone[i+1].item,attrSTART,strsave(buf));
			sprintf(buf,"%.4f",uphone[i+1].etime);
			PutAttrVal(uphone[i+1].item,attrSTOP,strsave(buf));
			i++;
		}
		else {
			sprintf(buf,"%.4f",uphone[i].etime);
			PutAttrVal(uphone[i].item,attrSTOP,strsave(buf));
			sprintf(buf,"%.4f",uphone[i].stime);
			PutAttrVal(uphone[i].item,attrSTART,strsave(buf));
			if (uphone[i].stopgap) {
				sprintf(buf,"%.4f",uphone[i].rtime);
				PutAttrVal(uphone[i].item,attrRELEASE,strsave(buf));
			}
		}
	}

	/* move times onto higher components */
	uptimes(ipitem,&start,&stop);
}

/* recurse down to find syllables and WREFs */
void scanforsyl(NSL_Item *item,char **wref,double *start,double *stop,int wscount)
{
	NSL_Data *data;
	const char	*wid;
	int	i;
	double	t;

	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"SYL")==0) {
				wid = GetAttrStringVal(item,attrWREF);
				for (i=0;i<wscount;i++)
					if (strcmp(wid,wref[i])==0) {
						t = myatof(GetAttrStringVal(item,attrSTART));
						if (t < start[i]) start[i]=t;
						t = myatof(GetAttrStringVal(item,attrSTOP));
						if (t > stop[i]) stop[i]=t;
					}
			}
			else
				scanforsyl(item,wref,start,stop,wscount);
		}
	}
}

/* copy times from IP item to WORDSEQ item */
void copytimes(NSL_Item *ipitem,NSL_Item *wsitem)
{
	NSL_Item	*item;
	NSL_Data 	*data;
	int		wscount;
	char		**wref;
	double		*start;
	double		*stop;
	char		buf[32];

	/* count the number of words */
	wscount=0;
	for (data=wsitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item=data->first;
			if (strcmp(item->label,"WORD")==0) wscount++;
		}
	}

	/* allocate some space to hold times */
	wref = (char **)calloc(wscount,sizeof(char *));
	start = (double *)calloc(wscount,sizeof(double));
	stop = (double *)calloc(wscount,sizeof(double));
	if (!wref || !start || !stop)
		error("out of memory");

	/* get the word IDs */
	wscount=0;
	for (data=wsitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item=data->first;
			wref[wscount] = strsave(GetAttrStringVal(item,attrID));
			start[wscount]=10E6;
			stop[wscount]=0.0;
			wscount++;
		}
	}

	/* scan through the IP syllables */
	scanforsyl(ipitem,wref,start,stop,wscount);

	/* and copy times onto words */
	wscount=0;
	for (data=wsitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item=data->first;
			if (strcmp(item->label,"WORD")==0) {
				sprintf(buf,"%.4f",start[wscount]);
				PutAttrVal(item,attrSTART,strsave(buf));
				sprintf(buf,"%.4f",stop[wscount]);
				PutAttrVal(item,attrSTOP,strsave(buf));
				wscount++;
			}
		}
	}

	free(wref);
	free(start);
	free(stop);
}

/* copy labels into SFS format */
void copylabels(NSL_Item *lbitem)
{
	NSL_Data	*data;
	NSL_Item	*item;
	int		i;
	double		t;

	/* free existing */
	if (an) free(an);

	/* initialise anitem */
	anitem.numframes=0;
	anitem.frameduration=1.0E-6;
	anitem.offset=0;

	/* count number of annotations */
	for (data=lbitem->data;data!=NULL;data=data->next) {
		if (data->type==NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"LABEL")==0)
				anitem.numframes++;
		}
	}

	/* allocate some memory */
	an = (struct an_rec *)calloc(anitem.numframes,sizeof(struct an_rec));

	/* transfer annotations */
	i=0;
	for (data=lbitem->data;data!=NULL;data=data->next) {
		if (data->type==NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"LABEL")==0) {
				t = myatof(GetAttrStringVal(item,attrTIME));
				an[i].posn = (int)(t/anitem.frameduration);
				an[i].size = 0;
				if (i>0) an[i-1].size = an[i].posn-an[i-1].posn;
				an[i].label = GetAttrStringVal(item,attrNAME);
				i++;
			}
		}
	}
	
}

/* process an utterance */
void uttprocess(NSL_Item *uttitem)
{
	NSL_Item 	*ipitem;
	NSL_Item 	*lbitem;
	NSL_Item 	*wordseqitem;

	/* locate annotations if not SFS */
	if (!dosfs) {
		lbitem = MyObtainItem(uttitem,"LBDATA");
		if (!lbitem)
			error("no LBDATA item in input (and no SFS file supplied)");
		copylabels(lbitem);
	}

	/* find the IP item */
	ipitem = MyObtainItem(uttitem,"IP");

	/* check */
	if (!ipitem)
		error("No IP item in input");

	/* OK, let's go */
	process(ipitem);

	/* copy the times from the IP item to the WORDSEQ item (if exists) */
	wordseqitem = MyObtainItem(uttitem,"WORDSEQ");
	if (wordseqitem!=NULL)
		copytimes(ipitem,wordseqitem);
}

void finduttprocess(NSL_Item *item)
{
	NSL_Data	*data;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"UTT")==0)
				uttprocess(item);
			else
				finduttprocess(item);
		}
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	inf;
	NSL_Item 	*uttitem;
	NSL_Bit		*bit;
	int		i;

	int		it;
	char		*ty;
	char		*antype="0";

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ii:s:D"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance alignment to SFS annotation Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'i' :	/* input specification */
		if (itspec(optarg,&it,&ty)==0) {
			if (it==AN_TYPE)
				antype=ty;
			else
				error("unsuitable item specification: '%s'",optarg);
		}
		else
			error("illegal item specification: '%s'",optarg);
		break;
	case 's':
		strcpy(sfilename,optarg);
		dosfs=1;
		break;
	case 'D':
		debug=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-i item) (-s input.sfs) (input.xml) (output.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((inf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(inf);
	SFclose(inf);

	/* get unique names for attributes */
	attrID = AttrUniqueName(odct,"ID",0);
	attrWREF = AttrUniqueName(odct,"WREF",0);
	attrSTART = AttrUniqueName(odct,"START",0);
	attrSTOP = AttrUniqueName(odct,"STOP",0);
	attrRELEASE = AttrUniqueName(odct,"RELEASE",0);
	attrTIME = AttrUniqueName(odct,"TIME",0);
	attrNAME = AttrUniqueName(odct,"NAME",0);
	for (i=0;i<NFEATNAME;i++)
		attrFEAT[i] = AttrUniqueName(odct,featname[i],0);

	/* open the SFS file */
	if (dosfs)
		getitem(sfilename,AN_TYPE,antype,&anitem,(void *)&an);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, odct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else {
		strcpy(filename,"STDIN");
		if ((inf = SFFopen(stdin, odct, NSL_read,filename))==NULL)
			error("unable to open '%s'",filename);
	}

	/* Open the output NSGML stream using the doctype declaration */
	optind++;
	if (optind < argc) {
		strcpy(ofilename,argv[optind]);
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
		outf = SFopen(tfilename, odct, NSL_write_pretty);
	}
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* read first tag - should be <UTT> or <PROSYNTH> */
	while ((bit=GetNextBit(inf))!=NULL) {

		if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
			error("Could not read <UTT> or <PROSYNTH> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		/* find and process UTT items */
		if (strcmp(uttitem->label,"UTT")==0)
			uttprocess(uttitem);
		else
			finduttprocess(uttitem);

		/* print the utterance with modified word sequence */
		PrintItem(outf,uttitem);
		ForceNewline(outf);
	}

	/* that's all folks */
	SFclose(inf);
	SFclose(outf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
