/* uttclass -- classify words in utterance */

/* Mark Huckvale - University College London */

/* version 1.1 - March 1998 */

/*
 * usage: uttclass (-I) -l lexicon.xml input.xml (output.xml)
 *
 * uttclass adds lexical reference information to the words in
 * a WORDSEQ item within an utterance annotation file.  This
 * information should be generated as a by-product of parsing
 * the text and assigning parts-of-speech to the words.  These
 * in turn allow the appropriate pronunciation for the word to be
 * retrieved from the lexicon by the the 'uttcomp' program.
 *
 * This program is merely a shell which locates the lexical
 * entry for each word and adds a reference to the first pronunciation
 * in the lexicon.
 */
 
#define PROGNAME "uttclass"
#define PROGVERS "1.1"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"

/* input file */
char		filename[256];	/* input file (from uttmake) */
char		ofilename[256];	/* output filename */
char		tfilename[256];	/* temporary output filename */

/* lexicon file */
char		lfilename[256];	/* lexicon file (from lexparse) */
NSL_File	lexf;
NSL_Doctype	ldct=NULL;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;
const char *	attrID;
const char *	lexattrID;
const char *	attrLREF;	/* lexical reference */
const char *	attrPREF;	/* pronunciation reference */
char		wordID[16];

/* word sequence to process */
NSL_Item	*wordseqitem;

/* structure to hold words */
struct word_rec {
	char		*str;
	NSL_Item	*worditem;
	struct word_rec	*next;
};

struct word_rec *wroot;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* check for a word match */
int	matchword(char *str)
{
	struct word_rec *w=wroot;

	while (w) {
		if (strcmp(w->str,str)==0) return(1);
		w = w->next;
	}
	return(0);
}

/* find words in lexicon */
void findwords()
{
	NSL_Bit		*bit;
	NSL_Item	*item;
	NSL_Data	*data;
	NSL_Item	*pitem;
	NSL_Data	*pdata;
	struct word_rec *w;
	const char	*lexID;
	char		lexIDbuf[512];
	const char 	*pref;

	/* read first tag - should be <LEXICON> */
	if ((bit=GetNextBit(lexf))==NULL)
		error("Could not read <LEXICON> start tag");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"LEXICON")))
		error("Could not read <LEXICON> start tag");
	FreeBit(bit);

	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(lexf))!=NULL) {
		
		if (bit->type==NSL_start_bit) {
			
			/* should be an entry */
			item=ItemParse(lexf,bit->value.item);
			if (strcmp(item->label,"ENTRY"))
				error("Expected <ENTRY> tag not found");

			/* check HEADWORD entry */
			for (data=item->data;data!=NULL;data = data->next) {
				if (data->type==NSL_item_data) {
					pitem = data->first;
					if (strcmp(pitem->label,"HW")==0) {
						/* found headword entry */
						pdata = pitem->data;
						if (pdata->type==NSL_text_data) {
							w = wroot;
							while (w) {
								if (strcmp(w->str,pdata->first)==0) {
									lexID = GetAttrStringVal(item,lexattrID);
									sprintf(lexIDbuf,"%s#ID(%s)",lfilename,lexID);
									PutAttrVal(w->worditem,attrLREF,strsave(lexIDbuf));
									pref = GetAttrStringVal(w->worditem,attrPREF);
									if (!pref || (*pref=='\0'))
										PutAttrVal(w->worditem,attrPREF,strsave("#1"));
								}
								w = w->next;
							}
							break;
						}
					}
				}
			}

			FreeItem(item);
		}
		else
			FreeBit(bit);
	}

}

/* process word sequence */
void	process(NSL_Item *wordseqitem)
{
	struct word_rec *w,*wn;
	struct word_rec **wlast=&wroot;
	NSL_Item	*item;
	NSL_Data	*data;
	NSL_Data	*pdata;
	const char	*str;
	int		wordmiss=0;
	
	/* load words into temporary structure */
	for (data=wordseqitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORD")==0) {

				/* store information */
				w = (struct word_rec *)calloc(1,sizeof(struct word_rec));
				w->worditem = item;	/* pointer to actual data */
				pdata = item->data;
				if (pdata->type == NSL_text_data)
					w->str = strsave(item->data->first);
				else
					w->str = strsave("");

				/* add to end of chain */
				*wlast = w;
				wlast = &w->next;
			}
		}
	}

	/* open the lexicon */
	if (lfilename[0]=='\0')
		error("no lexicon specified");
	if ((lexf = SFopen(lfilename,ldct,NSL_read))==NULL)
		error("unable to open '%s'",lfilename);
	ldct = DoctypeFromFile(lexf);
	lexattrID = AttrUniqueName(ldct,"ID",0);

	/* match words against lexicon */
	findwords();

	/* close lexicon */
	SFclose(lexf);

	/* check all words accounted for */
	w = wroot;
	while (w) {
		str = GetAttrStringVal(w->worditem,attrLREF);
		if (!str || !*str) {
			if (wordmiss==0) fprintf(stderr,"Missing words:\n");
			fprintf(stderr,"%s\n",w->str);
			wordmiss=1;
		}
		w = w->next;
	}

	/* free word rec structures */
	w = wroot;
	while (w) {
		wn = w->next;
		free(w->str);
		free(w);
		w = wn;
	}
	wroot=NULL;

	if (wordmiss) exit(1);
}

/* recursively search for and process WORDSEQ items */
void findprocess(NSL_Item *item)
{
	NSL_Data	*data;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORDSEQ")==0)
				process(item);
			else
				findprocess(item);
		}
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	dinf;
	NSL_File 	inf;
	NSL_Doctype 	idct=NULL;
	NSL_Item 	*item;
	NSL_Item 	*uttitem;
	NSL_Bit		*bit;
	const char	*attrbuf;
	NSL_ElementSummary defn;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Il:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: phrase composition from lexicon Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'l':	/* lexicon name */
		strcpy(lfilename,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) -l lexicon.xml input.xml (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((dinf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(dinf);
	SFclose(dinf);
	
	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, idct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else {
		strcpy(filename,"STDIN");
		if ((inf = SFFopen(stdin, idct, NSL_read,filename))==NULL)
			error("unable to open '%s'",filename);
	}
	idct = DoctypeFromFile(inf);

	/* add PREF and LREF attributes to input DTD */
	item = NewNullNSLItem(idct,"WORD",0);
	defn = (NSL_ElementSummary) item->defn;
	attrbuf = "PREF";
	FindAttrSumAndName(idct,&defn,item->label,&attrbuf,0);
	attrPREF = attrbuf;
	attrbuf = "LREF";
	FindAttrSumAndName(idct,&defn,item->label,&attrbuf,0);
	attrLREF = attrbuf;
	FreeItem(item);

	/* get output attribute names */
	attrID = AttrUniqueName(odct,"ID",0);

	/* Open the output stream using the output doctype declaration */
	optind++;
	if (optind < argc) {
		strcpy(ofilename,argv[optind]);
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
		outf = SFopen(tfilename, odct, NSL_write_pretty);
	}
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* read first tag - should be <UTT> or <PROSYNTH> */
	while ((bit=GetNextBit(inf))!=NULL) {
		if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
			error("Could not read <UTT> or <PROSYNTH> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		/* find and process the WORDSEQ item */
		findprocess(uttitem);

		/* print the utterance with modified word sequence */
		PrintItem(outf,uttitem);
		ForceNewline(outf);

	}

	/* close input file */
	SFclose(inf);

	/* that's all folks */
	SFclose(outf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
