/* uttcomp -- utterance composition from lexicon */

/* Mark Huckvale - University College London */

/* version 1.0 - March 1998 */
/* version 1.1 - November 1998
	- put initial weak foot in own AG
*/
/* version 1.2 - July 1999
	- each foot has own AG
	- initial AG is weak
*/
/* version 1.3 - August 2000
	- fix pick-up of strong/weak pronunciation variants
*/

/*
 * uttcomp composes a pronunciation for an utterance from
 * syllable structures stored in a lexicon.  Each WORD in
 * the WORDSEQ item for an utterance is looked up in the
 * lexicon file and the appropriate SYLSEQ entries are
 * concatenated into an IP + AG + FOOT structure with
 * minor modifications where necessary
 *
 * uttcomp is normally run after the parse of the utterance
 * so that each word can be given a part-of-speech tag and
 * hence the appropriate pronunciation for the word may
 * be selected from the lexicon.
 *
 * In fact, uttcomp only looks for LREF and PREF attributes
 * on the WORD tags.  LREF should be a reference to the
 * lexical entry for the word, while PREF should indicate
 * the ID of the appropriate pronunciation within that entry.
 * uttcomp does not care how these references have been
 * generated, and a program uttclass provides a default
 * implementation.
 */
 
#define PROGNAME "uttcomp"
#define PROGVERS "1.3"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"

/* input file */
char		filename[256];	/* input file (from uttmake + uttclass) */
char		ofilename[256];	/* output filename */
char		tfilename[256];	/* temporary output filename */
const char	*attrLREF;
const char	*attrPREF;
const char	*attrWREF;
const char 	*attrSTRESS;
const char 	*attrSTRENGTH;
const char 	*attrTYPE;
const char	*attrID;
const char	*attrFPOS;
const char	*attrRFPOS;
const char	*attrWPOS;
const char	*attrRWPOS;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;

/* token to process */
char	*token;

/* word records */
struct word_rec {
	char	*str;		/* word string */
	int	flags;		/* stress flags */
	char	strpat[16];	/* stress pattern */
	NSL_Item *worditem;	/* pointer to word item */
	NSL_Item *sylseq;	/* syllable sequence from lexicon */
	char	*lexname;	/* lexicon file */
	char	*lexid;		/* lexicon entry ID */
	char	*prnid;		/* pronunciation ID */
	const char	*wrdid;		/* word ID */
	struct word_rec *next;
};

#define WFLAG_STRESS_NONE	0
#define WFLAG_STRESS_FOOT	1
#define WFLAG_STRESS_ACCENT	2

struct word_rec *wroot;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory\n");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Data *data,const char *name)
{
	NSL_Item *item;
	for (;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
		}
	}
	return(NULL);
}

/* a working version of DeleteItems */
void MyDeleteItems(NSL_Item *parent,const char *name)
{
	NSL_Item *item;
	NSL_Data *data;
	NSL_Data *last=NULL;

	for (data=parent->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0) {
				if (!last)
					parent->data = data->next;
				else
					last->next = data->next;
				FreeItem(item);
			}
		}
		last = data;
	}
}

/* parse a lexical reference */
void parselexref(const char *str,char **lname,char **lid)
{
	char	name[256];
	char	id[256];
	int	i,j;

	name[0]='\0';
	id[0]='\0';
	for (i=0,j=0;(str[i]!='\0') && (str[i]!='#');i++,j++)
		name[j] = str[i];
	name[j]='\0';
	if (str[i]=='#') i++;
	if (str[i]=='I') i++;
	if (str[i]=='D') i++;
	if (str[i]=='(') {
		for (j=0,i++;(str[i]!='\0') && (str[i]!=')');i++,j++)
			id[j] = str[i];
		id[j]='\0';
	}
	*lname = strsave(name);
	*lid = strsave(id);
}

/* find given pronunciation within PRONSEQ item */
NSL_Item *findpron(NSL_Item *pitem,char *prnid)
{
	NSL_Data	*data;
	NSL_Item	*item;
	const char 	*pronID;

	if (!pitem) return(NULL);

	for (data=pitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"PRON")==0) {
				pronID = GetAttrStringVal(item,attrID);
				if (strcmp(pronID,prnid)==0)
					/* got it ! */
					return(CopyItem(MyObtainItem(item->data,"SYLSEQ")));
			}
		}
	}
	return(NULL);
}

/* find words in lexicon */
void findwords(char *lfilename)
{
	NSL_File	lexf;

	NSL_Bit		*bit;
	NSL_Item	*item;
	const char	*entryID;
	struct word_rec	*w;

	/* open the lexicon file */
	if ((lexf = SFopen(lfilename,odct,NSL_read))==NULL)
		error("unable to open lexicon file '%s'",lfilename);

	/* read first tag - should be <LEXICON> */
	if ((bit=GetNextBit(lexf))==NULL)
		error("Could not read lexicon file");
	while (bit->type==NSL_pi_bit)
		if ((bit=GetNextBit(lexf))==NULL)
			error("Could not read lexicon file");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"LEXICON")))
		error("Could not read <LEXICON> start tag");
	FreeBit(bit);

	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(lexf))!=NULL) {
		
		if (bit->type==NSL_start_bit) {
			
			/* should be an entry */
			item=ItemParse(lexf,bit->value.item);
			if (strcmp(item->label,"ENTRY"))
				error("Expected <ENTRY> tag not found");

			/* check if any word references this entry */
			entryID = GetAttrStringVal(item,attrID);

			w = wroot;
			while (w) {
				if ((w->sylseq==NULL) &&
				    (strcmp(w->lexname,lfilename)==0) &&
				    (strcmp(w->lexid,entryID)==0)) {
					/* yes - find pronunciation */
					w->sylseq = findpron(MyObtainItem(item->data,"PRONSEQ"),w->prnid);
				}
				w = w->next;
			}

			FreeItem(item);
		}
		else
			FreeBit(bit);
	}

	SFclose(lexf);
}

/* add reverse foot position attributes to syl seq under foot */
void addRFPOS(NSL_Item *foot,int footcnt)
{
	NSL_Data	*data;
	char		buf[32];

	for (data=foot->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			sprintf(buf,"%d",footcnt--);
			PutAttrVal(data->first,attrRFPOS,strsave(buf));
		}
	}	
}

void PutAttrValRecurse(NSL_Item *item,const char *attr,const char *val)
{
	NSL_Data *data;
	const char *oldval;

	oldval = GetAttrStringVal(item,attr);
	if (oldval && (strcmp(oldval,val)!=0))
		PutAttrVal(item,attr,strsave(val));

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			PutAttrValRecurse(data->first,attr,val);
		}
	}	
}

/* compose from lexicon */
NSL_Item *compose(NSL_Item *wordseqitem)
{
	struct word_rec *w;
	struct word_rec **wlast=&wroot;
	NSL_Data	*data;
	NSL_Data	*pdata;
	NSL_Item	*item;
	NSL_Item	*ipitem;
	NSL_Item	*agitem;
	NSL_Item	*ftitem;
	const char	*str;
	const char	*lref;
	const char	*pref;
	int		footcnt;
	int		wordcnt,strcnt;
	char		buf[32];
	int		firststress=1;

	/* copy words into processing structure */
	for (data=wordseqitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORD")==0) {

				/* store information */
				w = (struct word_rec *)calloc(1,sizeof(struct word_rec));
				w->worditem = item;	/* pointer to actual data */
				pdata = item->data;
				if (pdata->type == NSL_text_data)
					w->str = strsave(item->data->first);
				else
					w->str = strsave("");

				/* get lexical reference */
				lref = GetAttrStringVal(item,attrLREF);
				if (!lref)
					error("Word '%s' has no Lexical reference (LREF)\n",w->str);
				parselexref(lref,&w->lexname,&w->lexid);

				/* get pronunciation reference */
				pref = GetAttrStringVal(item,attrPREF);
				if (!pref)
					error("Word '%s' has no pronunciation reference (PREF)\n",w->str);
				if (pref[0]=='#')
					w->prnid = strsave(pref+1);
				else
					w->prnid = strsave(pref);

				/* get Word ID */
				w->wrdid = GetAttrStringVal(item,attrID);

				/* see if word carries stress */
				strcpy(w->strpat,GetAttrStringVal(item,attrSTRESS));
				if (strchr(w->strpat,'A')!=NULL)
					w->flags |= WFLAG_STRESS_ACCENT;
				else if (strchr(w->strpat,'F')!=NULL)
					w->flags |= WFLAG_STRESS_FOOT;
				else
					w->flags |= WFLAG_STRESS_NONE;

				/* add to end of chain */
				*wlast = w;
				wlast = &w->next;
			}
		}
	}

	/* match words against lexicon */
	w = wroot;
	while (w) {
		/* we call findwords for each missing SYLSEQ
		   but in fact findwords will find all references
		   to a given lexicon in one pass through WORDSEQ */
		if (!w->sylseq) findwords(w->lexname);
		w = w->next;
	}

	/* create new intonational phrase */
	ipitem = newItem("IP");

	/* create new accent group */
	agitem = newItem("AG");

	/* add words into FOOT */
	ftitem = newItem("FOOT");
	footcnt=0;
	w = wroot;
	while (w) {
		/* check we found word in lexicon */
		if (!w->sylseq)
			error("No lexicon entry found for '%s'",w->str);
		/* process each syllable in turn */
		wordcnt=0;
		strcnt=0;
		for (data=w->sylseq->data;data!=NULL;data=data->next) {
			if (data->type==NSL_item_data) {
				str = GetAttrStringVal(data->first,attrSTRESS);
				/* check if this is a stressed syllable */
				if (strcmp(str,"1")==0) {
					/* check if new AG required here */
					if ((w->strpat[strcnt]=='A') || (w->strpat[strcnt]=='F')) {

						/* start a new foot */
						if (ftitem->type!=NSL_empty) {
							/* save old one */
							addRFPOS(ftitem,footcnt);
							if (firststress) {
								/* initial weak foot has own weak AG */
								PutAttrVal(ftitem,attrSTRENGTH,"WEAK");
								PutAttrVal(agitem,attrTYPE,"WEAK");
								AddItemToEnd(agitem,ftitem);
								AddItemToEnd(ipitem,agitem);
								agitem = newItem("AG");
							}
							else if (w->strpat[strcnt]=='A') {
								/* start new AG */
								AddItemToEnd(agitem,ftitem);
								AddItemToEnd(ipitem,agitem);
								agitem = newItem("AG");
							}
							else {
								/* don't start new AG */
								AddItemToEnd(agitem,ftitem);
							}
							/* start new foot */
							ftitem = newItem("FOOT");
						}
						firststress=0;
						footcnt=0;
						strcnt++;
					}
				}
				else if (strcmp(str,"2")==0) {

					/* just start a new foot */
					if (ftitem->type!=NSL_empty) {
						/* save old one */
						addRFPOS(ftitem,footcnt);
						if (firststress) {
							/* initial weak foot has own weak AG */
							PutAttrVal(ftitem,attrSTRENGTH,"WEAK");
							PutAttrVal(agitem,attrTYPE,"WEAK");
							AddItemToEnd(agitem,ftitem);
							AddItemToEnd(ipitem,agitem);
							agitem = newItem("AG");
						}
						else {
							/* don't start new AG */
							AddItemToEnd(agitem,ftitem);
						}
						/* start new foot */
						ftitem = newItem("FOOT");
					}
					firststress=0;
					footcnt=0;
				}
				else {
					/* force strength to WEAK */
					PutAttrValRecurse(data->first,attrSTRENGTH,"WEAK");
				}
				/* add word ID into syllable */
				if (w->wrdid)
					PutAttrVal(data->first,attrWREF,w->wrdid);
					
				/* add position markers to syllable */
				sprintf(buf,"%d",++footcnt);
				PutAttrVal(data->first,attrFPOS,strsave(buf));
				sprintf(buf,"%d",++wordcnt);
				PutAttrVal(data->first,attrWPOS,strsave(buf));

				/* add syllable to foot */
				AddItemToEnd(ftitem,data->first);
			}
		}
		/* go back and put RWPOS features in */
		for (data=w->sylseq->data;data!=NULL;data=data->next) {
			if (data->type==NSL_item_data) {
				sprintf(buf,"%d",wordcnt--);
				PutAttrVal(data->first,attrRWPOS,strsave(buf));
			}
		}
		w = w->next;
	}

	/* tidy away last foot */
	if (ftitem->type != NSL_empty) {
		addRFPOS(ftitem,footcnt);
		AddItemToEnd(agitem,ftitem);
	}

	/* tidy away last AG */
	AddItemToEnd(ipitem,agitem);

	return(ipitem);

}

void uttprocess(NSL_Item *uttitem)
{
	NSL_Item 	*wordseqitem;
	NSL_Item 	*ipitem;

	/* find the WORDSEQ item */
	wordseqitem = MyObtainItem(uttitem->data,"WORDSEQ");

	/* check */
	if (!wordseqitem) return;

	/* delete existing IP item */
	MyDeleteItems(uttitem,"IP");

	/* create IP item */
	ipitem = compose(wordseqitem);

	/* add IP item to end of utterance */
	AddItemToEnd(uttitem,ipitem);

}

void finduttprocess(NSL_Item *item)
{
	NSL_Data	*data;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"UTT")==0)
				uttprocess(item);
			else
				finduttprocess(item);
		}
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	NSL_Doctype 	ddct=NULL;
	char		*p;
	
	NSL_File 	inf;
	NSL_File 	dinf;
	NSL_Item 	*uttitem;
	NSL_Bit		*bit;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"I"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance composition from lexicon Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (input.xml) (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((dinf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(dinf);
	SFclose(dinf);

	/* get unique names for attributes */
	attrID = AttrUniqueName(odct,"ID",0);
	attrLREF = AttrUniqueName(odct,"LREF",0);
	attrPREF = AttrUniqueName(odct,"PREF",0);
	attrWREF = AttrUniqueName(odct,"WREF",0);
	attrSTRESS = AttrUniqueName(odct,"STRESS",0);
	attrSTRENGTH = AttrUniqueName(odct,"STRENGTH",0);
	attrTYPE = AttrUniqueName(odct,"TYPE",0);
	attrFPOS = AttrUniqueName(odct,"FPOS",0);
	attrRFPOS = AttrUniqueName(odct,"RFPOS",0);
	attrWPOS = AttrUniqueName(odct,"WPOS",0);
	attrRWPOS = AttrUniqueName(odct,"RWPOS",0);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, odct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else {
		strcpy(filename,"STDIN");
		if ((inf = SFFopen(stdin, odct, NSL_read,filename))==NULL)
			error("unable to open '%s'",filename);
	}

	/* get output DTD */
	if ((dinf = SFopen(dtdname, NULL, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	ddct = DoctypeFromFile(dinf);
	SFclose(dinf);

	/* Open the output NSGML stream using the doctype declaration */
	optind++;
	if (optind < argc) {
		strcpy(ofilename,argv[optind]);
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
		outf = SFopen(tfilename, ddct, NSL_write_pretty);
	}
	else
		outf = SFFopen(stdout, ddct, NSL_write_pretty, "stdout");

	/* read first tag - should be <UTT> or <PROSYNTH> */
	while ((bit=GetNextBit(inf))!=NULL) {

		if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
			error("Could not read <UTT> or <PROSYNTH> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		/* find and process UTT items */
		if (strcmp(uttitem->label,"UTT")==0)
			uttprocess(uttitem);
		else
			finduttprocess(uttitem);

		/* print the utterance with modified word sequence */
		PrintItem(outf,uttitem);
		ForceNewline(outf);
	}

	/* that's all folks */
	SFclose(inf);
	SFclose(outf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
