/* uttdict -- use dictionary to find pronunciations for words */

/* Mark Huckvale - University College London */

/* version 1.0 - July 1999 */

/* version 1.1 - August 2000
	- look particularly for weak pronunciations
*/

/*
 * usage: uttdict (-I) -o lexicon.xml (input.xml) (output.xml)
 *
 * uttdict determines the pronunciations of words found in
 * a WORDSEQ item within an utterance annotation file.  This
 * information is the stored in a file which has the format of
 * a lexicon.  This means that subsequent programs will 'magically'
 * find the right words in the given lexicon.
 *
 */
 
#define PROGNAME "uttdict"
#define PROGVERS "1.1"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"
#include "lookup.h"
#include "samparse.h"

/* input file */
char		filename[256];	/* input file (from uttmake) */
char		ofilename[256];	/* output filename */
char		tfilename[256];	/* temporary output filename */

/* lexicon file */
char		lfilename[256];	/* lexicon file (from lexparse) */
NSL_File	lexf;
NSL_Doctype	ldct=NULL;
NSL_Item	*lexiconitem;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;
const char *	attrID;
const char *	lexattrID;
const char *	attrLREF;	/* lexical reference */
const char *	attrPREF;	/* pronunciation reference */
const char *	attrSTRESS;	/* stress level */
char		wordID[16];

/* word sequence to process */
NSL_Item	*wordseqitem;

/* structure to hold words */
struct word_rec {
	char		*str;
	int		npron;
	int		weak;
	char		*phn;
	NSL_Item	*worditem;
	struct word_rec	*next;
};

struct word_rec *wroot;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* check for a word match */
int	matchword(char *str)
{
	struct word_rec *w=wroot;

	while (w) {
		if (strcmp(w->str,str)==0) return(1);
		w = w->next;
	}
	return(0);
}

/* find pronunciations for words */
void	findpron(int nword)
{
	struct word_rec 	*wr;
	char	**w,**p;
	char	*pref[32];
	char	*pr,*lpref;
	int	npref;
	int	i,j,n;

	w = (char **)calloc(nword,sizeof(char *));
	p = (char **)calloc(nword,sizeof(char *));
	wr = wroot;
	i=0;
	while (wr) {
		w[i] = wr->str;
		i++;
		wr=wr->next;
	}
	lookup(w,nword,p,0);
	wr = wroot;
	i=0;
	while (wr) {
		/* split into pronunciations */
		pr = strtok(p[i],"\t");
		for (npref=0;(pr!=NULL)&&(*pr!='\0')&&(npref<32);npref++) {
			pref[npref] = pr;
			pr = strtok(NULL,"\t");
		}

		if (npref==0) {
			wr->phn=NULL;
		}
		else {
			/* take any specified counts into account if possible */
			n=wr->npron;

			/* are we looking for a weak one ? */
			lpref = pref[0];
			if (wr->weak) {
				for (j=0;(j<npref)&&(n>0);j++)
					if (strchr(pref[j],'\'')==NULL) {
						lpref = pref[j];
						n--;
					}
			}
			else {
				for (j=0;(j<npref)&&(n>0);j++) {
					lpref = pref[j];
					n--;
				}
			}
			wr->phn = strsave(lpref);
			if (wr->weak) {
				/* delete any stress marks in dictionary */
				for (pr=wr->phn,j=0;wr->phn[j];j++)
					if (wr->phn[j]!='\'')
						*pr++ = wr->phn[j];
				*pr = '\0';
			}
/* fprintf(stderr,"str='%s' phn='%s' npron=%d npref=%d\n",wr->str, wr->phn, wr->npron, npref); */
		}

		free(p[i]);
		p[i]=wr->phn;
		i++;
		wr=wr->next;
	}
}

/* process word sequence */
void	process(NSL_Item *wordseqitem)
{
	struct word_rec *w,*wn;
	struct word_rec **wlast=&wroot;
	NSL_Item	*item;
	NSL_Data	*data;
	NSL_Data	*pdata;
	int		wordmiss=0;
	int		nword=0;
	NSL_Item	*entryitem;
	NSL_Item	*hwitem;
	NSL_Item	*pronseqitem;
	NSL_Item	*pronitem;
	NSL_Item	*ipaitem;
	NSL_Item	*sylseqitem;
	char		wbuf[256];
	char		uwbuf[256];
	char		lbuf[1024];
	char		*p;
	const char *	pref;
	const char *	stress;

	/* load words into temporary structure */
	for (data=wordseqitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORD")==0) {

				/* store information */
				w = (struct word_rec *)calloc(1,sizeof(struct word_rec));
				w->worditem = item;	/* pointer to actual data */
				pdata = item->data;
				if (pdata->type == NSL_text_data)
					w->str = strsave(item->data->first);
				else
					w->str = strsave("");
				pref = GetAttrStringVal(item,attrPREF);
				if (pref && *pref) {
					if (*pref=='#')
						w->npron = atoi(pref+1);
					else
						w->npron = atoi(pref);
					if (w->npron==0) w->npron=1;
				}
				else
					w->npron = 1;
		
				stress = GetAttrStringVal(item,attrSTRESS);
				if (stress && *stress) {
					if (*stress=='W')
						w->weak=1;
					else
						w->weak=0;
				}
				else
					w->weak=0;

				/* add to end of chain */
				*wlast = w;
				wlast = &w->next;
				nword++;
			}
		}
	}

	/* match words against lexicon */
	findpron(nword);

	/* check all words accounted for */
	w = wroot;
	while (w) {
		if (!w->phn) {
			if (wordmiss==0) fprintf(stderr,"Missing words:\n");
			fprintf(stderr,"%s\n",w->str);
			wordmiss=1;
		}
		w = w->next;
	}

	w = wroot;
	while (w) {
		strcpy(wbuf,strtok(w->str,"#"));
		strcpy(uwbuf,wbuf);
		for (p=uwbuf;*p;p++) if (islower(*p)) *p=toupper(*p);
		entryitem=newItem("ENTRY");
		PutAttrVal(entryitem,attrID,strsave(uwbuf));
		sprintf(lbuf,"%s#ID(%s)",lfilename,uwbuf);
		PutAttrVal(w->worditem,attrLREF,strsave(lbuf));
		if (w->weak)
			PutAttrVal(w->worditem,attrPREF,strsave("#2"));
		else
			PutAttrVal(w->worditem,attrPREF,strsave("#1"));
		sprintf(lbuf,"<HW>%s</HW>",wbuf);
		hwitem=GetItemFromString(lbuf,ldct);
		AddItemToEnd(entryitem,hwitem);
		pronseqitem=newItem("PRONSEQ");
		pronitem=newItem("PRON");
		if (w->weak)
			PutAttrVal(pronitem,attrID,strsave("2"));
		else
			PutAttrVal(pronitem,attrID,strsave("1"));
		sprintf(lbuf,"<IPA>%s</IPA>",w->phn);
		ipaitem=GetItemFromString(lbuf,ldct);
		AddItemToEnd(pronitem,ipaitem);
		sylseqitem=segparse(w->phn);
		AddItemToEnd(pronitem,sylseqitem);
		AddItemToEnd(pronseqitem,pronitem);
		AddItemToEnd(entryitem,pronseqitem);
		AddItemToEnd(lexiconitem,entryitem);
		w=w->next;
	}
	
	/* free word rec structures */
	w = wroot;
	while (w) {
		wn = w->next;
		free(w->str);
		if (w->phn) free(w->phn);
		free(w);
		w = wn;
	}
	wroot=NULL;

	if (wordmiss) exit(1);
}

/* recursively search for and process WORDSEQ items */
void findprocess(NSL_Item *item)
{
	NSL_Data	*data;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORDSEQ")==0)
				process(item);
			else
				findprocess(item);
		}
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	dinf;
	NSL_File 	inf;
	NSL_Doctype 	idct=NULL;
	NSL_Item 	*item;
	NSL_Item 	*uttitem;
	NSL_Bit		*bit;
	const char	*attrbuf;
	NSL_ElementSummary defn;

	/* Initialise the NSL API */
	NSLInit(0);

	strcpy(lfilename,"uttdict.lex");

	/* option decoding */
	while ((c=getopt(argc,argv,"Io:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: pronunciation lookup Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'o':	/* lexicon name */
		strcpy(lfilename,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) -o lexicon.xml (input.xml) (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((dinf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(dinf);
	SFclose(dinf);
	
	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, idct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else {
		strcpy(filename,"STDIN");
		if ((inf = SFFopen(stdin, idct, NSL_read,filename))==NULL)
			error("unable to open '%s'",filename);
	}
	idct = DoctypeFromFile(inf);

	/* add PREF and LREF attributes to input DTD */
	item = NewNullNSLItem(idct,"WORD",0);
	defn = (NSL_ElementSummary) item->defn;
	attrbuf = "PREF";
	FindAttrSumAndName(idct,&defn,item->label,&attrbuf,0);
	attrPREF = attrbuf;
	attrbuf = "LREF";
	FindAttrSumAndName(idct,&defn,item->label,&attrbuf,0);
	attrLREF = attrbuf;
	attrbuf = "STRESS";
	FindAttrSumAndName(idct,&defn,item->label,&attrbuf,0);
	attrSTRESS = attrbuf;
	FreeItem(item);

	/* get output attribute names */
	attrID = AttrUniqueName(odct,"ID",0);

	/* Open the output stream using the output doctype declaration */
	optind++;
	if (optind < argc) {
		strcpy(ofilename,argv[optind]);
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
		outf = SFopen(tfilename, odct, NSL_write_pretty);
	}
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* initialise the segment parser */
	ldct = initparse(dtdname);

	/* create lexicon */
	lexf = SFopen(lfilename, ldct, NSL_write_pretty);
	lexiconitem = newItem("LEXICON");

	/* read first tag - should be <UTT> or <PROSYNTH> */
	while ((bit=GetNextBit(inf))!=NULL) {
		if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
			error("Could not read <UTT> or <PROSYNTH> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		/* find and process the WORDSEQ item */
		findprocess(uttitem);

		/* print the utterance with modified word sequence */
		PrintItem(outf,uttitem);
		ForceNewline(outf);

	}

	PrintItem(lexf,lexiconitem);
	SFclose(lexf);

	/* close input file */
	SFclose(inf);

	/* that's all folks */
	SFclose(outf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
