/* uttdict -- use dictionary to find pronunciations for words */

/* Mark Huckvale - University College London */

/* version 1.0 - July 1999 */

/* version 1.1 - August 2000
	- look particularly for weak pronunciations
*/

/* version 2.0 - April 2001
	- output phonemic transcription into WORD elements
	  rather than creating external lexicon file
*/

/*
 * usage: uttdict2 (-I) (-x exceptions.dic) (input.xml) (output.xml)
 *
 * uttdict determines the pronunciations of WORDs found in
 * a PROSYNTH/UTT/WORDSEQ element.  The pronunciations are 
 * stored in a PRON attribute of the WORD element, but any existing
 * PRON atribute is not over-written.  This allows the user to
 * override dictionary pronunciations.  An exceptions dictionary
 * can also be specified.  This is just a text file with two columns
 * separated by a TAB: the spelling and the SAMPA transcription.
 *
 */
 
#define PROGNAME "uttdict2"
#define PROGVERS "2.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"
#include "lookup.h"

#ifdef SVR4
#define _stricmp strcasecmp
#endif

/* input file */
char		filename[256];	/* input file (from uttmake) */
char		ofilename[256];	/* output filename */
char		tfilename[256];	/* temporary output filename */

/* exceptions file */
char		xfilename[256];	/* exceptions dictionary */
struct except_rec {
	char	*wrd;
	char	*phn;
} *xtab;
int xcnt;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;
const char *	attrID;
const char *	attrPREF;	/* pronunciation reference */
const char *	attrSTRESS;	/* stress level */
const char *	attrPRON;

/* word sequence to process */
NSL_Item	*wordseqitem;

/* structure to hold words */
struct word_rec {
	char		*str;		/* spelling */
	int		npron;		/* requested pronunciation # */
	int		weak;		/* weak preferred */
	char		*phn;		/* SAMPA transcription */
	NSL_Item	*worditem;	/* referring item */
	struct word_rec	*next;		/* link */
};
struct word_rec *wroot;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* compare two exceptions dictionary entries */
int exceptcmp(const void *v1,const void * v2)
{
	struct except_rec *x1 = (struct except_rec *)v1;
	struct except_rec *x2 = (struct except_rec *)v2;
	return(strcmp(x1->wrd,x2->wrd));
}

/* load exceptions dictionary */
void loadexcept(char *xname)
{
	FILE	*ip;
	char	iline[1024];
	char	*p;

	if ((ip=fopen(xname,"r"))==NULL)
		error("could not open '%s'",xname);
	xcnt=0;
	while (fgets(iline,0124,ip)) xcnt++;
	rewind(ip);

	xtab = (struct except_rec *)calloc(xcnt,sizeof(struct except_rec));
	xcnt=0;
	while (fgets(iline,1024,ip)) {
		p = strtok(iline,"\t\r\n");
		if (p && *p) {
			xtab[xcnt].wrd = strsave(p);
			p = strtok(NULL,"\r\n");
			if (p && *p) {
				xtab[xcnt].phn = strsave(p);
				xcnt++;
			}
		}
	}
	fclose(ip);

	/* sort it for luck */
	qsort(xtab,xcnt,sizeof(struct except_rec),exceptcmp);
}

/* look up word in exceptions dictionary */
struct except_rec *findexcept(char *str)
{
	int	i,j,k,c;

	if (xcnt==0) return(NULL);

	/* look for exact match */
	i=0;
	j=xcnt-1;
	do {
		k = (i+j)/2;
		c = strcmp(str,xtab[k].wrd);
		if (c < 0)
			j = k - 1;
		else if (c > 0)
			i = k + 1;
		else
			return(&xtab[k]);
	} while (i <= j);

	/* look for case insensitive match */
	i=0;
	j=xcnt-1;
	do {
		k = (i+j)/2;
		c = _stricmp(str,xtab[k].wrd);
		if (c < 0)
			j = k - 1;
		else if (c > 0)
			i = k + 1;
		else
			return(&xtab[k]);
	} while (i <= j);

	/* fail */
	return(NULL);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* check for a word match */
int	matchword(char *str)
{
	struct word_rec *w=wroot;

	while (w) {
		if (strcmp(w->str,str)==0) return(1);
		w = w->next;
	}
	return(0);
}

/* find pronunciations for words */
void	findpron(int nword)
{
	struct word_rec 	*wr;
	struct except_rec	*xr;
	char	**w,**p;
	char	*pref[32];
	char	*pr,*lpref;
	int	npref;
	int	i,j,n;

	/* convert words into a table */
	w = (char **)calloc(nword,sizeof(char *));
	p = (char **)calloc(nword,sizeof(char *));
	wr = wroot;
	i=0;
	while (wr) {
		w[i] = wr->str;
		i++;
		wr=wr->next;
	}

	/* look up words in lexicon */
	lookup(w,nword,p,0);

	/* transfer pronunciations */
	wr = wroot;
	i=0;
	while (wr) {

		/* only process empty pronunciations */
		if (wr->phn==NULL) {

			/* check against exceptions dictionary */
			xr = findexcept(wr->str);
			if (xr!=NULL) {
				if (strcmp(wr->str,xr->wrd)==0) {
					/* exact match, go with this */
					free(p[i]);
					p[i]=strsave(xr->phn);
				}
				else if (_stricmp(wr->str,xr->wrd)==0) {
					/* only case insensitive match
					   go with it if forced to */
					if (strcmp(p[i],"")==0) {
						free(p[i]);
						p[i]=strsave(xr->phn);
					}
				}
			}

			/* split into pronunciation alternatives */
			pr = strtok(p[i],"\t");
			for (npref=0;(pr!=NULL)&&(*pr!='\0')&&(npref<32);npref++) {
				pref[npref] = pr;
				pr = strtok(NULL,"\t");
			}

			if (npref==0) {
				/* oh dear, none! */
				wr->phn=NULL;
			}
			else {
				/* take any specified counts into account if possible */
				n=wr->npron;

				/* are we looking for a weak one ? */
				lpref = pref[0];
				if (wr->weak) {
					for (j=0;(j<npref)&&(n>0);j++)
						if (strchr(pref[j],'\'')==NULL) {
							lpref = pref[j];
							n--;
						}
				}
				else {
					for (j=0;(j<npref)&&(n>0);j++) {
						lpref = pref[j];
						n--;
					}
				}
				wr->phn = strsave(lpref);
				if (wr->weak) {
					/* delete any stress marks in dictionary */
					for (pr=wr->phn,j=0;wr->phn[j];j++)
						if (wr->phn[j]!='\'')
							*pr++ = wr->phn[j];
					*pr = '\0';
				}
/* fprintf(stderr,"str='%s' phn='%s' npron=%d npref=%d\n",wr->str, wr->phn, wr->npron, npref); */
			}

			free(p[i]);
			p[i]=wr->phn;
		}
		i++;
		wr=wr->next;
	}
}

/* process word sequence */
void	process(NSL_Item *wordseqitem)
{
	struct word_rec *w,*wn;
	struct word_rec **wlast=&wroot;
	NSL_Item	*item;
	NSL_Data	*data;
	NSL_Data	*pdata;
	int		wordmiss=0;
	int		nword=0;

	const char *	pref;
	const char *	stress;
	const char *	pron;

	/* load words into temporary structure */
	for (data=wordseqitem->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORD")==0) {

				/* store information */
				w = (struct word_rec *)calloc(1,sizeof(struct word_rec));
				w->worditem = item;	/* pointer to actual data */
				pdata = item->data;
				if (pdata->type == NSL_text_data)
					w->str = strsave(item->data->first);
				else
					w->str = strsave("");
				pref = GetAttrStringVal(item,attrPREF);
				if (pref && *pref) {
					if (*pref=='#')
						w->npron = atoi(pref+1);
					else
						w->npron = atoi(pref);
					if (w->npron==0) w->npron=1;
				}
				else
					w->npron = 1;
		
				stress = GetAttrStringVal(item,attrSTRESS);
				if (stress && *stress) {
					if (*stress=='W')
						w->weak=1;
					else
						w->weak=0;
				}
				else
					w->weak=0;

				pron = GetAttrStringVal(item,attrPRON);
				if (pron && *pron)
					w->phn = strsave(pron);
				else
					w->phn = NULL;

				/* add to end of chain */
				*wlast = w;
				wlast = &w->next;
				nword++;
			}
		}
	}

	/* match words against lexicon */
	findpron(nword);

	/* check all words accounted for */
	w = wroot;
	while (w) {
		if (!w->phn) {
			if (wordmiss==0) fprintf(stderr,"Missing words:\n");
			fprintf(stderr,"%s\n",w->str);
			wordmiss=1;
		}
		w = w->next;
	}

	/* add pronunciations */
	w = wroot;
	while (w) {
		PutAttrVal(w->worditem,attrPRON,strsave(w->phn));
		w=w->next;
	}
	
	/* free word rec structures */
	w = wroot;
	while (w) {
		wn = w->next;
		free(w->str);
		if (w->phn) free(w->phn);
		free(w);
		w = wn;
	}
	wroot=NULL;

	if (wordmiss) exit(1);
}

/* recursively search for and process WORDSEQ items */
void findprocess(NSL_Item *item)
{
	NSL_Data	*data;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"WORDSEQ")==0)
				process(item);
			else
				findprocess(item);
		}
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	dinf;
	NSL_File 	inf;
	NSL_Doctype 	idct=NULL;

	NSL_Item 	*uttitem;
	NSL_Bit		*bit;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ix:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: pronunciation lookup Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'x':	/* exceptions dictionary */
		strcpy(xfilename,optarg);
		loadexcept(xfilename);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-x exceptions.dic) (input.xml) (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((dinf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(dinf);
	SFclose(dinf);
	
	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, odct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else {
		strcpy(filename,"STDIN");
		if ((inf = SFFopen(stdin, odct, NSL_read,filename))==NULL)
			error("unable to open '%s'",filename);
	}

	/* get output attribute names */
	attrID = AttrUniqueName(odct,"ID",0);
	attrPREF = AttrUniqueName(odct,"PREF",0);
	attrSTRESS = AttrUniqueName(odct,"STRESS",0);
	attrPRON = AttrUniqueName(odct,"PRON",0);

	/* Open the output stream using the output doctype declaration */
	optind++;
	if (optind < argc) {
		strcpy(ofilename,argv[optind]);
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
		outf = SFopen(tfilename, odct, NSL_write_pretty);
	}
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* read first tag - should be <UTT> or <PROSYNTH> */
	while ((bit=GetNextBit(inf))!=NULL) {
		if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
			error("Could not read <UTT> or <PROSYNTH> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		/* find and process WORDSEQ items */
		findprocess(uttitem);

		/* print the utterance with modified word sequence */
		PrintItem(outf,uttitem);
		ForceNewline(outf);

	}

	/* close input file */
	SFclose(inf);

	/* that's all folks */
	SFclose(outf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
