/* uttfind.c -- find components of XML structures by tree match */

/* Mark Huckvale - University College London */

/* version 1.0 - July 1998 */
/* version 1.1 - August 1998
	- add '.' as alternative for space as indentation
	- add '!=' operator for feature values
*/
/* version 1.2 - April 1999
	- add switch to invert match
*/

#define PROGNAME "uttfind"
#define PROGVERS "1.2"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"
#include "uttfind.h"

/* tree of searches */
char	sfilename[256];
struct search_rec *sroot;
int	nfound=0;
int	doinvert=0;

/* instantiated parameters */
#define MAXPARAM 10
char	*param[MAXPARAM+1];

/* node stack */
struct nmatch_rec *nrstack[100];
int	nrsp=0;

/* ID attribute */
const char *attrID;

/* push and pop nodes */
void pushnode(struct nmatch_rec *nr)
{
	nrstack[nrsp++] = nr;
}
struct nmatch_rec *popnode()
{
	return(nrstack[--nrsp]);
}

/* new search */
struct search_rec *newsearch()
{
	struct search_rec *p=calloc(1,sizeof(struct search_rec));
	if (p==NULL)
		error("out of memory");
	return(p);
}

/* new node match */
struct nmatch_rec *newnmatch()
{
	struct nmatch_rec *p=calloc(1,sizeof(struct nmatch_rec));
	if (p==NULL)
		error("out of memory");
	return(p);
}

/* new attrribute match */
struct amatch_rec *newamatch()
{
	struct amatch_rec *p=calloc(1,sizeof(struct amatch_rec));
	if (p==NULL)
		error("out of memory");
	return(p);
}

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory\n");
	strcpy(ptr,str);
	return(ptr);
}

/* decode a line */
struct nmatch_rec *decode(char *p)
{
	struct nmatch_rec *this = newnmatch();
	struct amatch_rec *attr;
#define BUFSIZE	128
	char		buf[BUFSIZE];
	int		i;
	
	/* possible '*' */
	if (*p=='*') {
		this->flags |= NODE_REPORT;
		p++;
	}
	/* node name or possible (.) */
	if (strncmp(p,"(.)",3)==0) {
		this->flags |= NODE_RECURSE;
		p+=3;
	}
	else {
		for (i=0;(i<BUFSIZE) && (*p!='\0') && (*p!=' ') && (*p!='(');i++,p++) buf[i]=*p;
		buf[i]='\0';
		this->name = strsave(buf);
	}
	while (isspace(*p)) p++;
	/* possible attribute list */
	if (*p=='(') {
		p++;
		while (isspace(*p)) p++;
		while (*p && (*p!=')')) {
			/* start a new attribute match */
			attr = newamatch();
			/* get name of attribute */
			for (i=0;(i<BUFSIZE) && (*p!='\0') && (*p!=' ') && (*p!='=') && (*p!='!') && (*p!=')');i++,p++) buf[i]=*p;
			buf[i]='\0';
			attr->name = strsave(buf);
			if (*p=='=') {
				p++;
				if (*p=='=') p++; /* also allow '==' for equals */
				for (i=0;(i<BUFSIZE) && (*p!='\0') && (*p!=' ') && (*p!=')');i++,p++) buf[i]=*p;
				buf[i]='\0';
				attr->val = strsave(buf);
				attr->matchop = ATTR_MATCH_EQ;
			}
			else if ((p[0]=='!') && (p[1]=='=')) {
				p+=2;
				for (i=0;(i<BUFSIZE) && (*p!='\0') && (*p!=' ') && (*p!=')');i++,p++) buf[i]=*p;
				buf[i]='\0';
				attr->val = strsave(buf);
				attr->matchop = ATTR_MATCH_NE;
			}
			else
				error("bad attribute comparison operator: '%c'",*p);
			while (isspace(*p)) p++;
			attr->next = this->afirst;
			this->afirst = attr;
		}
		if (*p==')') p++;
	}
	while (isspace(*p)) p++;
	/* possible text match */
	if (*p=='/') {
		this->tmatch = strsave(strtok(p+1,"/"));
	}

	return(this);
}

/* parse queries */
void parsequery(FILE *ip)
{
	char	iline[1024];
	int	lineno=0;
	char	*p;
	int	level=0;
	int	tlevel;
	struct search_rec *snew;
	struct nmatch_rec *this=NULL;
	struct nmatch_rec *temp=NULL;

	while (fgets(iline,1024,ip)) {
	    lineno++;
	    if ((iline[0]!='\n') && (iline[0]!='#')) {
		p = strtok(iline,"\n\r");
		if (*p==' ')
			for (tlevel=0;isspace(*p);tlevel++) p++;
		else
			for (tlevel=0;*p=='.';tlevel++) p++;
		if (tlevel > level+1)
			error("%s (%d) level jump too big",sfilename,lineno);
		if (tlevel==0) {
			/* new search */
			if (this!=NULL) {
				/* clear last search */
				while (level > 0) {
					this = popnode();
					level--;
				}
				/* add to head of search chain */
				snew = newsearch();
				snew->first = this;
				snew->next = sroot;
				sroot = snew;
			}
			this = NULL;
		}
		/* get line */
		temp = decode(p);
		/* add in to right place */
		if (tlevel < level) {
			/* back up to right level */
			while (level > tlevel) {
				this = popnode();
				level--;
			}
			/* add in sequence at this level */
			this->next=temp;
			this = temp;
		}
		else if (tlevel > level) {
			/* down one level max */
			if (this==NULL)
				error("%s: (%d) push empty node",sfilename,lineno);
			this->first = temp;
			pushnode(this);
			this = temp;
		}
		else {
			/* same level add in sequence */
			if (this) {
				this->next = temp;
			}
			this = temp;
		}
		level = tlevel;
	    }
	}
	if (this!=NULL) {
		/* clear last search */
		while (level > 0) {
			this = popnode();
			level--;
		}
		/* add to head of search chain */
		snew = newsearch();
		snew->first = this;
		snew->next = sroot;
		sroot = snew;
	}
}

/* dump node */
void dumpnode(struct nmatch_rec *nr,int level)
{
	int	i;
	struct amatch_rec *ar;

	for (i=0;i<level;i++) printf(".");
	if (nr->flags & NODE_REPORT) printf("*");
	if (nr->flags & NODE_RECURSE)
		printf("(.)");
	else
		printf("%s",nr->name);
	if (nr->afirst) {
		printf("(");
		ar = nr->afirst;
		while (ar) {
			if (ar->matchop==ATTR_MATCH_EQ)
				printf("%s=%s",ar->name,ar->val);
			else
				printf("%s!=%s",ar->name,ar->val);
			ar = ar->next;
			if (ar) putchar(' ');
		}
		printf(")");
	}
	if (nr->tmatch)
		printf(" /%s/",nr->tmatch);
	printf("\n");
	if (nr->first)
		dumpnode(nr->first,level+1);
	if (nr->next)
		dumpnode(nr->next,level);
}

/* dump query */
void dumpquery()
{
	int	i=0;
	struct search_rec *sr;

	sr = sroot;
	while (sr) {
		printf("Query %d:\n",++i);
		dumpnode(sr->first,0);
		sr = sr->next;
	}
}

/* look up node IDs */
void lookupnodeid(NSL_Doctype dct,struct nmatch_rec *nr)
{
	struct amatch_rec *ar;

	if (nr->name)
		nr->id = ElementUniqueName(dct,nr->name,0);
	ar = nr->afirst;
	while (ar) {
		ar->id = AttrUniqueName(dct,ar->name,0);
		ar = ar->next;
	}
	if (nr->first)
		lookupnodeid(dct,nr->first);
	if (nr->next)
		lookupnodeid(dct,nr->next);
}

/* look up query IDs */
void lookupids(NSL_Doctype dct)
{
	struct search_rec *sr;

	sr = sroot;
	while (sr) {
		lookupnodeid(dct,sr->first);
		sr = sr->next;
	}

	attrID = AttrUniqueName(dct,"ID",0);
}

NSL_Item *found;

int checknode(struct nmatch_rec *nr,NSL_Item *item)
{
	const char *val;
	struct amatch_rec *ar;
	NSL_Data *data;
	int	i;

	/* check top level match for label */
	if (nr->id && (nr->id != item->label)) {
#ifdef EBUG
printf("checknode(%s,%s)=>0\n",nr->name,item->label);
#endif
		return(0);
	}

	/* check top level match for attributes */
	ar = nr->afirst;
	while (ar) {
		val = GetAttrStringVal(item,ar->id);
#ifdef EBUG
printf("checkparam(%s,%s,%s)\n",ar->id,ar->val,val);
#endif
		if (ar->val[0]=='$') {
			i = atoi(ar->val+1);
			if ((i<0) || (i>MAXPARAM))
				error("$n parameter too large");
			if (param[i]==NULL)
				param[i]=strsave(val);
			else if ((ar->matchop==ATTR_MATCH_EQ) && (strcmp(param[i],val)!=0))
				return(0);
			else if ((ar->matchop==ATTR_MATCH_NE) && (strcmp(param[i],val)==0))
				return(0);
		}
		else if (!val) {
			return(0);
		}
		else if ((ar->matchop==ATTR_MATCH_EQ) && (strcmp(val,ar->val)!=0)) {
			return(0);
		}
		else if ((ar->matchop==ATTR_MATCH_NE) && (strcmp(val,ar->val)==0)) {
			return(0);
		}
		ar = ar->next;
	}

	/* check any text match */
	if (nr->tmatch) {
#ifdef EBUG
printf("checking text\n");
#endif
		data = item->data;
		if (!data) return(0);
		if (data->type!=NSL_text_data) return(0);
#ifdef EBUG
printf("checktext(%s,%s)\n",nr->tmatch,data->first);
#endif
		if (strcmp(nr->tmatch,data->first)) return(0);
	}
	
	/* check what to return */
	if (nr->flags & NODE_REPORT) found=item;
	else if (found==NULL) found=item;

#ifdef EBUG
printf("checknode(%s,%s)=>1\n",nr->name,item->label);
#endif

	return(1);
}

/* match an item against single query */
int matchnode(struct nmatch_rec *nr,NSL_Data *data)
{
	NSL_Item *item;


	if (nr==NULL) return(1);
	if (data==NULL) return(0);
	if (data->type!=NSL_item_data) return(0);
	item = data->first;
#ifdef EBUG
printf("match %s:%s\n",nr->name,item->label);
#endif
	if (checknode(nr,item)) {
#ifdef EBUG
printf("Node OK\n");
#endif
		if (nr->first) {
			/* this node OK, also check children */
			if (matchnode(nr->first,item->data)) {
#ifdef EBUG
printf("Children OK\n");
#endif
				/* children OK, check next query v. next node */
				if (!nr->next) return(1);
				if (matchnode(nr->next,data->next)) {
#ifdef EBUG
printf("Siblings OK\n");
#endif
					return(1);
				}
			}
		}
		else {
			/* children OK, check next query v. next node */
			if (!nr->next) return(1);
			if (matchnode(nr->next,data->next)) {
#ifdef EBUG
printf("Siblings OK\n");
#endif
				return(1);
			}
		}
	}

	/* are we allowed to recurse ? */
	if (nr->flags & NODE_RECURSE) {
#ifdef EBUG
printf("Drilling down\n");
#endif
		if (matchnode(nr,item->data))
			return(1);
	}

	/* back up, this query v. next node */
#ifdef EBUG
printf("Backing up\n");
#endif
	return matchnode(nr,data->next);
}

/* clear parameters */
void clearparams()
{
	int	i;
	for (i=0;i<=MAXPARAM;i++) {
		if (param[i]) {
			free(param[i]);
			param[i]=NULL;
		}
	}
}

/* match an item against query list */
NSL_Item *matchquery(NSL_Item *item)
{
	struct search_rec *sr;
	int		match;

	sr = sroot;
	while (sr) {
		clearparams();
		if (checknode(sr->first,item)) {
			match = matchnode(sr->first->first,item->data);
			if (match) return(found);
		}
		sr = sr->next;
	}
	return(NULL);
}

/* scan a source file */
void scan(NSL_File inf,NSL_File outf,NSL_Doctype dct)
{
	NSL_Bit		*bit;
	NSL_Item	*item;
	NSL_Item	*match;
	NSL_Item	*wrapper;
	char		buf[128];
	const char	*p;

	while ((bit=GetNextBit(inf))!=NULL) {
		if (bit->type!=NSL_start_bit)
			error("file is not XML");

		/* load the whole thing into memory */
		item = ItemParse(inf,bit->value.item);

		/* compare with query */
		match = matchquery(item);

		/* print if found */
		if (doinvert) {
			if (!match) {
				PrintItem(outf,item);
				ForceNewline(outf);
				nfound++;
			}
		}
		else if (match) {
			if (match==item)
				PrintItem(outf,match);
			else {
				sprintf(buf,"<%s></%s>",bit->label,bit->label);
				wrapper = GetItemFromString(buf,dct);
				p = GetIdVal(item);
				if (p==NULL)
					p = GetAttrStringVal(item,attrID);
				if ((attrID!=NULL) && (p!=NULL))
					PutAttrVal(wrapper,attrID,p);
				AddItemToEnd(wrapper,match);
				PrintItem(outf,wrapper);
			}
			ForceNewline(outf);
			nfound++;
		}

		FreeItem(item);
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	int		debug=0;
	FILE		*ip;

	char		filename[256];
	NSL_File 	inf;
	NSL_File 	outf=NULL;
	NSL_Doctype 	dct=NULL;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"IDn"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: find components by tree query Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'D':
		debug=1;
		break;
	case 'n':
		doinvert=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-D) (-n) search.txt (*.xml)",PROGNAME);

	/* parse search text */
	if (optind < argc)
		strcpy(sfilename,argv[optind++]);
	else
		error("no search pattern file specified");

	if ((ip=fopen(sfilename,"r"))==NULL)
		error("could not open '%s'",sfilename);

	parsequery(ip);
	fclose(ip);

	if (debug) dumpquery();

	/* scan input files */
	if (argc==optind) {
		inf = SFFopen(stdin, dct, NSL_read, "stdin");
		dct = DoctypeFromFile(inf);
		/* open output file */
		if (outf==NULL)
			outf = SFFopen(stdout,dct,NSL_write,"stdout");
		lookupids(dct);
		scan(inf,outf,dct);
	}
	else for (;optind<argc;optind++) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, dct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
		dct = DoctypeFromFile(inf);
		/* open output file */
		if (outf==NULL)
			outf = SFFopen(stdout,dct,NSL_write,"stdout");
		lookupids(dct);
		scan(inf,outf,dct);
		SFclose(inf);
	}

	fprintf(stderr,"Found %d matches.\n",nfound);
	exit(0);
}
