/* uttfx -- transfer Phonetic Fx onto utterance */

/* Mark Huckvale - University College London */

/* version 1.0 - August 1998 */

/*
 * uttfx takes the aligned syllable structure from an
 * utterance file and models the Fx contour for a segment
 * as a straight line, recording the gradient and mid-point
 * as features FXGRD & FXMID on the utterance
 */
 
#define PROGNAME "uttfx"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include "SFSCONFG.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "sfs.h"
#include "lt-memory.h"
#include "nsl.h"

/* input file */
char		filename[256];	/* input file (from uttcomp) */
const char	*attrWREF;
const char	*attrID;
const char	*attrSTART;
const char	*attrSTOP;
const char	*attrFXGRD;
const char	*attrFXMID;
const char	*attrVOI;

/* SFS file */
char		sfilename[256];
struct item_header	fxitem;
short		*fx;

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;

/* phone records */
struct phone_rec {
	NSL_Item *item;
	int	voiced;		/* voiced flag */
	double	stime;		/* starting time */
	double	etime;		/* ending time */
	double	sdur;		/* segmental duration */
	double	fxgrd;		/* Fx gradient */
	double	fxmid;		/* Fx at mid point */
};

struct phone_rec *uphone;	/* utterance phones */
int	ucount;

/* section of FX contour */
#define MAXLENGTH	1000

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Data *data,const char *name)
{
	NSL_Item *item;
	for (;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
		}
	}
	return(NULL);
}

/* count number of CNS and VOC units */
int countphone(NSL_Item *item,int count)
{
	NSL_Data *data;

	if (!item) return(count);

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if ((strcmp(item->label,"VOC")==0) ||
			    (strcmp(item->label,"CNS")==0)) count++;
			else
				count = countphone(item,count);
		}
	}
	return(count);
}

/* safe version of atof() */
double myatof(const char *str)
{
	if (str && *str && isdigit(*str))
		return(atof(str));
	else {
		fprintf(stderr,"?atof(%s)\n",(str)?str:"<null>");
		return(0.0);
	}
}

/* safe version of strcmp */
int mystrcmp(const char *s,const char *t)
{
	if (s && t)
		return(strcmp(s,t));
	else if (s || t)
		return(1);
	else
		return(0);
}

/* copy over CNS and VOC units */
void copyphone(NSL_Item *item)
{
	NSL_Data *data;

	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if ((strcmp(item->label,"VOC")==0) ||
			    (strcmp(item->label,"CNS")==0)) {
			    	uphone[ucount].item = item;
		    		uphone[ucount].stime = myatof(GetAttrStringVal(item,attrSTART));
		    		uphone[ucount].etime = myatof(GetAttrStringVal(item,attrSTOP));
		    		uphone[ucount].sdur = uphone[ucount].etime - uphone[ucount].stime;
				if (strcmp(item->label,"VOC")==0)
					uphone[ucount].voiced=1;
				else if (mystrcmp(GetAttrStringVal(item,attrVOI),"Y")==0)
					uphone[ucount].voiced=1;
				else
					uphone[ucount].voiced=0;
			    	ucount++;
			}
			else
				copyphone(item);
		}
	}
}

/* fx modelling */
void fxmodel()
{
	double	x[MAXLENGTH];
	double	y[MAXLENGTH];
	int	n;
	int	i,j;
	double	t;
	double	xsum,ysum,xsum2,xysum;

	/* take each segment in turn */
	for (i=0;i<ucount;i++) {
		/* only process phonetically voiced segments */
		if (uphone[i].voiced==0) {
			uphone[i].fxmid=0;
			uphone[i].fxgrd=0;
			continue;
		}
		/* copy fx values into array */
		for (n=0,t=uphone[i].stime;(n<MAXLENGTH)&&(t<uphone[i].etime);t+=fxitem.frameduration) {
			j = (int)(0.5+(t-fxitem.offset)/fxitem.frameduration);
			if ((0 <= j) && (j < fxitem.numframes)) {
				x[n] = t - uphone[i].stime;
				y[n] = fx[j];
				if (y[n] > 0) n++;	/* only use non-zero values */
			}
		}
		if (n==0) {
			uphone[i].fxmid=0;
			uphone[i].fxgrd=0;
		}
		else if (n==1) {
			uphone[i].fxmid=y[0];
			uphone[i].fxgrd=0;
		}
		else {
			/* least squares fit */
			xsum=0;
			ysum=0;
			xsum2=0;
			xysum=0;
			for (j=0;j<n;j++) {
				xsum += x[j];
				ysum += y[j];
				xsum2 += x[j]*x[j];
				xysum += x[j]*y[j];
			}
			uphone[i].fxgrd = (n*xysum - xsum*ysum)/(n*xsum2 - xsum*xsum);
			uphone[i].fxmid = uphone[i].fxgrd * (uphone[i].etime-uphone[i].stime)/2 + (ysum - uphone[i].fxgrd*xsum)/n;
		}
	}
}

/* process an utterance */
void process(NSL_Item *ipitem)
{
	int		i;
	int		ulen;
	char		buf[32];
	
	/* load the segments into temporary structure */
	ulen = countphone(ipitem,0);

	/* get memory */
	uphone = (struct phone_rec *)calloc(ulen,sizeof(struct phone_rec));
	if (!uphone)
		error("out of memory");

	/* copy over the utterance phones */
	ucount=0;
	copyphone(ipitem);

	/* do FX modelling */
	fxmodel();

	/* store fx model values on segments */
	for (i=0;i<ucount;i++) {
		sprintf(buf,"%.1f",uphone[i].fxgrd);
		PutAttrVal(uphone[i].item,attrFXGRD,strsave(buf));
		sprintf(buf,"%.1f",uphone[i].fxmid);
		PutAttrVal(uphone[i].item,attrFXMID,strsave(buf));
	}

}

/* join fx contour across segments */
short joinfx(double fx1,double grd1,	/* starting fx and gradient */
	     double dur1,		/* duration of first half */
	     double fx2,double grd2,	/* ending fx and gradient */
	     double dur2,		/* duration of second half */
	     double t)			/* duration and time */
{
	double	a,b,d;

#if 0
printf("fx1=%5.1f grd1=%5.1f dur1=%5.3f fx2=%5.1f grd2=%5.1f dur2=%5.3f t=%5.3f\n",
	fx1,grd1,dur1,fx2,grd2,dur2,t);
#endif

	if (fx1==0) {
		/* starting linear */
		return((short)(fx2-grd2*(dur1+dur2-t)));
	}
	else if (fx2==0) {
		/* ending linear */
		return((short)(fx1+grd1*t));
	}
	else {
		/* fit cubic */
		d = dur1+dur2;
		a = (grd2*d-2*fx2+grd1*d+2*fx1)/(d*d*d);
		b = (grd2-grd1-3*a*d*d)/(2*d);
		return((short)(a*t*t*t + b*t*t + grd1*t + fx1));
	}
}

/* save modelled Fx back into SFS file */
void fxsave(char *fname)
{
	struct item_header opitem;
	int	i,j;
	double	t,dur1,dur2;
	
	sfsheader(&opitem,FX_TYPE,0,2,1,fxitem.frameduration,fxitem.offset,1,0,0);
	sprintf(opitem.history,"%s(%d.%02d;utt=%s)",
		PROGNAME,
		fxitem.datatype,fxitem.subtype,
		filename);

	/* overwrite existing Fx buffer */
	for (i=0;i<fxitem.numframes;i++) fx[i]=0;
#ifdef LINEARFX
	for (i=0;i<ucount;i++) {
		startfx = uphone[i].fxmid - uphone[i].fxgrd*(uphone[i].etime-uphone[i].stime)/2;
		for (t=uphone[i].stime;t<uphone[i].etime;t+=fxitem.frameduration) {
			j = (int)(0.5+(t-fxitem.offset)/fxitem.frameduration);
			if ((0 <= j) && (j < fxitem.numframes)) {
				fx[j] = (short)(startfx + uphone[i].fxgrd * (t - uphone[i].stime));
			}
		}
	}
#else
	for (i=0;i<ucount;i++) if (uphone[i].voiced) {
		/* do first half */
		if (i==0)
			dur1 = 0;
		else
			dur1 = uphone[i-1].sdur/2;
		dur2 = uphone[i].sdur/2;
		for (t=0;t<dur2;t+=fxitem.frameduration) {
			j = (int)(0.5+(uphone[i].stime+t-fxitem.offset)/fxitem.frameduration);
			if ((0 <= j) && (j < fxitem.numframes)) {
				if (i==0)
					fx[j] = joinfx(0.0,0.0,0.0,
						uphone[i].fxmid,uphone[i].fxgrd,dur2,t);
				else
					fx[j] = joinfx(uphone[i-1].fxmid,uphone[i-1].fxgrd,dur1,
						uphone[i].fxmid,uphone[i].fxgrd,dur2,dur1+t);
			}
		}
		/* do second half */
		dur1 = uphone[i].sdur/2;
		if (i==ucount-1)
			dur2 = 0;
		else
			dur2 = uphone[i+1].sdur/2;
		for (t=0;t<dur1;t+=fxitem.frameduration) {
			j = (int)(0.5+(uphone[i].stime+dur1+t-fxitem.offset)/fxitem.frameduration);
			if ((0 <= j) && (j < fxitem.numframes)) {
				if (i==ucount-1)
					fx[j] = joinfx(uphone[i].fxmid,uphone[i].fxgrd,dur1,
						0.0,0.0,0.0,t);
				else
					fx[j] = joinfx(uphone[i].fxmid,uphone[i].fxgrd,dur1,
						uphone[i+1].fxmid,uphone[i+1].fxgrd,dur2,t);
			}
		}
	}
#endif
	putitem(fname,&opitem,fxitem.numframes,fx);
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p;
	
	NSL_File 	inf;
	NSL_Item 	*uttitem;
	NSL_Item 	*ipitem;
	NSL_Bit		*bit;

	int		it;
	char		*ty;
	char		*fxtype="0";
	int		savefx=0;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ii:s:o"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance alignment to SFS annotation Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'i' :	/* input specification */
		if (itspec(optarg,&it,&ty)==0) {
			if (it==FX_TYPE)
				fxtype=ty;
			else
				error("unsuitable item specification: '%s'",optarg);
		}
		else
			error("illegal item specification: '%s'",optarg);
		break;
	case 's':
		strcpy(sfilename,optarg);
		break;
	case 'o':
		/* save modelled fx contour */
		savefx=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-i item) (-o) -s input.sfs input.xml (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((inf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(inf);
	SFclose(inf);

	/* get unique names for attributes */
	attrID = AttrUniqueName(odct,"ID",0);
	attrWREF = AttrUniqueName(odct,"WREF",0);
	attrSTART = AttrUniqueName(odct,"START",0);
	attrSTOP = AttrUniqueName(odct,"STOP",0);
	attrFXMID = AttrUniqueName(odct,"FXMID",0);
	attrFXGRD = AttrUniqueName(odct,"FXGRD",0);
	attrVOI = AttrUniqueName(odct,"VOI",0);

	/* open the SFS file */
	if (!sfilename[0])
		error("no SFS file given");
	getitem(sfilename,FX_TYPE,fxtype,&fxitem,(void *)&fx);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, odct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else
		error("no input file specified");

	/* read first tag - should be <UTT> or <PROSYNTH> */
	if ((bit=GetNextBit(inf))==NULL)
		error("Could not read input file");

	if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
		error("Could not read <UTT> or <PROSYNTH> start tag");

	/* load the whole thing into memory */
	uttitem = ItemParse(inf,bit->value.item);

	/* find the IP item */
	ipitem = MyObtainItem(uttitem->data,"IP");

	/* check */
	if (!ipitem)
		error("No IP item in input");

	/* OK, let's go */
	process(ipitem);

	/* save modelled Fx if required */
	if (savefx) fxsave(sfilename);
	SFclose(inf);

	/* Open the output NSGML stream using the doctype declaration */
	optind++;
	if (optind < argc)
		outf = SFopen(argv[optind], odct, NSL_write_pretty);
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* print the utterance with modified word sequence */
	PrintItem(outf,uttitem);

	/* that's all folks */
	SFclose(outf);
	return 0;
}
