/* uttfxdat -- measure utterance structure and Fx contour */

/* Mark Huckvale - University College London */

/* version 1.0 - February 1999 */
/* version 1.1 - February 1999
	- stats package output option
	- change peak Fx threshold to 6%
*/

/*
 * uttfxdat takes the aligned syllable structure from an
 * utterance file and measures times and Fx contour from the
 * associated SFS file
 */
 
#define PROGNAME "uttfxdat"
#define PROGVERS "1.1"
char *progname=PROGNAME;

#include "SFSCONFG.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "sfs.h"
#include "dig.h"
#include "digdata.h"
#include "digitem.h"
#include "lt-memory.h"
#include "nsl.h"

#define FXTHRESH	0.96	/* range of -4% valid at peak */

/* input file */
char		filename[256];	/* input file (from uttcomp) */
int		formstat=0;	/* stats output format */
char		transcription[1024];
const char	*attrHREF;
const char	*attrID;
const char	*attrSTART;
const char	*attrSTOP;
const char	*attrVOI;
const char	*attrSON;

/* SFS file */
char		*sfilename;
struct item_header	txitem;
int32		*tx;
struct item_header	fxitem;
short		*fx;
int		lofx,hifx;

/* tokens and times */
struct token_rec {
	char	*str;		/* name */
	double	stime;		/* starting time */
	double	etime;		/* ending time */
};

/* foot records */
struct token_rec *flist;	/* utterance feet */
int	fcount;

/* syl records */
struct token_rec *slist;	/* utterance syllables */
int	scount;

/* VOC records */
struct token_rec *vlist;	/* utterance vowels */
int	vcount;

/* CNS records */
struct token_rec *clist;	/* utterance consonants */
int	ccount;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* safe version of atof() */
double myatof(const char *str)
{
	if (str && *str && isdigit(*str))
		return(atof(str));
	else {
		return(0.0);
	}
}

/* safe version of strcmp */
int mystrcmp(const char *s,const char *t)
{
	if (s && t)
		return(strcmp(s,t));
	else if (s || t)
		return(1);
	else
		return(0);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Item *item,const char *name)
{
	NSL_Data	*data;
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
			else if ((item=MyObtainItem(item,name))!=NULL)
				return(item);
		}
	}
	return(NULL);
}

/* return last named item below unit */
NSL_Item *MyObtainLastItem(NSL_Item *item,const char *name)
{
	NSL_Item *last=NULL;
	NSL_Data *data;
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				last = item;
			else if ((item=MyObtainLastItem(item,name))!=NULL)
				last = item;
		}
	}
	return(last);
}

/* find first voiced CNS */
NSL_Item *FirstVoiced(NSL_Data *data)
{
	NSL_Item *item;
	for (;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"CNS")==0)
				if (mystrcmp(GetAttrStringVal(item,attrVOI),"Y")==0)
					return(item);
		}
	}
	return(NULL);
}

/* return last voiced consonant */
NSL_Item *LastVoiced(NSL_Data *data)
{
	NSL_Item *last=NULL;
	NSL_Item *item;
	for (;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"CNS")==0)
				if (mystrcmp(GetAttrStringVal(item,attrVOI),"Y")==0)
					last = item;
		}
	}
	return(last);
}

double fxval(double t)
{
	int	idx = (int)(0.5+(t-fxitem.offset)/fxitem.frameduration);
	if ((0 <= idx) && (idx < fxitem.numframes))
		return((double)fx[idx]);
	else
		return((double)0);
}

double peaktime(double stime,double etime)
{
	int	idx;
	int	sidx= (int)(0.5+(stime-fxitem.offset)/fxitem.frameduration);
	int	eidx= (int)(0.5+(etime-fxitem.offset)/fxitem.frameduration);
	int	max=0;
	int	midx=0;
	double	t;

	for (idx=sidx;idx <= eidx;idx++) {
		if ((0 <= idx) && (idx < fxitem.numframes)) {
			if (fx[idx] > max) {
				max = fx[idx];
				midx = idx;
			}
		}
	}

	t = fxitem.offset+midx*fxitem.frameduration;
	if (t < stime) t += fxitem.frameduration;
	return(t);
}

double earlypeaktime(double stime,double ptime,double fxval)
{
	int	idx;
	int	sidx= (int)(0.5+(stime-fxitem.offset)/fxitem.frameduration);
	int	pidx= (int)(0.5+(ptime-fxitem.offset)/fxitem.frameduration);
	int	midx=pidx;
	double	t;

	for (idx=pidx;idx >= sidx;idx--) {
		if ((0 <= idx) && (idx < fxitem.numframes)) {
			if (fx[idx] > (int)(fxval*FXTHRESH)) {
				midx = idx;
			}
		}
	}

	t = fxitem.offset+midx*fxitem.frameduration;
	if (t < stime) t += fxitem.frameduration;
	return(t);
}

double latepeaktime(double ptime,double etime,double fxval)
{
	int	idx;
	int	pidx= (int)(0.5+(ptime-fxitem.offset)/fxitem.frameduration);
	int	eidx= (int)(0.5+(etime-fxitem.offset)/fxitem.frameduration);
	int	midx=pidx;
	double	t;

	for (idx=pidx;idx <= eidx;idx++) {
		if ((0 <= idx) && (idx < fxitem.numframes)) {
			if (fx[idx] > (int)(fxval*FXTHRESH)) {
				midx = idx;
			}
		}
	}

	t = fxitem.offset+midx*fxitem.frameduration;
	if (t < ptime) t += fxitem.frameduration;
	return(t);
}

/* find final fx value from average of last 50ms */
double finalfx(double stime,double etime)
{
	int	idx;
	int	sidx= (int)(0.5+(stime-fxitem.offset)/fxitem.frameduration);
	int	eidx= (int)(0.5+(etime-fxitem.offset)/fxitem.frameduration);
	int	midx=sidx;
	int	fxsum,fxcnt;

	/* find time of last valid fx value */
	for (idx=sidx;idx <= eidx;idx++) {
		if ((0 <= idx) && (idx < fxitem.numframes)) {
			if (fx[idx] > 0) {
				midx = idx;
			}
		}
	}

	/* jump back 50ms */
	sidx = midx - (int)(0.5+0.05/fxitem.frameduration);

	/* get mean */
	fxsum=0;
	fxcnt=0;
	for (idx=sidx;idx <= midx;idx++) {
		if ((0 <= idx) && (idx < fxitem.numframes)) {
			if (fx[idx] > 0) {
				fxsum += fx[idx];
				fxcnt++;
			}
		}
	}

	/* return mean */
	return((double)fxsum/fxcnt);
}

/* find time at which peak falls 75% to tail */
double earlytailtime(double ptime,double etime,double pkval,double tlval)
{
	int	idx;
	int	pidx= (int)(0.5+(ptime-fxitem.offset)/fxitem.frameduration);
	int	eidx= (int)(0.5+(etime-fxitem.offset)/fxitem.frameduration);
	int	midx=pidx;
	double	fxval = tlval + 0.25*(pkval-tlval);
	double	t;

	for (idx=pidx;idx <= eidx;idx++) {
		if ((0 <= idx) && (idx < fxitem.numframes) && (fx[idx]!=0)) {
			if (fx[idx] <= fxval) {
				midx = idx;
				break;
			}
		}
	}

	t = fxitem.offset+midx*fxitem.frameduration;
	if (t < ptime) t += fxitem.frameduration;
	return(t);
}

void transcribe(NSL_Item *item)
{
	NSL_Data *data;

	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_text_data)
			strcat(transcription,(char *)(data->first));
		else if (data->type == NSL_item_data)
			transcribe((NSL_Item *)(data->first));
	}
}

void reportfname(char *fname,double stime,double etime,char *tscript)
{
	if (formstat)
		printf("%s\t%s\t%g\t%g",fname,tscript,stime,etime);
	else {
		printf("# analysis of '%s'\n",fname);
		printf("# Foot contains %s\n",tscript);
		printf("# Foot extends from %.3f to %.3f\n",stime,etime);
		printf("# Label   Time FxVal\n");
	}
}
void reportnumber(char *name,double tval,double fval)
{
	if (formstat)
		printf("\t%g\t%g",tval,fval);
	else
		printf("%s:\t%6.3f\t%5.1f\n",name,tval,fval);
}
void reportstring(char *name,const char *str1,const char *str2)
{
	if (formstat)
		printf("\t%s\t%s",str1,str2);
	else
		printf("%s:\tVOI=%s\tSON=%s\n",name,str1,str2);
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	NSL_File 	inf;
	NSL_Doctype 	dct=NULL;
	NSL_Item 	*uttitem;
	NSL_Item	*fileitem;
	NSL_Item 	*ipitem;
	NSL_Item 	*agitem;
	NSL_Item 	*ftitem;
	NSL_Item 	*sylitem;
	NSL_Item	*onitem;
	NSL_Item	*cnsitem;
	NSL_Item	*rhitem;
	NSL_Item	*ncitem;
	NSL_Item	*cditem;
	NSL_Item	*acditem;
	NSL_Bit		*bit;

	int		it;
	char		*ty;
	char		*txtype="0";
	char		*fxtype="0";
	char		*antype="0";
	int		iptype=FX_TYPE;
	double		stime,etime,ptime,t;
	
	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ii:s"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance alignment to SFS annotation Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'i' :	/* input specification */
		if (itspec(optarg,&it,&ty)==0) {
			if (it==TX_TYPE) {
				iptype=TX_TYPE;
				txtype=ty;
			}
			else if (it==FX_TYPE) {
				iptype=FX_TYPE;
				fxtype=ty;
			}
			else
				error("unsuitable item specification: '%s'",optarg);
		}
		else
			error("illegal item specification: '%s'",optarg);
		break;
	case 's':
		formstat=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-i item) (-s) input.xml",PROGNAME);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, dct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else
		error("no input file specified");

	/* get unique names for attributes */
	dct = DoctypeFromFile(inf);
	attrID = AttrUniqueName(dct,"ID",0);
	attrHREF = AttrUniqueName(dct,"HREF",0);
	attrSTART = AttrUniqueName(dct,"START",0);
	attrSTOP = AttrUniqueName(dct,"STOP",0);
	attrVOI = AttrUniqueName(dct,"VOI",0);
	attrSON = AttrUniqueName(dct,"SON",0);

	/* read first tag - should be <UTT> */
	if ((bit=GetNextBit(inf))==NULL)
		error("Could not read input file");

	if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
		error("Could not read <UTT> or <PROSYNTH> start tag");

	/* load the whole thing into memory */
	uttitem = ItemParse(inf,bit->value.item);

	/* find name of SFS file */
	fileitem = MyObtainItem(uttitem,"FILE");
	sfilename = strsave(GetAttrStringVal(fileitem,attrHREF));

	/* open the SFS file and load fx contour */
	if (!sfilename || !sfilename[0])
		error("no SFS file given");
	getitem(sfilename,FX_TYPE,fxtype,&fxitem,(void *)&fx);

	/* find the last IP item */
	ipitem = MyObtainLastItem(uttitem,"IP");
	if (!ipitem)
		error("No IP item in input");

	/* find the last AG item */
	agitem = MyObtainLastItem(ipitem,"AG");
	if (!agitem)
		error("Missing AG item in input");
	
	/* find the first FOOT item */
	ftitem = MyObtainItem(agitem,"FOOT");
	if (!ftitem)
		error("Missing FOOT item in input");
	
	/* find the first syllable in foot */
	sylitem = MyObtainItem(ftitem,"SYL");
	if (!sylitem)
		error("Missing SYL item in input");

	/* find the onset of syllable */
	onitem = MyObtainItem(sylitem,"ONSET");

	/* find the rhyme of syllable */
	rhitem = MyObtainItem(sylitem,"RHYME");
	if (!rhitem)
		error("Missing RHYME item in input");

	/* find the nucleus of syllable */
	ncitem = MyObtainItem(rhitem,"NUC");
	if (!ncitem)
		error("Missing NUC item in input");

	/* find the coda of syllable */
	cditem = MyObtainItem(rhitem,"CODA");

	/* print details of what we've found */
	strcpy(transcription,"");
	transcribe(ftitem);
	reportfname(filename,
		stime=myatof(GetAttrStringVal(ftitem,attrSTART)),
		etime=myatof(GetAttrStringVal(ftitem,attrSTOP)),
		transcription);

	/* 1. print relative time of onset and fx value */
	t=stime;
	reportnumber("Onset",t-stime,fxval(t));

	/* 2. print relative time of nucleus and fx value */
	t=myatof(GetAttrStringVal(ncitem,attrSTART));
	reportnumber("Nuc",t-stime,fxval(t));

	/* 3. print relative time of coda and fx value */
	if (cditem)
		t=myatof(GetAttrStringVal(cditem,attrSTART));
	else
		t=myatof(GetAttrStringVal(ncitem,attrSTOP));
	reportnumber("Coda",t-stime,fxval(t));

	/* 4. print relative time of tail start and fx value */
	if (cditem)
		t=myatof(GetAttrStringVal(cditem,attrSTOP));
	else
		t=myatof(GetAttrStringVal(ncitem,attrSTOP));
	reportnumber("Tail",t-stime,fxval(t));

	/* 5. print relative time of end and fx value */
	t=myatof(GetAttrStringVal(ftitem,attrSTOP));
	reportnumber("Ends",t-stime,fxval(t));

	/* 6. print start of first voiced consonant in onset */
	if (onitem) {
		cnsitem = FirstVoiced(onitem->data);
		if (cnsitem) {
			t=myatof(GetAttrStringVal(cnsitem,attrSTART));
			reportnumber("VOT",t-stime,fxval(t));
		}
		else {
			t=myatof(GetAttrStringVal(ncitem,attrSTART));
			reportnumber("VOT",t-stime,fxval(t));
		}
	}
	else {
		t=myatof(GetAttrStringVal(sylitem,attrSTART));
		reportnumber("VOT",t-stime,fxval(t));
	}
	
	/* 7. print end of last voiced consonant in coda */
	if (cditem) {
		cnsitem = LastVoiced(cditem->data);
		if (cnsitem) {
			t=myatof(GetAttrStringVal(cnsitem,attrSTOP));
			reportnumber("VET",t-stime,fxval(t));
		}
		else {
			t=myatof(GetAttrStringVal(ncitem,attrSTOP));
			reportnumber("VET",t-stime,fxval(t));
		}
	}
	else {
		t=myatof(GetAttrStringVal(sylitem,attrSTOP));
		reportnumber("VET",t-stime,fxval(t));
	}
	
	/* 8. print relative time of peak and fx value */
	ptime=t=peaktime(stime,etime);
	reportnumber("Peak",t-stime,fxval(t));

	/* 9. print relative time of peak and fx value */
	t=earlypeaktime(stime,ptime,fxval(ptime));
	reportnumber("POT",t-stime,fxval(t));

	/* 10. print relative time of peak and fx value */
	t=latepeaktime(ptime,etime,fxval(ptime));
	reportnumber("PET",t-stime,fxval(t));

	/* 11. print tail onset time */
	t=earlytailtime(ptime,etime,fxval(ptime),finalfx(ptime,etime));
	reportnumber("TOT",t-stime,fxval(t));
	
	/* 12. print onset attributes */
	if (onitem) {
		cnsitem = MyObtainLastItem(onitem,"CNS");
		if (cnsitem) {
			reportstring("ONSET",
				GetAttrStringVal(cnsitem,attrVOI),
				GetAttrStringVal(cnsitem,attrSON));
		}
		else {
			reportstring("ONSET","N","N");
		}
	}
	else {
		reportstring("ONSET","N","N");
	}

	/* 13. print coda attributes */
	if (cditem) {
		cnsitem = MyObtainItem(cditem,"CNS");
		if (cnsitem) {
			reportstring("CODA",
				GetAttrStringVal(cnsitem,attrVOI),
				GetAttrStringVal(cnsitem,attrSON));
		}
		else {
			acditem = MyObtainItem(cditem,"ACODA");
			if (acditem) {
				cnsitem = MyObtainItem(acditem,"CNS");
				if (cnsitem) {
					reportstring("CODA",
						GetAttrStringVal(cnsitem,attrVOI),
						GetAttrStringVal(cnsitem,attrSON));
				}
				else
					reportstring("CODA","N","N");
			}
			else
				reportstring("CODA","N","N");
		}
	}
	else {
		reportstring("CODA","N","N");
	}
	if (formstat) printf("\n");

	/* that's all folks */
	SFclose(inf);
	return 0;
}
