/* uttfxds -- display utterance structure with Fx contour */

/* Mark Huckvale - University College London */

/* version 1.0 - October 1998 */
/* version 1.1 - November 1998
	- add option to display annotations
*/
/* version 1.2 - January 2000
	- add option to display spectrogram
*/

/*
 * uttfxds takes the aligned syllable structure from an
 * utterance file and displays the Fx contour from the
 * associated SFS file
 */
 
#define PROGNAME "uttfxds"
#define PROGVERS "1.2"
char *progname=PROGNAME;

#include "SFSCONFG.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "sfs.h"
#include "dig.h"
#include "digdata.h"
#include "digitem.h"
#include "lt-memory.h"
#include "nsl.h"

/* input file */
char		filename[256];	/* input file (from uttcomp) */
const char	*attrHREF;
const char	*attrID;
const char	*attrSTART;
const char	*attrSTOP;

/* SFS file */
char		*sfilename;
struct item_header	spitem;
short		*sp;
struct item_header	fxitem;
short		*fx;
int		lofx,hifx;
struct item_header	anitem;
struct an_rec	*an;
int		doan=0;
int		dosp=0;

extern int digit_xgrid[];
extern int digit_nxgrid;

/* tokens and times */
struct token_rec {
	char	*str;		/* name */
	double	stime;		/* starting time */
	double	etime;		/* ending time */
};

/* word records */
struct token_rec *wlist;	/* utterance words */
int	wcount;

/* ip records */
struct token_rec *ilist;	/* utterance intonation phrases */
int	icount;

/* ag records */
struct token_rec *alist;	/* utterance accents */
int	acount;

/* foot records */
struct token_rec *flist;	/* utterance feet */
int	fcount;

/* syl records */
struct token_rec *slist;	/* utterance syllables */
int	scount;

/* VOC records */
struct token_rec *vlist;	/* utterance vowels */
int	vcount;

/* CNS records */
struct token_rec *clist;	/* utterance consonants */
int	ccount;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Item *item,const char *name)
{
	NSL_Data	*data;
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
			else if ((item=MyObtainItem(item,name))!=NULL)
				return(item);
		}
	}
	return(NULL);
}

/* count number of named units */
int countunit(NSL_Item *item,char *name,int count)
{
	NSL_Data *data;

	if (!item) return(count);

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				count++;
			else
				count = countunit(item,name,count);
		}
	}
	return(count);
}

/* safe version of atof() */
double myatof(const char *str)
{
	if (str && *str && isdigit(*str))
		return(atof(str));
	else {
		fprintf(stderr,"?atof(%s)\n",(str)?str:"<null>");
		return(0.0);
	}
}

/* safe version of strcmp */
int mystrcmp(const char *s,const char *t)
{
	if (s && t)
		return(strcmp(s,t));
	else if (s || t)
		return(1);
	else
		return(0);
}

/* copy over specified unit */
void copyunit(NSL_Item *item,char *name,struct token_rec *list,int *count)
{
	NSL_Data *data;
	NSL_Data *subdata;

	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0) {
				subdata = item->data;
				if ((subdata!=NULL) && (subdata->type==NSL_text_data))
					list[*count].str = strsave(subdata->first);
				else
					list[*count].str = strsave(name);
				list[*count].stime = myatof(GetAttrStringVal(item,attrSTART));
				list[*count].etime = myatof(GetAttrStringVal(item,attrSTOP));
			    	(*count)++;
			}
			else
				copyunit(item,name,list,count);
		}
	}
}

/* draw one level */
void drawlevel(int bundles,float xl,float yb,float xr,float yt,double stime,double etime,struct token_rec *list,int count,char *label)
{
	int	i,iy;
	float	x0,x1,x2,x3,x,y;

	/* sort out bundles */
	digitbundle(bundles);

	/* init box */
	digitbox(xl,yb,xr,yt,stime,etime,DIGITEMBOX|DIGITEMOVERLAY);
	xl += digitab.ixoff/digdata.xscale;
	yt = yval(digitab.iyt);

	/* write axis label */
	iy = (digitab.iyt + digitab.iyb - digdata.chheight)/2;
	digtextp(digitab.lbun,digitab.ixl+3,iy,label);

	/* for each label */
	for (i=0;i<count;i++) {
		/* draw edges */
		x1 = (float)(xl + (xr-xl)*(list[i].stime-stime)/(etime-stime));
		digline(digitab.gbun,x1,yb,x1,yt);
		x2 = (float)(xl + (xr-xl)*(list[i].etime-stime)/(etime-stime));
		digline(digitab.gbun,x2,yb,x2,yt);
		/* check for ambisyllabicity */
		if (i > 0) {
			if ((list[i-1].stime < list[i].stime) && (list[i-1].etime > list[i].stime)) {
				x0 = (float)(xl + (xr-xl)*(list[i-1].etime-stime)/(etime-stime));
				digline(digitab.gbun,x1,(yb+yt)/2,x0,(yb+yt)/2);
				x1 = x0;
			}
		}
		if (i < count-1) {
			if ((list[i+1].etime > list[i].etime) && (list[i+1].stime < list[i].etime)) {
				x3 = (float)(xl + (xr-xl)*(list[i+1].stime-stime)/(etime-stime));
				digline(digitab.gbun,x2,(yb+yt)/2,x3,(yb+yt)/2);
				x2 = x3;
			}
		}
		/* draw label */
		x = (x1+x2)/2 - digtextlen(list[i].str)/2;
		y = (yb+yt)/2 - (digdata.chheight/digdata.yscale)/2;
		digtext(digitab.lbun,x,y,list[i].str);
	}

}

/* draw picture */
void draw()
{
	char	title[256];
	double	tsize;
	double	stime,etime;
	double	fxpos;
	int	i;
	
	/* find total picture duration */
	stime = ilist[0].stime;
	etime = ilist[icount-1].etime;

	/* open graphics */
	digstart(DIG_DEFAULT_TERM,NULL,1);

	/* display/redraw loop */
do {
	digscale((float)1.0,(float)1.0,0);

	/* get suitable basic size unit */
	tsize = 2 * digdata.chheight / digdata.yscale;

	digclearscreen();
	sprintf(title,"file=%s",filename);
	digtext(20,(float)0.0,(float)(1.0-2*tsize/3),title);

	/* display top timescale */
	digitemtime(20,(float)0.0,(float)(1.0-2*tsize),(float)1.0,(float)(1.0-tsize),stime,etime,1);

	/* display IP list */
	drawlevel(20,(float)0.0,(float)(1.0-3*tsize),(float)1.0,(float)(1.0-2*tsize),stime,etime,ilist,icount,"IP");

	/* display AG list */
	drawlevel(20,(float)0.0,(float)(1.0-4*tsize),(float)1.0,(float)(1.0-3*tsize),stime,etime,alist,acount,"AG");

	/* display FOOT list */
	drawlevel(20,(float)0.0,(float)(1.0-5*tsize),(float)1.0,(float)(1.0-4*tsize),stime,etime,flist,fcount,"FOOT");

	/* display SYL list */
	drawlevel(20,(float)0.0,(float)(1.0-6*tsize),(float)1.0,(float)(1.0-5*tsize),stime,etime,slist,scount,"SYL");

	/* display VOC list */
	drawlevel(20,(float)0.0,(float)(1.0-7*tsize),(float)1.0,(float)(1.0-6*tsize),stime,etime,vlist,vcount,"");

	/* display CNS list */
	drawlevel(20,(float)0.0,(float)(1.0-7*tsize),(float)1.0,(float)(1.0-6*tsize),stime,etime,clist,ccount,"SEG");

	/* locate segment bundaries */
	digit_nxgrid=0;
	for (i=0;i<vcount;i++) {
		digit_xgrid[digit_nxgrid++] = (int)(digitab.ixl+digitab.ixoff+(digitab.ixr-digitab.ixl-digitab.ixoff)*(vlist[i].stime-stime)/(etime-stime));
		digit_xgrid[digit_nxgrid++] = (int)(digitab.ixl+digitab.ixoff+(digitab.ixr-digitab.ixl-digitab.ixoff)*(vlist[i].etime-stime)/(etime-stime));
	}
	for (i=0;i<ccount;i++) {
		digit_xgrid[digit_nxgrid++] = (int)(digitab.ixl+digitab.ixoff+(digitab.ixr-digitab.ixl-digitab.ixoff)*(clist[i].stime-stime)/(etime-stime));
		digit_xgrid[digit_nxgrid++] = (int)(digitab.ixl+digitab.ixoff+(digitab.ixr-digitab.ixl-digitab.ixoff)*(clist[i].etime-stime)/(etime-stime));
	}

	if (doan) {
		digitemAN(20,(float)0.0,(float)(1.0-8*tsize),(float)1.0,(float)(1.0-7*tsize),&anitem,an,stime,etime,DIGITEMBOX);
		fxpos = (1.0-8*tsize);
	}
	else
		fxpos = (1.0-7*tsize);

	if (dosp) {
		digitemFFT(20201626,(float)0.0,(float)(2*tsize),(float)1.0,(float)fxpos,&spitem,sp,stime,etime,DIGITEMBOX|DIGITEMGRID);
	}
	else {
		/* draw Fx contour */
		if ((lofx!=0)&&(hifx!=0)) {
			digitab.lo = lofx;
			digitab.hi = hifx;
			digitemFX(20,(float)0.0,(float)(2*tsize),(float)1.0,(float)fxpos,&fxitem,fx,stime,etime,DIGITEMBOX|DIGITEMFIX|DIGITEMGRID);
		}
		else
			digitemFX(20,(float)0.0,(float)(2*tsize),(float)1.0,(float)fxpos,&fxitem,fx,stime,etime,DIGITEMBOX|DIGITEMGRID);
	}

	/* draw WORD list */
	drawlevel(20,(float)0.0,(float)tsize,(float)1.0,(float)(2*tsize),stime,etime,wlist,wcount,"WORD");
	
	/* put up second timescale */
	digitemtime(20,(float)0.0,(float)0.0,(float)1.0,(float)tsize,stime,etime,1);
	digflush();
	
} while (digwait()==DIG_REDRAW);

	digend();
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	NSL_File 	inf;
	NSL_Doctype 	dct=NULL;
	NSL_Item 	*uttitem;
	NSL_Item	*fileitem;
	NSL_Item 	*ipitem;
	NSL_Bit		*bit;

	int		it;
	char		*ty;
	char		*fxtype="0";
	char		*antype="0";
	char		*sptype="0";

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ii:f:ag"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance alignment to SFS annotation Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'i' :	/* input specification */
		if (itspec(optarg,&it,&ty)==0) {
			if (it==FX_TYPE)
				fxtype=ty;
			else if (it==AN_TYPE) {
				antype=ty;
				doan=1;
			}
			else if (it==SP_TYPE) {
				sptype=ty;
				dosp=1;
			}
			else
				error("unsuitable item specification: '%s'",optarg);
		}
		else
			error("illegal item specification: '%s'",optarg);
		break;
	case 'f':	/* fix fx range */
		lofx = atoi(optarg);
		optarg = argv[optind++];
		hifx = atoi(optarg);
		break;
	case 'a':	/* display annotation item */
		doan=1;
		break;
	case 'g':	/* display spectrogram item */
		dosp=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-i item) (-f lofx hifx) (-a) input.xml",PROGNAME);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, dct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else
		error("no input file specified");

	/* get unique names for attributes */
	dct = DoctypeFromFile(inf);
	attrID = AttrUniqueName(dct,"ID",0);
	attrHREF = AttrUniqueName(dct,"HREF",0);
	attrSTART = AttrUniqueName(dct,"START",0);
	attrSTOP = AttrUniqueName(dct,"STOP",0);

	/* read first tag - should be <UTT> */
	if ((bit=GetNextBit(inf))==NULL)
		error("Could not read input file");

	if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT")) && (strcmp(bit->label,"PROSYNTH"))))
		error("Could not read <UTT> or <PROSYNTH> start tag");

	/* load the whole thing into memory */
	uttitem = ItemParse(inf,bit->value.item);

	/* find name of SFS file */
	fileitem = MyObtainItem(uttitem,"FILE");
	sfilename = strsave(GetAttrStringVal(fileitem,attrHREF));

	/* open the SFS file and load fx contour */
	if (!sfilename || !sfilename[0])
		error("no SFS file given");
	if (doan)
		getitem(sfilename,AN_TYPE,antype,&anitem,(void *)&an);
	if (dosp)
		getitem(sfilename,SP_TYPE,antype,&spitem,(void *)&sp);
	else
		getitem(sfilename,FX_TYPE,fxtype,&fxitem,(void *)&fx);

	/* find the IP item */
	ipitem = MyObtainItem(uttitem,"IP");
	if (!ipitem)
		error("No IP item in input");

	/* load WORD data */
	wcount = countunit(uttitem,"WORD",0);
	wlist = (struct token_rec *)calloc(wcount,sizeof(struct token_rec));
	wcount=0;
	copyunit(uttitem,"WORD",wlist,&wcount);

	/* load IP data */
	icount = countunit(uttitem,"IP",0);
	ilist = (struct token_rec *)calloc(icount,sizeof(struct token_rec));
	icount=0;
	copyunit(uttitem,"IP",ilist,&icount);

	/* load AG data */
	acount = countunit(uttitem,"AG",0);
	alist = (struct token_rec *)calloc(acount,sizeof(struct token_rec));
	acount=0;
	copyunit(uttitem,"AG",alist,&acount);

	/* load FOOT data */
	fcount = countunit(uttitem,"FOOT",0);
	flist = (struct token_rec *)calloc(fcount,sizeof(struct token_rec));
	fcount=0;
	copyunit(uttitem,"FOOT",flist,&fcount);

	/* load SYL data */
	scount = countunit(uttitem,"SYL",0);
	slist = (struct token_rec *)calloc(scount,sizeof(struct token_rec));
	scount=0;
	copyunit(uttitem,"SYL",slist,&scount);

	/* load VOC data */
	vcount = countunit(uttitem,"VOC",0);
	vlist = (struct token_rec *)calloc(vcount,sizeof(struct token_rec));
	vcount=0;
	copyunit(uttitem,"VOC",vlist,&vcount);

	/* load CNS data */
	ccount = countunit(uttitem,"CNS",0);
	clist = (struct token_rec *)calloc(ccount,sizeof(struct token_rec));
	ccount=0;
	copyunit(uttitem,"CNS",clist,&ccount);

#ifdef IAG
	/* dump everything */
	printf("WORDs:\n");
	for (i=0;i<wcount;i++) printf("%12s %8.4f %8.4f\n",wlist[i].str,wlist[i].stime,wlist[i].etime);
	printf("IPs:\n");
	for (i=0;i<icount;i++) printf("%12s %8.4f %8.4f\n",ilist[i].str,ilist[i].stime,ilist[i].etime);
	printf("AGs:\n");
	for (i=0;i<acount;i++) printf("%12s %8.4f %8.4f\n",alist[i].str,alist[i].stime,alist[i].etime);
	printf("FOOTs:\n");
	for (i=0;i<fcount;i++) printf("%12s %8.4f %8.4f\n",flist[i].str,flist[i].stime,flist[i].etime);
	printf("SYLs:\n");
	for (i=0;i<scount;i++) printf("%12s %8.4f %8.4f\n",slist[i].str,slist[i].stime,slist[i].etime);
	printf("VOCs:\n");
	for (i=0;i<vcount;i++) printf("%12s %8.4f %8.4f\n",vlist[i].str,vlist[i].stime,vlist[i].etime);
	printf("CNSs:\n");
	for (i=0;i<ccount;i++) printf("%12s %8.4f %8.4f\n",clist[i].str,clist[i].stime,clist[i].etime);
	printf("Fx contour:\n");
	for (i=0;i<fxitem.numframes;i++) printf("%3d, ",fx[i]);
	printf("\n");
#endif

	/* draw picture */
	draw();

	/* that's all folks */
	SFclose(inf);
	return 0;
}
