/* uttmake -- generate a new utterance file */

/* Mark Huckvale - University College London */

/* version 1.0 - March 1998 */
/* version 1.1 - March 1999
	- allow text to come on standard input
*/
/* version 1.2 - May 1999
	- process paragraphs of IP labelled text
*/
/* version 1.3 - January 2000
	- allow multiple stress marks/word
	- use ` and \ to mark AG and FOOT boundaries
*/
/* version 1.4 - April 2001
	- allow entry of filenames into <FILE> element with -s, -l and -x switches
*/
/* version 1.5 - August 2001
	- input XML text tokens now wrapped in <TOKEN>
*/

#define PROGNAME "uttmake"
#define PROGVERS "1.5"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"

/* input file */
char		filename[256];	/* input file (from sdump -H) */
char		ofilename[256];	/* output filename */
char		tfilename[256];	/* temporary output filename */
char		sfilename[256];	/* signal filename */
char		lfilename[256];	/* label filename */
char		xfilename[256];	/* analysis filename */
char		idname[256];	/* utterance id (=filename without extension) */
char		idname2[256];	/* utterance id (=filename without extension) */

/* output file */
NSL_File 	outf;
NSL_Doctype 	odct=NULL;
const char *	attrHREF;
const char *	attrID;
const char *	attrSTRESS;
const char *	attrPREF;
const char *	attrSIGNAL;
const char *	attrLABELS;
const char *	attrSPDATA;

/* token to process */
char	*token;
char	*tbuf;
char	wordID[16];

/* word records */
struct word_rec {
	char	*str;
	int	flags;
	int	pronid;
	char	strpat[16];
	struct word_rec *next;
};

#define WFLAG_STRESS_NONE	0
#define WFLAG_STRESS_FOOT	1
#define WFLAG_STRESS_ACCENT	2

struct word_rec *wroot;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* create a new item of a supplied type */
NSL_Item * newItem(char *ty)
{
	NSL_Item *itm;
	itm = NewNullNSLItem(odct,ty,0);
	itm->type = NSL_empty;
	return(itm);
}

/* check a word for stress mark */
int	checkword(char *src,char *strpat,int *pronid)
{
	int	stress=WFLAG_STRESS_NONE;
	char	*dst=src;
	int	val;

	*pronid=1;

	while (*src) {
		if (*src=='`') {
			stress=WFLAG_STRESS_ACCENT;
			*strpat++ = 'A';
			src++;
		}
		else if (*src=='\\') {
			stress=WFLAG_STRESS_FOOT;
			*strpat++ = 'F';
			src++;
		}
		else if (*src=='#') {
			val=0;
			src++;
			while (isdigit(*src)) {
				val = 10*val + *src - '0';
				src++;
			}
			*pronid = val;
		}
		else if (!strchr("?",*src))
			*dst++ = *src++;
		else
			src++;
	}
	*dst='\0';
	if (stress==WFLAG_STRESS_NONE) *strpat++='N';
	*strpat='\0';
	return(stress);
}

/* make a word item from token */
NSL_Item *compose(char *token)
{
	char		*p;
	struct word_rec *w,*wtmp;
	struct word_rec **wlast=&wroot;
	NSL_Item	*wordseqitem;
	NSL_Item	*worditem;
	int		wno=0;

	/* split token into words */
	p = strtok(token," \t\r\n");
	while (p && *p) {
		/* get a new word record */
		if ((w = (struct word_rec *)calloc(1,sizeof(struct word_rec)))==NULL)
			error("out of memory");
		w->str = strsave(p);
		w->flags = checkword(w->str,w->strpat,&(w->pronid));
		w->next = NULL;

		/* add to end of chain */
		*wlast = w;
		wlast = &w->next;
		p = strtok(NULL," \t\r\n");
	}

	wordseqitem = newItem("WORDSEQ");

	w = wroot;
	while (w) {
		worditem = newItem("WORD");
		worditem->data = NewTextNSLData(odct,strsave(w->str),0,FALSE,NULL,FALSE);
		worditem->type = NSL_non_empty;
		if (w->flags != WFLAG_STRESS_NONE) PutAttrVal(worditem,attrSTRESS,strsave(w->strpat));
		if (w->pronid!=1) {
			sprintf(wordID,"#%d",w->pronid);
			PutAttrVal(worditem,attrPREF,strsave(wordID));
		}
		sprintf(wordID,"WORD%d",++wno);
		PutAttrVal(worditem,attrID,strsave(wordID));
		AddItemToEnd(wordseqitem,worditem);
		w = w->next;
	}

	/* free up memory used */
	w = wroot;
	while (w) {
		free(w->str);
		wtmp = w->next;
		free(w);
		w=wtmp;
	}
	wroot=NULL;
	return(wordseqitem);
}

/* read a string from standard input */
char *readinput()
{
	char	*buf=malloc(4096);
	int	buflen=0;
	int	bufsiz=4096;
	char	xbuf[4000];
	char	*p;

	buf[0]='\0';
	while (fgets(xbuf,4000,stdin)!=NULL) {
		p = strtok(xbuf,"\r\n");
		if (p && *p) {
			if ((buflen+strlen(p))>=bufsiz) {
				buf = realloc(buf,bufsiz+4096);
				bufsiz += 4096;
			}
			if (*buf) strcat(buf," ");
			strcat(buf,p);
			buflen = strlen(buf);
		}
	}
	return(buf);
}

/* parse a char string into IPs */
static char *ippostext;

char *getnextip()
{
	char	*sptr = ippostext;
	char	*eptr;

	/* skip to first non-/ */
	while (*sptr && (isspace(*sptr) || (*sptr=='/'))) sptr++;
	if (*sptr=='\0') return NULL;
	ippostext=sptr;

	/* skip to end of IP */
	while (*ippostext && (*ippostext!='/')) ippostext++;
	eptr = ippostext;
	if (*ippostext=='/') ippostext++;
	if (*eptr=='/') {
		while ((eptr>sptr)&&isspace(*(eptr-1))) eptr--;
		*eptr='\0';
	}
	return(sptr);
}

char *getfirstip(char *iptxt)
{
	ippostext = iptxt;
	return getnextip();
}


/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		dtdname[256];
	char		*p,*iptxt;

	NSL_File 	inf;
	NSL_Doctype 	idct=NULL;
	NSL_Item 	*item;
	NSL_Item 	*psitem;
	NSL_Item 	*uttitem;
	NSL_Item 	*seqitem;
	NSL_Item 	*filitem;
	NSL_Data	*data;
	NSL_Bit		*bit;
	int		ipno=0;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"ITt:s:l:x:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: initialise utterance file Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 't':	/* token supplied on command line */
		token = strsave(optarg);
		break;
	case 'T':	/* token supplied on standard input */
		token=readinput();
		break;
	case 's':	/* signal file name */
		strcpy(sfilename,optarg);
		break;
	case 'l':	/* label file name */
		strcpy(lfilename,optarg);
		break;
	case 'x':	/* analysis file name */
		strcpy(xfilename,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-T|-t token|input.xml) (-s sigfile) (-l labfile) (-x datfile) (out.xml)",PROGNAME);

	/* get the name of the DTD */
	if (argv[0][1]==':')
		strcpy(dtdname,&argv[0][2]);
	else
		strcpy(dtdname,argv[0]);
	for (p = dtdname;*p;p++) if (*p=='\\') *p = '/';
	if ((p=strrchr(dtdname,'.'))!=NULL)
		strcpy(p,".dtd");
	else
		strcat(dtdname,".dtd");

	/* open the dtd file */
	if ((inf = SFopen(dtdname, odct, NSL_read))==NULL)
		error("unable to open DTD file '%s'",dtdname);
	odct = DoctypeFromFile(inf);
	SFclose(inf);

	/* get input filename */
	if (!token && (optind < argc)) {
		strcpy(filename,argv[optind]);
		optind++;
	}
	else
		filename[0]='\0';

	/* get output attribute names */
	attrHREF = AttrUniqueName(odct,"HREF",0);
	attrID = AttrUniqueName(odct,"ID",0);
	attrSTRESS = AttrUniqueName(odct,"STRESS",0);
	attrPREF = AttrUniqueName(odct,"PREF",0);
	attrSIGNAL = AttrUniqueName(odct,"SIGNAL",0);
	attrLABELS = AttrUniqueName(odct,"LABELS",0);
	attrSPDATA = AttrUniqueName(odct,"SPDATA",0);

	/* Open the output NSGML stream using the segparse doctype declaration */
	if (optind < argc) {
		strcpy(ofilename,argv[optind]);
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
		outf = SFopen(tfilename, odct, NSL_write_pretty);
	}
	else
		outf = SFFopen(stdout, odct, NSL_write_pretty, "stdout");

	/* get token to process */
	if (token) {

		/* create prosynth item */
		psitem = newItem("PROSYNTH");

		if (iptxt=getfirstip(token)) do {
		 	/* create output utterance item */
			uttitem = newItem("UTT");
			sprintf(idname,"UTT%03d",++ipno);
			PutAttrVal(uttitem,attrID,strsave(idname));

			/* add a file item */
			if (*sfilename || *lfilename || *xfilename) {
				filitem = newItem("FILE");
				if (*sfilename) PutAttrVal(filitem,attrSIGNAL,sfilename);
				if (*lfilename) PutAttrVal(filitem,attrLABELS,lfilename);
				if (*xfilename) PutAttrVal(filitem,attrSPDATA,xfilename);
				AddItemToEnd(uttitem,filitem);
			}

			/* add token item */
			tbuf = malloc(strlen(iptxt)+32);
			sprintf(tbuf,"<TOKEN>%s</TOKEN>",iptxt);
			AddItemToEnd(uttitem,GetItemFromString(tbuf,odct));
			free(tbuf);

			/* create word item */
			seqitem = compose(iptxt);
			AddItemToEnd(uttitem,seqitem);

			AddItemToEnd(psitem,uttitem);

		} while (iptxt=getnextip());

		/* print the utterance */
		PrintItem(outf,psitem);

	}
	else {
		/* open input file */
		if (filename[0]=='\0') {
			if ((inf = SFFopen(stdin, odct, NSL_read, "stdin"))==NULL)
				error("unable to open input stream");
			strcpy(filename,"STDIN");
		}
		else {
			if ((inf = SFopen(filename, odct, NSL_read))==NULL)
				error("unable to open '%s'",filename);
		}

		/* read first tag - should be <SFSFILE> or <PROSYNTH> */
		if ((bit=GetNextBit(inf))==NULL)
			error("Could not read <SFSFILE> or <PROSYNTH> start tag");

		/* choose operation on basis of first tag */
		if ((bit->type==NSL_start_bit) && (strcmp(bit->label,"SFSFILE")==0)) {
			FreeBit(bit);

			/* create prosynth item */
			psitem = newItem("PROSYNTH");

		 	/* create output utterance annotation referenced to input file */
			uttitem = newItem("UTT");
			PutAttrVal(uttitem,attrHREF,filename);
			strcpy(idname,filename);
			if ((p=strrchr(idname,'.'))!=NULL) *p='\0';
			PutAttrVal(uttitem,attrID,idname);

			/* Loop round reading components of the input text */
			while ((bit=GetNextBit(inf))!=NULL) {
				if ((bit->type==NSL_start_bit) || (bit->type==NSL_empty_bit)) {
					/* scan entries looking for FILE & TOKEN */
					item=ItemParse(inf,bit->value.item);

					if (strcmp(item->label,"FILE")==0) {

						/* add any extra filenames */
						if (*sfilename) PutAttrVal(item,attrSIGNAL,sfilename);
						if (*lfilename) PutAttrVal(item,attrLABELS,lfilename);
						if (*xfilename) PutAttrVal(item,attrSPDATA,xfilename);

						/* add the FILE entry  */
						AddItemToEnd(uttitem,CopyItem(item));
					}
					else if (strcmp(item->label,"TOKEN")==0) {

						/* add the TOKEN entry straight */
						AddItemToEnd(uttitem,CopyItem(item));

						/* process the TOKEN data */
						data = item->data;
						if (data->type==NSL_text_data) {
							token = strsave(data->first);
							seqitem = compose(token);
							AddItemToEnd(uttitem,seqitem);
							free(token);
						}
					}
					/* else skip item */
					FreeItem(item);
				}
				else
					FreeBit(bit);
			}

			/* print the utterance */
			AddItemToEnd(psitem,uttitem);
			PrintItem(outf,psitem);

		}
		else if ((bit->type==NSL_start_bit) && (strcmp(bit->label,"PROSYNTH")==0)) {
			FreeBit(bit);

			/* create prosynth item */
			psitem = newItem("PROSYNTH");

			/* Loop round reading components of the input text */
			while ((bit=GetNextBit(inf))!=NULL) {
				if ((bit->type==NSL_start_bit) || (bit->type==NSL_empty_bit)) {
					/* scan entries looking for IP */
					item=ItemParse(inf,bit->value.item);

					if ((strcmp(item->label,"IP")==0)||(strcmp(item->label,"UTT")==0)||(strcmp(item->label,"TOKEN")==0)) {

						/* save all bits of text we find */
						data = item->data;
						if ((data!=NULL)&&(data->type==NSL_text_data)) {

						 	/* create output utterance annotation referenced to input file */
							uttitem = newItem("UTT");
							PutAttrVal(uttitem,attrHREF,filename);
							strcpy(idname,filename);
							if ((p=strrchr(idname,'.'))!=NULL) *p='\0';
							sprintf(idname2,"%s%03d",idname,++ipno);
							PutAttrVal(uttitem,attrID,strsave(idname2));

							/* process the IP data */
							token = strsave(data->first);

							/* convert IP entry to TOKEN */
							tbuf = malloc(strlen(token)+32);
							sprintf(tbuf,"<TOKEN>%s</TOKEN>",token);
							AddItemToEnd(uttitem,GetItemFromString(tbuf,odct));
							free(tbuf);

							/* process word sequence */
							seqitem = compose(token);
							AddItemToEnd(uttitem,seqitem);
							free(token);
						}

						/* append the utterance */
						AddItemToEnd(psitem,uttitem);
					}
					/* else skip item */
					FreeItem(item);
				}
				else
					FreeBit(bit);
			}
			PrintItem(outf,psitem);
			ForceNewline(outf);
		}
		else
			error("Could not read <SFSFILE> or <PROSYNTH> start tag");
		SFclose(inf);
	}

	/* that's all folks */
	SFclose(outf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
