/* uttmbrola -- XML utterance to MBROLA format transfer */

/* Mark Huckvale - University College London */

/* version 1.0 - August 1998 */
/* version 1.1 - February 1999
	- option to get Fx from SFS file
*/
/* version 1.2 - March 1999
	- use DUR, MINDUR & INHDUR if present
*/
/* version 1.3 - July 1999
	- rip out old FXMID/FXGRD method
	- support IPITCH/FPITCH/POF/PON/LON attributes
*/
/* version 1.4 - March 2000
	- add support for nucleus containing "ju"
	- leave Fx interpolation to MBROLA
*/

/*
 * uttmbrola takes the aligned and Fx loaded syllable structure from an
 * utterance file and outputs MBROLA commands
 */
 
#define PROGNAME "uttmbrola"
#define PROGVERS "1.4"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"
#include "mbrola.h"
#include "sfs.h"

/* input file */
char		filename[256];	/* input file */

/* SFS file */
char		*sfilename;
int		realfx=0;
char		*fxtype="0";
struct item_header fxitem;
short		*fx;

NSL_Doctype	dct=NULL;
const char	*attrHREF;
const char	*attrSTART;
const char	*attrSTOP;
const char	*attrVOI;
const char	*attrWEIGHT;
const char	*attrDUR;
const char	*attrMINDUR;
const char	*attrINHDUR;
const char	*attrID;
const char	*attrIPITCH;
const char	*attrFPITCH;
const char	*attrPON;
const char	*attrPOF;
const char	*attrLON;
const char	*attrAMBI;

/* phone features */
const char *featname[]={
	"CNT",
	"VOI",
	"NAS",
	"STR",
	"SON",
	"RHO",
	"CNSGRV",
	"CNSCMP",
#if 0
	"VOCGRV",
	"VOCHEIGHT",
	"VOCRND",
#endif
	"GRV",
	"HEIGHT",
	"RND",
};
#define NFEATNAME (sizeof(featname)/sizeof(char *))

const char *attrFEAT[NFEATNAME];

/* Fx records */
struct fx_rec {
	double	time;
	double	fxval;
};
struct fx_rec *fxtab;
int		fxtablen;
int		fxtabpos;

#define MAXNUMFXSEG	10
/* phone records */
struct phone_rec {
	char	*name;
	int	dble;
	const char *feat[NFEATNAME];
	const char *feat2[NFEATNAME];
	int	voiced;
	int	heavy;
	int	split;
	double	stime;
	double	etime;
	double	sdur;
	struct fx_rec fx[MAXNUMFXSEG];
	double	stime2;
	double	etime2;
	double	sdur2;
	struct fx_rec fx2[MAXNUMFXSEG];
};

struct phone_rec *uphone;	/* utterance phones */
int	ucount;
double	dstime=0.0;

struct phone_rec *mphone;	/* MBROLA phones */
int	mcount;

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Data *data,const char *name)
{
	NSL_Item *item;
	for (;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
		}
	}
	return(NULL);
}

/* count number of CNS and VOC units */
int countphone(NSL_Item *item,int count)
{
	NSL_Data *data;

	if (!item) return(count);

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if ((strcmp(item->label,"VOC")==0) ||
			    (strcmp(item->label,"CNS")==0)) count++;
			else
				count = countphone(item,count);
		}
	}
	return(count);
}

/* distance metric */
int distance(const char **feat1,const char **feat2)
{
	int	i;
	int	dist=0;

	for (i=0;i<NFEATNAME;i++) {
		if (feat1[i]) {
			if (feat2[i]) {
				if (strcmp(feat1[i],feat2[i])==0)
					dist += 0; /* same feature - same value */
				else {
					dist += 1; /* same feature - different value */
#if 0
printf("%s:%s:%s\n",featname[i],feat1[i],feat2[i]);
#endif
				}
			}
			else
				dist += 0; /* unspecified feature */
		}
		else {
			if (feat2[i]) {
				dist += 2; /* absent feature */
#if 0
printf("%s:NULL:%s\n",featname[i],feat2[i]);
#endif
			}
			else
				dist += 0; /* neither has feature */
		}
	}
	return(dist);
}

/* safe version of atof() */
double myatof(const char *str)
{
	if (str && *str && (isdigit(*str)||(*str=='-')))
		return(atof(str));
	else {
		return(0.0);
	}
}

/* safe version of strcmp */
int mystrcmp(const char *s,const char *t)
{
	if (s && t)
		return(strcmp(s,t));
	else if (s || t)
		return(1);
	else
		return(0);
}

/* version of get attr that returns NULLs */
const char *MyGetAttrStringVal(NSL_Item *item,const char *attr)
{
	const char *p=GetAttrStringVal(item,attr);
	if (!p || !*p)
		return(NULL);
	else
		return(p);
}

/* look up time in fx contour */
int findfx(double t)
{
	int	idx;

	if (fx==NULL) return(0);
	idx = (int)((t-fxitem.offset)/fxitem.frameduration);
	if (idx < 0)
		return(fx[0]);
	else if (idx < fxitem.numframes)
		return(fx[idx]);
	else
		return(fx[fxitem.numframes-1]);
}

/* calculate duration from DUR, MINDUR & INHDUR attributes */
double calcdur(NSL_Item *item)
{
	double dur = myatof(MyGetAttrStringVal(item,attrDUR));
	double mindur = myatof(MyGetAttrStringVal(item,attrMINDUR));
	double inhdur = myatof(MyGetAttrStringVal(item,attrINHDUR));
	return (mindur + (inhdur-mindur)*dur);
}

/* copy over CNS and VOC units */
void copyphone(NSL_Item *item,int heavy,int onset,double *start,double *stop)
{
	int	i;
	NSL_Data *data;
	const char *pweight;
	char	sbuf[32];
	double	tstart,tstop;

	*start = dstime;
	*stop = dstime;
	
	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"NUC")==0) {
				pweight = GetAttrStringVal(item,attrWEIGHT);
				if (pweight && (strcmp(pweight,"HEAVY")==0)) heavy=1;
				copyphone(item,heavy,onset,&tstart,&tstop);
				sprintf(sbuf,"%.03f",tstart);
				PutAttrVal(item,attrSTART,strsave(sbuf));
				sprintf(sbuf,"%.03f",tstop);
				PutAttrVal(item,attrSTOP,strsave(sbuf));
			}
			else if (strcmp(item->label,"VOC")==0) {
			    	for (i=0;i<NFEATNAME;i++)
			    		uphone[ucount].feat[i] = MyGetAttrStringVal(item,attrFEAT[i]);
		    		uphone[ucount].stime = myatof(GetAttrStringVal(item,attrSTART));
		    		if (uphone[ucount].stime==0) uphone[ucount].stime = dstime;
		    		uphone[ucount].etime = myatof(GetAttrStringVal(item,attrSTOP));
		    		if (uphone[ucount].etime==0) {
		    			uphone[ucount].sdur = calcdur(item);
					uphone[ucount].etime = uphone[ucount].stime + uphone[ucount].sdur;
		    		}
		    		else
			    		uphone[ucount].sdur = uphone[ucount].etime-uphone[ucount].stime;
			    	dstime = uphone[ucount].etime;

				data = data->next;
				if (data->type != NSL_item_data)
					error("unpaired VOC segment");
			    	item = data->first;
			    	for (i=0;i<NFEATNAME;i++)
			    		uphone[ucount].feat2[i] = MyGetAttrStringVal(item,attrFEAT[i]);
		    		uphone[ucount].stime2 = myatof(GetAttrStringVal(item,attrSTART));
		    		if (uphone[ucount].stime2==0) uphone[ucount].stime2 = dstime;
		    		uphone[ucount].etime2 = myatof(GetAttrStringVal(item,attrSTOP));
		    		if (uphone[ucount].etime2==0) {
		    			uphone[ucount].sdur2 = calcdur(item);
		    			uphone[ucount].etime2 = uphone[ucount].stime2 + uphone[ucount].sdur2;
		    		}
		    		else
			    		uphone[ucount].sdur2 = uphone[ucount].etime2-uphone[ucount].stime2;
			    	dstime = uphone[ucount].etime2;
				uphone[ucount].dble=1;
				uphone[ucount].voiced=1;
				uphone[ucount].heavy=heavy;
				uphone[ucount].split=0;
			    	ucount++;
			}
			else if (strcmp(item->label,"CNS")==0) {
				if (onset && (strcmp(MyGetAttrStringVal(item,attrAMBI),"Y")==0)) {
					/* ambisyllabic consonant in onset - skip */
				}
				else {
					for (i=0;i<NFEATNAME;i++)
				    		uphone[ucount].feat[i] = MyGetAttrStringVal(item,attrFEAT[i]);
			    		uphone[ucount].stime = myatof(GetAttrStringVal(item,attrSTART));
			    		if (uphone[ucount].stime==0) uphone[ucount].stime = dstime;
			    		uphone[ucount].etime = myatof(GetAttrStringVal(item,attrSTOP));
			    		if (uphone[ucount].etime==0) {
			    			uphone[ucount].sdur = calcdur(item);
			    			uphone[ucount].etime = uphone[ucount].stime + uphone[ucount].sdur;
			    		}
			    		else
				    		uphone[ucount].sdur = uphone[ucount].etime-uphone[ucount].stime;
				    	dstime = uphone[ucount].etime;
			    		uphone[ucount].dble=0;
					uphone[ucount].heavy=0;
					uphone[ucount].split=0;
					if (mystrcmp(GetAttrStringVal(item,attrVOI),"Y")==0)
						uphone[ucount].voiced=1;
					else
						uphone[ucount].voiced=0;
					if ((ucount==0)||(uphone[ucount].stime==0.0)||(uphone[ucount-1].stime < uphone[ucount].stime))
					    	ucount++;
				}
			}
			else if (strcmp(item->label,"ONSET")==0) {
				copyphone(item,heavy,1,&tstart,&tstop);
				sprintf(sbuf,"%.03f",tstart);
				PutAttrVal(item,attrSTART,strsave(sbuf));
				sprintf(sbuf,"%.03f",tstop);
				PutAttrVal(item,attrSTOP,strsave(sbuf));
			}
			else {
				copyphone(item,heavy,onset,&tstart,&tstop);
				sprintf(sbuf,"%.03f",tstart);
				PutAttrVal(item,attrSTART,strsave(sbuf));
				sprintf(sbuf,"%.03f",tstop);
				PutAttrVal(item,attrSTOP,strsave(sbuf));
			}	
		}
	}
	*stop = dstime;
}

/* create table of MBROLA phones */
void createmtable()
{
	int	i,j;
	NSL_Item	*item;

	/* allocate enough space for table */
	mphone = (struct phone_rec *)calloc(2*NUM_MBROLA_SEG,sizeof(struct phone_rec));
	if (!mphone)
		error("out of memory");

	/* copy MBROLA segment definitions into table */
	mcount=0;
	for (i=0;i<NUM_MBROLA_SEG;i++) {
		mphone[mcount].name = strsave(mbrola_segtab[i].sym);
		item = GetItemFromString(mbrola_segtab[i].elem,dct);
	    	for (j=0;j<NFEATNAME;j++)
		    	mphone[mcount].feat[j] = MyGetAttrStringVal(item,attrFEAT[j]);
		if (*mbrola_segtab[i].elem2) {
			item = GetItemFromString(mbrola_segtab[i].elem2,dct);
		    	for (j=0;j<NFEATNAME;j++)
			    	mphone[mcount].feat2[j] = MyGetAttrStringVal(item,attrFEAT[j]);
			mphone[mcount].dble=1;
		}
		else
			mphone[mcount].dble=0;
		mphone[mcount].heavy = mbrola_segtab[i].heavy;
		mcount++;
	}
}

/* smooth table of fx values */
void smoothfx(short *buff,int len)
{
	int	i,idx=0;
	int	sidx,sval;
	int	eidx,eval;

/*
printf("Before:");
for (i=0;i<len;i++) printf(" %d",buff[i]);
printf("\n");
*/
	/* fill in start with flat contour */
	while ((idx < len) && (buff[idx]==0)) idx++;
	if (idx==len) return;
	for (i=0;i<idx;i++) buff[i] = buff[idx];

	/* fill in end with flat contour */
	idx = len-1;
	while ((idx > 0) && (buff[idx]==0)) idx--;
	if (idx==0) return;
	for (i=idx+1;i<len;i++) buff[i] = buff[idx];

	/* fill in gaps in middle */
	idx=0;
	while (idx < len) {
		if (buff[idx]==0) {
			/* found a gap - scan to end */
			while ((idx < len) && (buff[idx]==0)) idx++;
			eidx = idx;
			eval = buff[idx];
			for (i=1;i<(eidx-sidx);i++)
				buff[sidx+i] = (int)(0.5 + sval + i*(double)(eval-sval)/(double)(eidx-sidx));
		}
		sidx = idx;
		sval = buff[idx];
		idx++;
	}
/*
printf("After:");
for (i=0;i<len;i++) printf(" %d",buff[i]);
printf("\n");
*/
}

/* set an Fx value at given time */
void setfx(double t,double f)
{

	if (f==0) return;
printf("; setfx(%.1f,%g)\n",1000*t,f);
	if (fxtabpos==fxtablen)
		error("internal Fx buffer overflow");
	fxtab[fxtabpos].time = t;
	fxtab[fxtabpos].fxval = f;
	fxtabpos++;
}

/* process fx from attributes */
void procfx(NSL_Item *item)
{
	NSL_Data	*data;
	double		d;

	/* process any values on this node */
	if (MyGetAttrStringVal(item,attrPON)&&MyGetAttrStringVal(item,attrPOF)&&MyGetAttrStringVal(item,attrLON)) {
		d = myatof(GetAttrStringVal(item,attrSTOP)) - myatof(GetAttrStringVal(item,attrSTART));
		setfx(myatof(GetAttrStringVal(item,attrSTART))+
			myatof(GetAttrStringVal(item,attrPON))*d/100,
			myatof(GetAttrStringVal(item,attrIPITCH)));
		setfx(myatof(GetAttrStringVal(item,attrSTART))+
			myatof(GetAttrStringVal(item,attrPOF))*d/100,
			myatof(GetAttrStringVal(item,attrIPITCH)));		
		setfx(myatof(GetAttrStringVal(item,attrSTART))+
			myatof(GetAttrStringVal(item,attrLON))*d/100,
			myatof(GetAttrStringVal(item,attrFPITCH)));
	}
	else if (MyGetAttrStringVal(item,attrPON)&&MyGetAttrStringVal(item,attrPOF)) {
		d = myatof(GetAttrStringVal(item,attrSTOP)) - myatof(GetAttrStringVal(item,attrSTART));
		setfx(myatof(GetAttrStringVal(item,attrSTART))+
			myatof(GetAttrStringVal(item,attrPON))*d/100,
			myatof(GetAttrStringVal(item,attrIPITCH)));
		setfx(myatof(GetAttrStringVal(item,attrSTART))+
			myatof(GetAttrStringVal(item,attrPOF))*d/100,
			myatof(GetAttrStringVal(item,attrFPITCH)));		
	}
	else if (MyGetAttrStringVal(item,attrPON)) {
		d = myatof(GetAttrStringVal(item,attrSTOP)) - myatof(GetAttrStringVal(item,attrSTART));
		setfx(myatof(GetAttrStringVal(item,attrSTART))+
			myatof(GetAttrStringVal(item,attrPON))*d/100,
			myatof(GetAttrStringVal(item,attrIPITCH)));
	}
	else {
		setfx(myatof(GetAttrStringVal(item,attrSTART)),myatof(GetAttrStringVal(item,attrIPITCH)));
		setfx(myatof(GetAttrStringVal(item,attrSTOP)),myatof(GetAttrStringVal(item,attrFPITCH)));
	}

	/* process any subordinate nodes */
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			procfx(item);
		}
	}
}

/* calculate fx from attributes */
void calcfx(NSL_Item *item,double totime)
{
	if (realfx) {
		/* get fx buffer */
		fxitem.offset = 0;
		fxitem.frameduration = 0.005;
		fxitem.numframes = (int)(1+totime/0.005);
		fx = (short *)calloc(fxitem.numframes,sizeof(short));
printf("; buffer ends at %g\n",fxitem.numframes*0.005);
	}
	else {
		/* get fx position list - ucount should be big enough */
		fxtablen = (ucount < 20) ? 20 : ucount;
		fxtab = (struct fx_rec *)calloc(fxtablen,sizeof(struct fx_rec));
		fxtabpos=0;
	}

	procfx(item);

	if (realfx) smoothfx(fx,fxitem.numframes);
}

/* copy fx into phone table */
void copyfx()
{
	int	i,j,k;

	if (realfx) {
		for (j=0;j<ucount;j++) {
	    		for (i=0;i<5;i++) {
				uphone[j].fx[i].time = uphone[j].stime+i*uphone[j].sdur/4;
				uphone[j].fx[i].fxval = findfx(uphone[j].fx[i].time);
			}
			if (uphone[j].dble)
		    		for (i=0;i<5;i++) {
					uphone[j].fx2[i].time = uphone[j].stime2+i*uphone[j].sdur2/4;
					uphone[j].fx2[i].fxval = findfx(uphone[j].fx[2].time);
				}
		}
	}
	else {
		for (j=0;j<ucount;j++) {
			k=0;
			for (i=0;(i<fxtabpos)&&(k<MAXNUMFXSEG);i++) {
				if ((uphone[j].stime <= fxtab[i].time) && (fxtab[i].time < uphone[j].stime+uphone[j].sdur))
					uphone[j].fx[k++] = fxtab[i];
			}
			if (uphone[j].dble) {
				k=0;
				for (i=0;(i<fxtabpos)&&(k<MAXNUMFXSEG);i++) {
					if ((uphone[j].stime2 <= fxtab[i].time) && (fxtab[i].time < uphone[j].stime2+uphone[j].sdur2))
						uphone[j].fx2[k++] = fxtab[i];
				}
			}
		}
	}
}
	
/* process an utterance */
void process(NSL_Item *ipitem,FILE *op,const char *idname)
{
	int		i,j,k;
	int		ulen;
	int		min,val,idx;
	double		sdur,tstart,tstop;
	char		sbuf[32];

	/* load the segments into temporary structure */
	ulen = countphone(ipitem,0);

	/* get memory */
	uphone = (struct phone_rec *)calloc(ulen,sizeof(struct phone_rec));
	if (!uphone)
		error("out of memory");

	/* copy over the utterance phones, doing duration calculation */
	ucount=0;
	dstime=0.0;
	copyphone(ipitem,0,0,&tstart,&tstop);
	sprintf(sbuf,"%.03f",tstart);
	PutAttrVal(ipitem,attrSTART,strsave(sbuf));
	sprintf(sbuf,"%.03f",tstop);
	PutAttrVal(ipitem,attrSTOP,strsave(sbuf));

	/* calculate the Fx contour and copy into phone table */
	if (!realfx) calcfx(ipitem,tstop);
	copyfx();

	/* match phones to MBROLA symbols */
	for (i=0;i<ucount;i++) {
		/* match features on segment to MBROLA symbols */
		min=1000;
		idx=0;
		for (j=0;j<mcount;j++) {
			val = distance(uphone[i].feat,mphone[j].feat);
			if (uphone[i].dble)
				val += distance(uphone[i].feat2,mphone[j].feat2);
			if ((uphone[i].heavy&1)!=(mphone[j].heavy&1)) val++;
			if (val < min) {
				min = val;
				idx = j;
			}
#if 0
printf("i=%d j=%d val=%d min=%d idx=%d\n",i,j,val,min,idx);
#endif
		}
		uphone[i].name = mphone[idx].name;
		/* test to see if single XML symbol must be split */
		if (mphone[idx].heavy&2) {
			uphone[i].dble=0;
			uphone[i].split=1;
		}
	}
	
	/* output MBROLA control data */
	if (idname)
		fprintf(op,"; MBROLA output from %s item %s",filename,idname);
	else
		fprintf(op,"; MBROLA output from %s",filename);
	if (realfx) fprintf(op," and %s(%d.%02d)",sfilename,fxitem.datatype,fxitem.subtype);
	fprintf(op,"\n");
	fprintf(op,"_  100\n");

	/* list segments */
	for (i=0;i<ucount;i++) {
		if (uphone[i].dble) {
			sdur = uphone[i].sdur + uphone[i].sdur2;
			fprintf(op,"%-2s %5.1f",uphone[i].name,(1000*sdur));
			k=0;
			while ((k<MAXNUMFXSEG)&&(uphone[i].fx[k].fxval>0)) {
				fprintf(op," %2d %d", (int)(100*(uphone[i].fx[k].time-uphone[i].stime)/sdur),(int)uphone[i].fx[k].fxval);
				k++;
			}
			k=0;
			while ((k<MAXNUMFXSEG)&&(uphone[i].fx2[k].fxval>0)) {
				fprintf(op," %2d %d", (int)(100*(uphone[i].fx2[k].time-uphone[i].stime)/sdur),(int)uphone[i].fx2[k].fxval);
				k++;
			}
			fprintf(op,"\n");
			tstop = uphone[i].stime2+uphone[i].sdur2;
		}
		else if (uphone[i].split) {
			strcpy(sbuf,uphone[i].name);
			sdur = uphone[i].sdur;
			fprintf(op,"%-2s %5.1f",strtok(sbuf,"+"),(1000*sdur));
			k=0;
			while ((k<MAXNUMFXSEG)&&(uphone[i].fx[k].fxval>0)) {
				fprintf(op," %2d %d", (int)(100*(uphone[i].fx[k].time-uphone[i].stime)/uphone[i].sdur),(int)uphone[i].fx[k].fxval);
				k++;
			}
			fprintf(op,"\n");
			sdur = uphone[i].sdur2;
			fprintf(op,"%-2s %3d",strtok(NULL,"+"),(int)(1000*sdur));
			k=0;
			while ((k<MAXNUMFXSEG)&&(uphone[i].fx2[k].fxval>0)) {
				fprintf(op," %2d %d", (int)(100*(uphone[i].fx2[k].time-uphone[i].stime2)/uphone[i].sdur2),(int)uphone[i].fx2[k].fxval);
				k++;
			}
			fprintf(op,"\n");
			tstop = uphone[i].stime2+uphone[i].sdur2;
		}
		else {
			sdur = uphone[i].sdur;
			fprintf(op,"%-2s %5.1f",uphone[i].name,(1000*sdur));
			k=0;
			while ((k<MAXNUMFXSEG)&&(uphone[i].fx[k].fxval>0)) {
				fprintf(op," %2d %d", (int)(100*(uphone[i].fx[k].time-uphone[i].stime)/uphone[i].sdur),(int)uphone[i].fx[k].fxval);
				k++;
			}
			fprintf(op,"\n");
			tstop = uphone[i].stime+uphone[i].sdur;
		}
	}

	/* do any last Fx point */
	fprintf(op,"_  100");
	for (i=0;i<fxtabpos;i++) {
		if ((tstop <= fxtab[i].time) && (fxtab[i].time < (tstop+0.1)))
			fprintf(op," %2d %d", (int)(100*(fxtab[i].time-tstop)/0.1),(int)fxtab[i].fxval);
	}
	fprintf(op,"\n");
}

void uttprocess(FILE *op,NSL_Item *uttitem)
{
	NSL_Item 	*fileitem;
	NSL_Item 	*ipitem;

	/* find name of SFS file */
	if (realfx) {
		fileitem = MyObtainItem(uttitem->data,"FILE");
		sfilename = strsave(GetAttrStringVal(fileitem,attrHREF));

		/* open the SFS file and load fx contour */
		if (!sfilename || !sfilename[0])
			error("no SFS file given");
		getitem(sfilename,FX_TYPE,fxtype,&fxitem,(void *)&fx);
		smoothfx(fx,fxitem.numframes);
	}

	/* find the IP item */
	ipitem = MyObtainItem(uttitem->data,"IP");

	/* check */
	if (!ipitem)
		error("No IP item in input");

	/* OK, let's go */
	process(ipitem,op,GetAttrStringVal(uttitem,attrID));

	/* free Fx buffer */
	if (fx) {
		free(fx);
		fx=NULL;
	}
	if (fxtab) {
		free(fxtab);
		fxtablen=0;
	}
}

void finduttprocess(FILE *op,NSL_Item *item)
{
	NSL_Data	*data;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,"UTT")==0)
				uttprocess(op,item);
			else
				finduttprocess(op,item);
		}
	}
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	int		it;
	char		*ty;
	NSL_File 	inf;
	
	NSL_Item 	*uttitem;
	NSL_Bit		*bit;
	int		i;
	FILE		*op;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ii:f"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance output as MBROLA commands Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'i' :	/* input specification */
		if (itspec(optarg,&it,&ty)==0) {
			if (it==FX_TYPE)
				fxtype=ty;
			else
				error("unsuitable item specification: '%s'",optarg);
		}
		else
			error("illegal item specification: '%s'",optarg);
		break;
	case 'f':
		realfx=1;
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-i item) (-f) (input.xml) (out.pho)",PROGNAME);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, dct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
		optind++;
	}
	else {
		strcpy(filename,"STDIN");
		if ((inf = SFFopen(stdin, dct, NSL_read,filename))==NULL)
			error("unable to open '%s'",filename);
	}
	dct = DoctypeFromFile(inf);

	/* get unique names for attributes */
	attrHREF = AttrUniqueName(dct,"HREF",0);
	attrID = AttrUniqueName(dct,"ID",0);
	attrSTART = AttrUniqueName(dct,"START",0);
	attrSTOP = AttrUniqueName(dct,"STOP",0);
	attrVOI = AttrUniqueName(dct,"VOI",0);
	attrWEIGHT = AttrUniqueName(dct,"WEIGHT",0);
	attrDUR = AttrUniqueName(dct,"DUR",0);
	attrMINDUR = AttrUniqueName(dct,"MINDUR",0);
	attrINHDUR = AttrUniqueName(dct,"INHDUR",0);
	attrIPITCH = AttrUniqueName(dct,"IPITCH",0);
	attrFPITCH = AttrUniqueName(dct,"FPITCH",0);
	attrPON = AttrUniqueName(dct,"PON",0);
	attrPOF = AttrUniqueName(dct,"POF",0);
	attrLON = AttrUniqueName(dct,"LON",0);
	attrAMBI = AttrUniqueName(dct,"AMBI",0);
	for (i=0;i<NFEATNAME;i++)
		attrFEAT[i] = AttrUniqueName(dct,featname[i],0);

	/* create table of MBROLA symbols */
	createmtable();

	/* get input file */
	if (optind < argc) {
		if ((op=fopen(argv[optind],"w"))==NULL)
			error("unable to open '%s'",argv[optind]);
	}
	else
		op = stdout;

	/* read first tag - should be <UTT> or <PROSYNTH> */
	while ((bit=GetNextBit(inf))!=NULL) {

		if ((bit->type!=NSL_start_bit) || ((strcmp(bit->label,"UTT"))&&(strcmp(bit->label,"PROSYNTH"))))
			error("Could not read <UTT> or <PROSYNTH> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		if (strcmp(uttitem->label,"UTT")==0)
			uttprocess(op,uttitem);
		else
			finduttprocess(op,uttitem);
			

	}

	/* that's all folks */
	if (op!=stdout) fclose(op);
	SFclose(inf);
	return 0;
}
